#include <string.h>
#include <stdio.h>

#include "constantes.h"
#include "types.h"
#include "allocs.h"  
#include "utils.h"  

extern domain_t Domain;

/*****************************************************************************/
void * Malloc (size_t size)
{
    void * p;

    if (NULL == (p = malloc (size)))
        do_error ("allocation problem (malloc)");

    return p;

} // Malloc ()


/*****************************************************************************/
char * fdpstrdup (const char * s)
{
    char * d;

    if (NULL == (d = strdup (s)))
        do_error ("string duplication failed");

    return d;

} // fdpstrdup ()

/*****************************************************************************/

/**
 * Libre les maillons (seulement) de la liste nl
 * NE LIBERE PAS LES NOMS POINTES PAR LA LISTE
 **/

void free_name_list (name_list nl)
{
    name_list n;

    for (n = nl; nl; nl = n)
    {
        n = nl->next;
        free (nl);
    }

} // free_name_list ()

/*****************************************************************************/

lit_list new_literal_list (pred_list pred, term_list terms, lit_list next)
{
    lit_list ll = (lit_list) Malloc (sizeof (lit_list_elt));

    ll->signe       = POSITIF;
    ll->predicate   = pred;
    ll->terms       = terms;
    ll->next        = next;

    return ll;

} // new_literal_list()


/*****************************************************************************/

func_term new_functional_term (func_list func, term_list terms)
{
    func_term ft = (func_term) Malloc (sizeof (func_term_elt));

    ft->function    = func;
    ft->terms       = terms;

    return ft;

} // new_functional_term()


/*****************************************************************************/
/*
    Le prdicat nom prend en arguments des objets dont le type est prcis
    par l'lment correspondant de tvl
*/

pred_list new_predicate_list (dico_list name, tvar_list tvl, pred_list next)
{
    tvar_list tmp_tvl;
    var_list  tmp_vl;
    pred_list tmp_pl;

    tmp_pl = (pred_list) Malloc (sizeof (pred_list_elt));

    /* arity computation */
    int arity = 0;
    for (tmp_tvl = tvl; tmp_tvl; tmp_tvl = tmp_tvl->next)
    {
        for (tmp_vl = tmp_tvl->vars; tmp_vl; tmp_vl = tmp_vl->next) ++arity;

    }

    tmp_pl->name      = name;
    tmp_pl->arguments = tvl;
    tmp_pl->arity     = arity;
    tmp_pl->next      = next;

    return tmp_pl;

} // new_predicate_list()


/*****************************************************************************/

func_list new_function_list (dico_list name, tvar_list tvl, name_list type,
                             func_list next)
{
    tvar_list tmp_tvl;
    var_list  tmp_vl;
    func_list tmp_fl;

    tmp_fl = (func_list) Malloc (sizeof (func_list_elt));

    /* arity computation */
    int arity = 0;
    for (tmp_tvl = tvl; tmp_tvl; tmp_tvl = tmp_tvl->next)
    {
        for (tmp_vl = tmp_tvl->vars; tmp_vl; tmp_vl = tmp_vl->next) ++arity;

    }

    tmp_fl->name      = name;
    tmp_fl->arguments = tvl;
    tmp_fl->arity     = arity;
    tmp_fl->type      = names2reftypes (type);
    tmp_fl->values    = NULL;
    tmp_fl->next      = next;

    return (tmp_fl);

} // new_function_list()

/*****************************************************************************/

numeric_term_value new_numeric_term_value (term_list terms, int value, 
                                           numeric_term_value next)
{
    numeric_term_value ntv;

    ntv = (numeric_term_value) Malloc (sizeof (numeric_term_value_elt));

    ntv->terms = terms;
    ntv->value = value;
    ntv->next  = next;

    return (ntv);

} // new_numeric_term_value()

/*****************************************************************************/

numeric_term new_numeric_term (int type, void * value)
{
    numeric_term nt;

    nt = (numeric_term) Malloc (sizeof (numeric_term_elt));
    
    nt->type = type;

    if (type == NUMERIC_TERM_VALUE)
        nt->value = *((int *) value);
    else
        nt->functerm = (func_term) value;

    return (nt);
}

/*****************************************************************************/

/**
 * comme str2token(), ne sert que pour les requirements.
 * Libre la liste et les chanes pointes.
 **/
void free_token_list (token_list tlist)
{
    token_list temp;
    while(tlist)
    {
        temp = tlist->next;
        free (tlist->item);  // libration de la chane
        free (tlist);        // libration maillon
        tlist = temp;
    }

} // free_token_list()

/*****************************************************************************/

/**
 * str2token() ne sert que pour les requirements (et ce qui ne doit pas tre
 * plac dans le dictionnaire).
 **/
token_list str2token (char * str, token_list next)
{
  token_list tlist;

  tlist = (token_list) Malloc (sizeof (token_list_elt));
  tlist->item = str;
  tlist->next = next;
  return tlist;

} // str2token()

/*****************************************************************************/

dico_list new_dico_entry (const char * s, dico_list next)
{
    dico_list new = (dico_list) Malloc (sizeof (dico_list_elt));

    new->entry = fdpstrdup (s);
    new->next  = next;

    return new;

} // new_dico_entry ()

/*****************************************************************************/

type_list new_type (dico_list name, type_list next)
{
    type_list t = (type_list) Malloc (sizeof (type_list_elt));

    t->name     = name;
    t->is_union = 0;
    t->subtypes = NULL;
    t->objects  = NULL;
    t->next     = next;

    return t;

} // new_type()

/*****************************************************************************/

name_list dico2name (dico_list d, name_list n)
{
    name_list new = (name_list) Malloc (sizeof (name_list_elt));

    new->item = d;
    new->next = n;

    return new;

} // dico2name()

/*****************************************************************************/

tname_list new_tname_list (name_list names, name_list type, tname_list next)
{
    tname_list tnl = Malloc (sizeof (tname_list_elt));

    tnl->names = names;
    tnl->type  = type;
    tnl->next  = next;

    return tnl;

} // new_tname_list()

/*****************************************************************************/

ref_type_list new_ref_type_list (type_list type, ref_type_list next)
{
    ref_type_list rtf = (ref_type_list) Malloc (sizeof (ref_type_list_elt));
    rtf->type = type;
    rtf->next = next;
    return rtf;

} // new_ref_type_list()

/*****************************************************************************/

act_list new_act_list (dico_list name, int status,
                       tvar_list params, lit_list pre,
                       lit_list  add,  lit_list  del, 
                       numeric_term cost,
                       int      contradicts, 
                       act_list next)
{
    act_list al = (act_list) Malloc (sizeof (act_list_elt));

    al->name        = name;
    al->status      = status;
    al->contradicts = contradicts;
    al->params      = params;
    al->pre         = pre;
    al->add         = add;
    al->del         = del;
    al->cost        = cost;
    al->next        = next;

    return al;

} // new_act_list()


/*****************************************************************************/
/*
    cre une liste de variable types par 'type' qui peut tre une union
    Transforme 'type' en une liste de rfrences
 */

tvar_list new_tvar_list (var_list vars, name_list type, tvar_list next)
{
    var_list v;

    tvar_list tvl = (tvar_list) Malloc (sizeof (tvar_list_elt));
    tvl->vars = vars;
    tvl->type = names2reftypes (type);
    tvl->next = next;

    free_name_list (type);

    /* Ajout du 17/03/05. Chaque variable doit pointer sur son type */
    /* afin de vrifier qu'elle est bien unifie avec un bon objet  */
    for (v = vars; v; v = v->next) v->type = tvl->type;

    return tvl;

} // new_tvar_list()


/*****************************************************************************/

var_list dico2variable (dico_list name, var_list next)
{
    var_list vl = (var_list) Malloc (sizeof (var_list_elt));

    vl->name          = name;
    vl->instance      = NULL;   // dico_list
    vl->constraints   = NULL;   // const_list
    vl->type          = NULL;   // ref_type_list
    vl->next          = next;

    return vl;

} // dico2variable()


/*****************************************************************************/
/* Un terme est soit un objet, soit une variable                             */
/* type : TERM_OBJ ou TERM_VAR                                               */
/* item : name_list ou var_list                                              */
/*****************************************************************************/
term_list new_term_list (int type, void * item, term_list next)
{
    term_list tl = (term_list) Malloc (sizeof (term_list_elt));
    tl->type     = type;
    tl->item     = item;
    tl->next     = next;

    return tl;

} // new_term_list()


/*****************************************************************************/
constraint new_constraint (int type, var_list m1, void * m2)
{
    constraint c = (constraint) Malloc (sizeof (constraint_elt));
    c->type      = type;
    c->m1        = m1;
    c->m2        = m2;

    return c;

} // new_constraint()


/*****************************************************************************/
const_list new_const_list (constraint cstr, const_list next)
{
    const_list cl = (const_list) Malloc (sizeof (const_list_elt));
    cl->cstr      = cstr;
    cl->next      = next;

    return cl;

} // new_const_list()


/*****************************************************************************/
sorted_lit_list new_sorted_lit_list (lit_list litx, sorted_lit_list next)
{
    sorted_lit_list sll = (sorted_lit_list) Malloc (sizeof (sorted_lit_list));
    sll->litx           = litx;
    sll->next           = next;

    return sll;

} // new_sorted_lit_list()


/*****************************************************************************/
lit_list dup_litteral (lit_list l)
{
    term_list t, * nt;

    lit_list nl     = new_literal_list (l->predicate, NULL, l->next);
    nl->signe       = l->signe;
    nt              = &(nl->terms);

    for (t = l->terms; t; t = t->next)
    {
        *nt = new_term_list (TERM_OBJ, dico2name (term_dico (t), NULL), NULL);
        nt  = &((*nt)->next);
    }

    *nt = NULL;

    return nl;

} // dup_litteral()

