%{
/*****************************************************************/
/*  PDDL parser for FDP                                          */
/*                                                               */
/*           Cyril Pain-Barre and Stephane Grandcolas            */
/*  cyril.pain-barre@lsis.org     stephane.grandcolas@lsis.org   */
/*****************************************************************/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "constantes.h"
#include "types.h"
#include "allocs.h"
#include "utils.h"
#include "stack.h"
#include "structures.h"

extern lit_list current_litx;
extern char * NomFicDomaine;
extern char * NomFicProbleme;
extern FILE * FicDomaine;
extern FILE * FicProbleme;
extern int nSteps;
extern int nStepsInProblem;
char * nomficourant;

stack_t   stack;  /* pile des modifications */
stack_t   file;   /* file des variables modifies  traiter */
problem_t Problem;
domain_t  Domain;
char error  [200];
char chaine [200];
int yylex (void);
int yyerror(char * s);
int num_value = 0;
char *endptr;

extern char * yytext;
extern int    yylineno;
extern FILE * yyin;

void check_metric_spec      (dico_list met);
void check_problem      (dico_list dom);
void check_requirements  (token_list flags);
void append_functions_in_domain (func_list funcl);
void skip_argument (void);

numeric_term crt_action_cost = NULL;  // current action cost
tvar_list crt_vars = NULL;  // parameters of the current action
pred_list crt_pred = NULL;  // predicate of the current literal
func_list crt_func = NULL;  // function of the current numeric term
int crt_arity = 0; // arity of the predicate/function (of the current literal
                   //     or numeric term) 
char * crt_pred_or_func_name_entry;

struct                      // argument en cours de traitement dans rfrence
{                           //  prdicat
    var_list   var;
    tvar_list  tvl;
} crt_arg = { NULL, NULL };


act_list lastact = NULL;

%}

%union {
    tname_list   tnamel;
    tvar_list    tvarl;
    var_list     varl;
    act_list     actl;
    name_list    namel;
    dico_list    dico;
    token_list   tokenl;
    pred_list    predl;
    func_list    funcl;
    lit_list     litl;
    term_list    terml;
    entry_t      str;
    numeric_term numt;
}

%start pddl

%token DEFINE
%token DOMAIN REQUIREMENTS CONSTANTS TYPES PREDICATES
%token PROBLEM OBJECTS INIT LENGTH PARALLEL SERIAL
%token CONTROL DEFPREDICATE EXCLUDE WFFCTRL SCOPE BODY 
%token ACTIONS PARAMETERS PRECONDITION EFFECT
%token AND 
%token FORALL EXISTS NOEXISTS 
%token FUNCTIONS INCREASE METRIC MINIMIZE
/*
    Mis en commentaire le 06/10/2005 car le prdicat
    'next' existe dans le domaine zeno-travel de IPC-3

    %token NEXT 
*/
%token EQ 
%token NOT GOAL
%token EITHER FLUENT
%token ID VAR

%%


pddl : 
        domain
     |  problem
     ;


domain :  '(' DEFINE '(' DOMAIN id ')' domain_body ')'
            {
                Domain.name = $<dico>5->entry;
            }
       ;

domain_body : 
        | domain_body '(' ':' domain_member ')' 
        ;

domain_member :
      REQUIREMENTS require_keys
        {
            check_requirements ($<tokenl>2);
            free_token_list ($<tokenl>2);
        }
    | TYPES typed_name_list
        {   
            trt_types_def ($<tnamel>2); 
        }
    | PREDICATES atomic_formulas_skeletons
        {
            Domain.predicates->next = $<predl>2;
            
        }
    | FUNCTIONS typed_function_skeletons
        {
            append_functions_in_domain ($<funcl>2); 
        }
    | CONSTANTS typed_name_list
        {
            trt_objects_def ($<tnamel>2);
        }
    | ACTIONS
        {
            crt_action_cost = NULL;
        }
      action_suite
    ;

typed_function_skeletons :
            {   $<funcl>$ = NULL;    }
    |  typed_function_skeleton typed_function_skeletons
            {
                $<funcl>$       = $<funcl>1;
                $<funcl>$->next = $<funcl>2;
            }
    ;

typed_function_skeleton :
    '(' id typed_variable_list ')'
        {
            $<funcl>$ = new_function_list ($<dico>2, $<tvarl>3,
                                dico2name(get_dico_entry("NUMBER"), NULL), 
                                NULL);
        }
    |
    '(' id typed_variable_list ')' '-' type
        {
            $<funcl>$ = new_function_list ($<dico>2,  $<tvarl>3, 
                                            $<namel>6, NULL);
        }
    ;


action_suite :
      id
      effect
          {
            $<actl>$ = trt_act_def ($<dico>1, NULL, NULL, $<litl>2, 
                                    crt_action_cost);
            if (lastact == NULL)
                Domain.actions = $<actl>$;
            else
                lastact->next = $<actl>$;
            lastact = $<actl>$;
          }
    |
      id
      parameter
      ':' PRECONDITION conjunction 
      effect
        {
            $<actl>$ = trt_act_def ($<dico>1, $<tvarl>2, $<litl>5, $<litl>6, 
                                    crt_action_cost);
            if (lastact == NULL)
                Domain.actions = $<actl>$;
            else
                lastact->next = $<actl>$;
            lastact = $<actl>$;
        }
    |
      id
      parameter
      effect
        {   
            $<actl>$ = trt_act_def ($<dico>1, $<tvarl>2, NULL, $<litl>3, 
                                    crt_action_cost);
            if (lastact == NULL)
                Domain.actions = $<actl>$;
            else
                lastact->next = $<actl>$;
            lastact = $<actl>$;
        }
    ;


cost : '(' INCREASE '(' ID
            {
                if (strcasecmp (yytext, "total-cost"))
                {
                    snprintf(error, sizeof(error), 
                             "increase of %s is not allowed",
                             yytext);
                    do_error(error);
                }
            }
       ')' numeric_term ')'
            {
                $<numt>$ = $<numt>7;
            }
    ;

numeric_term : 
        ID
            {
                num_value = strtol(yytext, &endptr, 10);
                if (*endptr != '\0')
                {
                    snprintf(error, sizeof(error), 
                             "wrong number %s", yytext);
                    do_error(error);
                }
                $<numt>$ = new_numeric_term(NUMERIC_TERM_VALUE, 
                                            (void *) &num_value);
            }
    |
        '(' ID
            {
                crt_arity = 0;
                crt_func = find_function (yytext);
                crt_pred_or_func_name_entry = crt_func->name->entry;
                if (NULL == crt_func->arguments)
                {
                    crt_arg.var = NULL;
                    crt_arg.tvl = NULL;
                }
                else
                {
                    crt_arg.var = crt_func->arguments->vars;
                    crt_arg.tvl = crt_func->arguments;
                }
            }
        terms ')'
            {
                if (crt_arity != crt_func->arity)
                {
                    snprintf(error, sizeof (error), 
                             "illegal use of function %s: unmatching arities",
                             crt_func->name->entry);
                    do_error(error);
                }
                $<numt>$ = new_numeric_term(NUMERIC_TERM_FUNCT,
                                    new_functional_term(crt_func, $<terml>4));
            }
    ;

require_keys :
        require_key
            {   $<tokenl>$ = str2token ($<str>1, NULL); }
    |   require_key require_keys
            { 
                $<tokenl>$ = str2token ($<str>1, $<tokenl>2);
            }
    ;


require_key : ':' ID
        { $<str>$ = fdpstrdup (yytext); }
    ;


typed_name_list :
        { $<tnamel>$ = NULL; }
    | id_list
        { $<tnamel>$ = new_tname_list ($<namel>1, NULL, NULL); }

    | id_list '-' type typed_name_list
        { $<tnamel>$ = new_tname_list ($<namel>1, $<namel>3, $<tnamel>4); }
    ;

id_list :
      id
        { $<namel>$ = dico2name ($<dico>1, NULL); }
    | id id_list
        { $<namel>$ = dico2name ($<dico>1, $<namel>2); }
    ;


type :
      id
        { $<namel>$ = dico2name ($<dico>1, NULL); }
    | '(' EITHER types ')'
        { $<namel>$ = $<namel>3; }
    ;


types :
      type
        {   $<namel>$ = $<namel>1; }
    | type types
        {
            $<namel>$       = $<namel>1;
            $<namel>$->next = $<namel>2;
        }
    ;

object_declaration :
    '(' ':' OBJECTS typed_name_list ')'
        {
            trt_objects_def ($<tnamel>4);
        }
    ;


atomic_formulas_skeletons :
        atomic_formula_skeleton
            {   
                $<predl>$ = $<predl>1;
            } 
    |
        atomic_formula_skeleton atomic_formulas_skeletons
            { 
                $<predl>$       = $<predl>1;
                $<predl>$->next = $<predl>2;
            }
    ;

atomic_formula_skeleton :
        '(' id typed_variable_list ')'
            {   
                $<predl>$ = new_predicate_list ($<dico>2, $<tvarl>3, NULL);  
            }  
    ;

typed_variable_list :
        { $<tvarl>$ = NULL; }
    | var_list
        {   $<tvarl>$ = new_tvar_list ($<varl>1, NULL, NULL); }

    | var_list '-' type typed_variable_list
        {   $<tvarl>$ = new_tvar_list ($<varl>1, $<namel>3, $<tvarl>4); }
    ;


var_list :
      variable
        {  $<varl>$ = dico2variable ($<dico>1, NULL); }

    | variable var_list
        {   $<varl>$ = dico2variable ($<dico>1, $<varl>2); }
    ;


id :
      ID
        { $<dico>$ = get_dico_entry (yytext); } 
    ;


variable :
      VAR 
        { $<dico>$ = get_dico_entry (yytext); }
    ;


parameter :
    ':' PARAMETERS '(' typed_variable_list ')'
        { 
            crt_vars = $<tvarl>$ = $<tvarl>4; 
        }
    ;


effect :
    ':' EFFECT conjunction_or_cost
        { 
            $<litl>$ = $<litl>3; 
        }
    ;

conjunction_or_cost :
        { $<litl>$ = NULL; }
    |
      literal_or_cost
        { $<litl>$ = $<litl>1; }
    | '(' AND literal_or_cost ')'
        { $<litl>$ = $<litl>3; }
    | '(' AND literal_or_cost literal_or_cost_list ')'
        { 
            if ($<litl>3 == NULL)
            {
                $<litl>$ = $<litl>4;
            }
            else
            {
                $<litl>$       = $<litl>3;
                $<litl>$->next = $<litl>4;
            }
        }
    ;

conjunction :
        { $<litl>$ = NULL; }
    |
      literal
        { $<litl>$ = $<litl>1; }
    | '(' AND literal ')'
        { $<litl>$ = $<litl>3; }
    | '(' AND literal literal_list ')'
        { 
            $<litl>$       = $<litl>3;
            $<litl>$->next = $<litl>4;
        }
    ;

literal_or_cost :
      cost
        {
            if (crt_action_cost != NULL)
            {
                snprintf(error, sizeof(error), 
                         "only one action cost is allowed per action");
                do_error(error);
            }
            crt_action_cost = $<numt>1;
            $<litl>$ = NULL;
        }
    | 
              atomic_formula
        {   $<litl>$ = $<litl>1; }
    | '(' NOT atomic_formula ')'
        { 
            $<litl>$        = $<litl>3; 
            $<litl>$->signe = NEGATIF;
        }
    ;

literal : 
              atomic_formula
        {   $<litl>$ = $<litl>1; }
    | '(' NOT atomic_formula ')'
        { 
            $<litl>$        = $<litl>3; 
            $<litl>$->signe = NEGATIF;
        }
    ;

literal_or_cost_list :
      literal_or_cost
        { $<litl>$ = $<litl>1; }
    | literal_or_cost literal_or_cost_list
        { 
            $<litl>$       = $<litl>1;
            $<litl>$->next = $<litl>2;
        }
    ;

literal_list :
      literal
        { $<litl>$ = $<litl>1; }
    | literal literal_list
        { 
            $<litl>$       = $<litl>1;
            $<litl>$->next = $<litl>2;
        }
    ;

atomic_formula :
    '(' ID
        {
            crt_arity = 0;
            crt_pred = find_predicate (yytext);
            crt_pred_or_func_name_entry = crt_pred->name->entry;
            if (NULL == crt_pred->arguments)
            {
                crt_arg.var = NULL;
                crt_arg.tvl = NULL;
            }
            else
            {
                crt_arg.var = crt_pred->arguments->vars;
                crt_arg.tvl = crt_pred->arguments;
            }
        } 
    terms ')'
        {
            if (crt_arity != crt_pred->arity)
            {
                snprintf(error, sizeof (error), 
                         "illegal use of predicate %s : unmatching arities",
                         crt_pred->name->entry);
                do_error(error);
            }
            $<litl>$ = new_literal_list (crt_pred, $<terml>4, NULL);
        }
    ;

terms :
        { $<namel>$ = NULL; }
    | term
        { skip_argument (); crt_arity++; } 
      terms
        { 
            $<terml>$       = $<terml>1;
            $<terml>$->next = $<terml>3;
        }
    ;

term :
      id
        {
            if (NULL == crt_arg.var)
            {
                snprintf (error, sizeof(error),
                          "%s: too many parameters for predicate/function %s",
                          yytext, crt_pred_or_func_name_entry);
                do_error (error);
            }

            $<terml>$ = new_term_list (TERM_OBJ,
                                       get_object ($<dico>1, 
                                                   crt_arg.tvl->type), 
                                       NULL);
        }
    | variable
        {
            if (NULL == crt_arg.var)
            {
                snprintf (error, sizeof(error),
                          "%s: too many parameters for predicate/function %s",
                          yytext, crt_pred_or_func_name_entry);
                do_error (error);
            }

            $<terml>$ = new_term_list (TERM_VAR,
                                       get_variable ($<dico>1,
                                                     crt_vars,
                                                     crt_arg.tvl->type),
                                       NULL);
        }
    ;

/**************************************************************************/
/* DEFINITION PROBLEME */
/**************************************************************************/

problem : '(' DEFINE '(' PROBLEM id ')'
                    {
                        Problem.name   = $<dico>5->entry;
                    }
                problem_body
          ')'
        ;

problem_body : 
                            problem_member
        |   problem_body    problem_member
        ;

problem_member : 
            '(' ':' DOMAIN id ')'
                {   
                    check_problem ($<dico>4);
                    Problem.domain = Domain;
                }
        |   object_declaration
            {
            }
        |   init
            {
                Problem.init   = $<litl>1; 
            }
        |   goals
            {
                Problem.goal   = $<litl>1;
            }
        |   length_spec
        |   metric_spec
        ;


length_spec : '(' ':' LENGTH length_body ')'
        ;

length_body :
                    length_member
    | length_body   length_member
    ;

length_member :
        '(' ':' SERIAL   ID 
            {
                nSteps = atoi (yytext);
                nStepsInProblem = 1;
            }
        ')'
    |   '(' ':' PARALLEL ID ')'
    ;

metric_spec : '(' ':' METRIC MINIMIZE '(' id ')' ')'
            {
                check_metric_spec ($<dico>6);
            }
    ;

ground_atomic_formula :
      '(' ID
            {
                crt_pred = find_predicate (yytext);
                if (NULL == crt_pred->arguments)
                {
                    crt_arg.var = NULL;
                    crt_arg.tvl = NULL;
                }
                else
                {
                    crt_arg.var = crt_pred->arguments->vars;
                    crt_arg.tvl = crt_pred->arguments;
                }
            } 
      names ')'
        { $<litl>$ = new_literal_list (crt_pred, $<terml>4, NULL); }
    ;

ground_atomic_formula_or_init_cost :
      '(' ID
            {
                crt_arity = 0;
                crt_pred = find_predicate (yytext);
                crt_pred_or_func_name_entry = crt_pred->name->entry;
                if (NULL == crt_pred->arguments)
                {
                    crt_arg.var = NULL;
                    crt_arg.tvl = NULL;
                }
                else
                {
                    crt_arg.var = crt_pred->arguments->vars;
                    crt_arg.tvl = crt_pred->arguments;
                }
            }
        ground_atomic_formula_or_init_cost_suite
            {
                $<litl>$ = $<litl>4;
            }
        ;

ground_atomic_formula_or_init_cost_suite :
      '(' ID  
        {
            if (crt_pred != Domain.equals)
            {
                snprintf (error, sizeof(error),
                          "only '=' is allowed in :init for functions");
                do_error (error);
            }

            crt_arity = 0;
            crt_func = find_function(yytext);
            crt_pred_or_func_name_entry = crt_func->name->entry;
            if (NULL == crt_func->arguments)
            {
                crt_arg.var = NULL;
                crt_arg.tvl = NULL;
            }
            else
            {
                crt_arg.var = crt_pred->arguments->vars;
                crt_arg.tvl = crt_pred->arguments;
            }
        }
      names ')' ID
        {
            num_value = strtol (yytext, &endptr, 10);
            if (*endptr != '\0')
            {
                snprintf (error, sizeof (error),
                         "%s : illegal value, integer expected",
                          yytext);
                do_error (error);
            }
        }
      ')'
        {
            if (crt_arity != crt_func->arity)
            {
                snprintf (error, sizeof (error),
                         "illegal use of function %s : unmatching arities",
                         crt_func->name->entry);
                do_error (error);
            }

            add_value_for_numeric_term (crt_func, $<terml>4, num_value);
            $<litl>$ = NULL;
        }
    | 
      names ')'
        {
            if (crt_arity != crt_pred->arity)
            {
                snprintf (error, sizeof (error),
                         "illegal use of predicate %s : unmatching arities",
                         crt_func->name->entry);
                do_error (error);
            }
                 
            $<litl>$ = new_literal_list (crt_pred, $<terml>1, NULL);
        }
    ;

names :
        { $<namel>$ = NULL; }
    | name
        { skip_argument (); crt_arity++; } 
      names
        { 
            $<terml>$       = $<terml>1;
            $<terml>$->next = $<terml>3;
        }
    ;

name :
      id
        {
            if (NULL == crt_arg.var)
            {
                snprintf (error, sizeof(error),
                         "exceedind parameter %s in predicate/function %s",
                         yytext, crt_pred_or_func_name_entry);
                do_error (error);
            }

            $<terml>$ = new_term_list (
                            TERM_OBJ,
                            find_or_create_object ($<dico>1, 
                                                   crt_arg.tvl->type),
                            NULL);
        }
    ;

ground_literal :
                  ground_atomic_formula
            {   $<litl>$ = $<litl>1;   }
        | '(' NOT ground_atomic_formula ')'
            { 
                $<litl>$        = $<litl>3; 
                $<litl>$->signe = NEGATIF;
            }
        ;

ground_literal_list : 
            ground_literal
                { $<litl>$ = $<litl>1; }
        |   ground_literal ground_literal_list
                {
                    $<litl>$       = $<litl>1;
                    $<litl>$->next = $<litl>2;
                }
        ;

ground_literal_or_init_cost :
          ground_atomic_formula_or_init_cost
            {   $<litl>$ = $<litl>1;   }
        | '(' NOT ground_atomic_formula ')'
            { 
                $<litl>$        = $<litl>3; 
                $<litl>$->signe = NEGATIF;
            }
        ;
            

ground_literal_or_init_cost_list :
            { $<litl>$ = NULL; }
        |
            ground_literal_or_init_cost ground_literal_or_init_cost_list
                {
                    if ($<litl>1 == NULL)
                        $<litl>$ = $<litl>2;
                    else
                    {
                        $<litl>$       = $<litl>1;
                        $<litl>$->next = $<litl>2;
                    }
                }
        ;

 

init    : '(' ':' INIT ground_literal_or_init_cost_list ')'
            { $<litl>$ = $<litl>4; }
        ;

goals   :  '(' ':' GOAL '(' AND ground_literal_list ')' ')'
            {
                $<litl>$ = $<litl>6; 
            }
        | '(' ':' GOAL ground_literal_list ')'
            {
                $<litl>$ = $<litl>4;
            }
        ;


%%


void check_metric_spec (dico_list met)
{
    if ( strcasecmp (met->entry, "total-cost") )
    {
        snprintf(error, sizeof (error), 
                 "metric minimization of %s not supported",
                 met->entry);
        do_error (error);
    }

} // check_metric_spec()

void check_problem (dico_list dom)
{
    if ( strcasecmp (dom->entry, Domain.name) )
    {
        snprintf (error, sizeof (error),
                  "Problem %s requires unknown domain %s",
                  Problem.name, dom->entry);
        do_error (error);
    }

} // check_problem()

void check_requirements (token_list flags)
{
    for (; flags; flags = flags->next) 
    {
        if (     strcasecmp("strips",   flags->item) != 0
    	      && strcasecmp("typing",   flags->item) != 0
              && strcasecmp("equality", flags->item) != 0 
              && strcasecmp("action-costs", flags->item) != 0 
           )
        {
            snprintf(error, sizeof (error), 
                     "requirement: %s is not supported", 
                     flags->item);
            do_error (error);
        } 
    }
} // check_requirements()


/* modifie crt_arg pour qu'il dsigne le prochain argument s'il y en a un */
void skip_argument (void)
{
    crt_arg.var = crt_arg.var->next;
    if (NULL == crt_arg.var)
    {
        crt_arg.tvl = crt_arg.tvl->next;
        if (NULL != crt_arg.tvl)
            crt_arg.var = crt_arg.tvl->vars;
    }

} // skip_argument ()

void append_functions_in_domain (func_list funcl)
{
    func_list dfl;

    if (Domain.functions == NULL)
    {
        Domain.functions = funcl;
        return;
    }

    for (dfl = Domain.functions; dfl->next; dfl = dfl->next);
    dfl->next = funcl;

} // append_functions_in_domain()

int yyerror(char * s)
{
  fprintf(stderr,"%s\n",s);
  fprintf(stderr,"Error at or near line %d file %s\n", yylineno, nomficourant);
  exit(1);
}

int analyser ()
{
    stack = alloc_stack (1000);  /* pile des modifications */
    file  = alloc_stack (1000);  /* pile des variables  vrifier */

    lf_InitState.tete = lf_InitState.queue = NULL;
    lf_Goals.tete     = lf_Goals.queue     = NULL;

    Domain.object = new_type (get_dico_entry ("OBJECT"), NULL);
    Domain.types  = new_type (get_dico_entry ("NUMBER"), Domain.object);

    /* 
        Ajout du predicat d'galit, comme s'il avait t dclar ainsi :
        (:predicates  (= ?obj1 ?obj2 - OBJECT))
    */
    Domain.equals = 
        new_predicate_list 
            (   
                get_dico_entry ("="),
                new_tvar_list 
                    (   
                        dico2variable 
                            (   
                                get_dico_entry ("?obj1"), 
                                dico2variable (get_dico_entry ("?obj2"), NULL)
                            ),    
                        dico2name (get_dico_entry ("OBJECT"), NULL), 
                        NULL
                    ),
                NULL
            );
    Domain.predicates = Domain.equals;

    Domain.functions  = NULL;

    #if (DEBUG_PDDL_PARSER)    
        printf ("Analyse du domaine : en cours\n");
    #endif

    nomficourant = NomFicDomaine;
    yyin = FicDomaine;
    yyparse ();

    #if (DEBUG_PDDL_PARSER)
        printf ("Analyse du domaine : termine\n");
        printf ("Analyse du probleme : en cours\n");
    #endif

    nomficourant = NomFicProbleme;
    yyin = FicProbleme;
    yylineno = 1;
    yyparse ();

    #if (DEBUG_PDDL_PARSER)
        printf ("Analyse du problme : termine\n");
    #endif

    printf ("(%s/%s) ", Domain.name, Problem.name);
    fflush (stdout);

    display_domain (); 
    display_problem ();

    current_litx = Problem.init;

    InitState = FromListToBooleans  (lf_InitState.tete, SizeAllocFaits);
    Goals     = FromGoalsToBooleans (lf_Goals.tete,     SizeAllocFaits);

    #if DEBUG_VERY_VERBOSE
        int i;
        printf ("Etat initial apres FromListToBooleans :\n");
        for (i=0; i < nFaits*2; ++i)
            if (InitState[i]) printf ("%s\n", LesFaits[i].nom);
        printf ("Fin de l'etat initial\n\n");


        printf ("Buts apres FromGoalsToBooleans :\n");
        for (i=0; i < nFaits*2; ++i)
            if (Goals[i]) printf ("%s\n", LesFaits[i].nom);
        printf ("Fin des buts initial\n\n");
    #endif

    return nSteps;

} // analyser()
