#ifndef    __TYPES_H__
#define    __TYPES_H__

#include "constantes.h"
#include "structures.h"


typedef char * entry_t;


typedef struct TOKENLIST
{
    entry_t             item;
    struct TOKENLIST *  next;

} token_list_elt, * token_list;


typedef struct DICOLIST
{
    entry_t           entry;
    struct DICOLIST * next;

} dico_list_elt, * dico_list;


typedef struct NAMELIST
{
    dico_list           item;
    struct NAMELIST   * next;

} name_list_elt, * name_list;


typedef struct TNAMELIST
{
    name_list           names;
    name_list           type;
    struct TNAMELIST *  next;

} tname_list_elt, * tname_list;


typedef struct REFTYPELIST
{
    struct TYPELIST    *    type;
    struct REFTYPELIST *    next;

} ref_type_list_elt, * ref_type_list;


typedef struct TYPELIST
{
    dico_list           name;
    int                 is_union;
    ref_type_list       subtypes;
    name_list           objects;
    struct TYPELIST *   next;

} type_list_elt, * type_list;


typedef struct VARLIST
{
    dico_list          name;
    dico_list          instance;
    struct CONSTLIST * constraints;
    ref_type_list      type;    /* rajout le 17/03/05 pour vrifier le type */
                                /* de l'objet avec lequel une variable est   */
                                /* unifie                                   */
    struct VARLIST   * next;

} var_list_elt, * var_list;


typedef struct TVARLIST
{
    var_list            vars;
    ref_type_list       type;
    struct TVARLIST *   next;

} tvar_list_elt, * tvar_list;


typedef struct PREDLIST
{
    dico_list           name;
    int                 arity;
    tvar_list           arguments;
    struct PREDLIST *   next;

} pred_list_elt, * pred_list;

typedef struct NUMERICTERMVALUE
{
    struct TERMLIST *         terms;
    int                       value;
    struct NUMERICTERMVALUE * next;

} numeric_term_value_elt, * numeric_term_value;

typedef struct FUNCLIST
{
    dico_list           name;
    int                 arity;
    tvar_list           arguments;
    ref_type_list       type;
    numeric_term_value  values;
    struct FUNCLIST *   next;

} func_list_elt, * func_list;


typedef struct LITLLIST
{
    int                 signe;      // POSITIF ou NEGATIF
    pred_list           predicate;
    struct TERMLIST *   terms;
    struct LITLLIST *   next;

} lit_list_elt, * lit_list;


typedef struct ACTLIST
{
    dico_list            name;
    int                  status;
    int                  contradicts; /* are some effect contradictory */
    tvar_list            params;
    lit_list             pre;
    lit_list             add;
    lit_list             del;
    struct NUMERICTERM * cost;
    struct ACTLIST *     next;

} act_list_elt, * act_list;

typedef struct FUNCTERM
{
    func_list           function;
    struct TERMLIST *   terms;
    
} func_term_elt, * func_term;

typedef struct NUMERICTERM
{
    int     type;
    union {
        int         value;
        func_term   functerm;  
    };

} numeric_term_elt, * numeric_term;


typedef struct DOMAINE
{
    entry_t   name;
    dico_list dictionary;
    type_list types;
    type_list object;
    pred_list predicates;
    pred_list equals;
    func_list functions;
    act_list  actions;

} domain_t;


typedef struct PROBLEME
{
    domain_t    domain;
    entry_t     name;
    lit_list    init;
    lit_list    goal;

} problem_t;


typedef struct SITUATION
{
    FLIST   tete, queue;

} list_fact_t;


#define term_variable(t)    ( (var_list)  ((t)->item) )
#define term_object(t)      ( (name_list) ((t)->item) )
/* element d'un terme selon son type (var_list ou name_list = objet) */
#define term_dico(t)  ( (t)->type == TERM_OBJ ?                    \
                            (term_object(t))->item :               \
                            (term_variable(t))->instance )
#define term_entry(t) ( term_dico(t)->entry )


typedef struct TERMLIST
{
    int                 type;  // TERM_OBJ, TERM_VAR
    void *              item;  // var_list ou name_list
    struct TERMLIST *   next;

} term_list_elt, * term_list;


typedef struct CONSTRAINT
{
    int                 type;  // EQU_VAR, NEQ_VAR, NEQ_OBJ
    var_list            m1;    // m1 est forcment une variable
    void *              m2;    // m2 est soit var_list, soit name_list

} constraint_elt, * constraint;


typedef struct CONSTLIST
{
    constraint          cstr;
    struct CONSTLIST *  next;

} const_list_elt, * const_list;


typedef struct SORTEDLITLIST
{
    lit_list     litx;
    struct SORTEDLITLIST * next;

} sorted_lit_list_elt, * sorted_lit_list;

#endif  /* __TYPES_H__ */
