#include <stdio.h>   // printf()
#include <string.h>  // strdup(), strcasecmp()
#include <assert.h>

#define INCORPORATE_FDP  1
#include "constantes.h"
#include "types.h"
#include "allocs.h"
#include "utils.h"
#include "stack.h"
#include "structures.h"

extern int IPCOutput; /* pour le nom des actions */
extern struct fait *LesFaits;
extern int nActionsIgnorees;
extern int nActionsUnknownDel;
extern stack_t   stack;
extern stack_t   file;
extern domain_t  Domain;
extern problem_t Problem;
extern char error [];

lit_list    current_litx      = NULL;  // littraux dj rencontrs
lit_list    new_litx          = NULL;  // littraux nouveaux aprs extension
int         levels            = 0;     // niveau courant du graphe
unsigned    nbact             = 0;     // nombre d'actions du niveau
unsigned    nblitx            = 0;     // nombre de littraux du niveau

char tmp_chlit  [10000];          // pour un littral
char tmp_chact  [10000];          // pour nom d'action instancie


#if defined    INCORPORATE_FDP

    list_fact_t lf_InitState, lf_Goals; // tats inital et final

    list_fact_t pre, add, del; // listes pour une action
    list_fact_t raz = {NULL, NULL}; // pour remise  zro de pre, add et del

    FLIST new_fact_maillon (int fait, FLIST prec, FLIST suiv)
    {
        FLIST fl = (FLIST) Malloc (sizeof (struct fact_maillon));
        fl->fait = fait;
        fl->prec = prec;
        fl->suiv = suiv;
        return fl;

    } // new_fact_maillon()

    void free_list_fact (list_fact_t * lst)
    {
        FLIST p, q;

        for (p = lst->tete; p; p = q)
        {
            q = p->suiv;
            free (p);
        }

        lst->tete = lst->queue = NULL;

    } // free_list_fact()


    /**
     *  Le fait s doit tre rajout dans la situation sit.
     *  signe vaut 0 si (not s) et 1 sinon.
     **/
    void add_numfact_in_list (int num, int signe, list_fact_t * lst)
    {
        FLIST fl = new_fact_maillon ((signe == POSITIF ? num : OPPOSE (num)), 
                                     lst->queue, NULL);
        if (NULL == lst->tete)
        {
            lst->tete = fl;
        }
        if (lst->queue != NULL)
        {
            lst->queue->suiv = fl;
        }
        lst->queue = fl;

    } // add_numfact_in_list()

    
    /**
     *  Le fait s doit tre rajout dans la situation sit.
     *  signe vaut 0 si (not s) et 1 sinon.
     **/
    void add_fact_in_list (char * s, int signe, int but, list_fact_t * lst, int noadd)
    {
        int   n;

        if (NONE == (n = SearchFact (s, signe)))
        {
            if (noadd != 0)
            {
                printf ("\n%s inexistant mais supprim !!\n",
                        s);
                assert (noadd == 0);
            }
            n = AddFact (s, signe, but);
            nblitx += 1 - but;
        }

        add_numfact_in_list (n, POSITIF, lst);

    } // add_fact_in_list()

    
#endif      /* INCORPORATE_FDP */


int check_and_propagate_unification (var_list v);


void do_error (char *s) 
{
    printf("%s\n",s); exit( 1 );
}


void add_value_for_numeric_term (func_list func, term_list terms, int value)
{
    func->values = new_numeric_term_value (terms, value, func->values);

} // new_value_for_numeric_term()


/* s doit tre dj alloue */
void atomic_formula2string (lit_list l, char * s)
{
    term_list tl;

    sprintf (s, "(%s", l->predicate->name->entry);

    for (tl = l->terms; tl; tl = tl->next)
        sprintf (s, "%s %s", s, term_entry (tl));

    sprintf (s, "%s)", s);

} // atomic_formula2string()

dico_list get_dico_entry (const char * s)
{
    dico_list p, q, n;
    int cmp;

    if (  ( !  Domain.dictionary)
       || ((cmp = strcasecmp (s, Domain.dictionary->entry)) < 0) )
    {
        return (Domain.dictionary = new_dico_entry (s, Domain.dictionary));
    }

    if (! cmp) return Domain.dictionary;

    for (p = Domain.dictionary, q = p->next; q; p = q, q = q->next)
    {
        cmp = strcasecmp (s, q->entry);
        if (! cmp) return q;
        if (cmp < 0) break;
    }

    n = new_dico_entry (s, q);
    p->next = n;

    return n;

} // get_dico_entry ()

/**
 * t est un sous-type de T, dclar ainsi : t - T
 * Rajoute t dans la liste des sous-types de T
 * Cela permettra ensuite de dterminer les objets d'un type en parcourant
 * ses sous-types.
 **/

void add_subtype_of (type_list t, type_list T)
{
    T->subtypes = new_ref_type_list (t, T->subtypes);

} // add_subtype_of()


type_list find_type (const char * s)
{
    type_list t;
    for (t = Domain.types; t ; t = t->next)
    {
        int cmp = strcasecmp (s, t->name->entry);
        if ( ! cmp ) return t;
        if (cmp < 0) return NULL;
    }

    return NULL;

} // find_type()



/**
 *
 * Ajoute le type 'type'  la liste des types du domaine
 * Cette liste est trie par ordre alphabtique
 *
 **/

void add_type (type_list type)
{
    type_list p, q;
    int cmp;

    if ( NULL == Domain.types ||
        (cmp = strcasecmp (type->name->entry, Domain.types->name->entry)) < 0)
    {
        type->next = Domain.types;
        Domain.types = type;
        return;
    }

    if ( ! cmp )
    {
        sprintf (error, "type %s already exists", type->name->entry);
        do_error (error);
    }

    for (p = Domain.types, q = p->next; q; p = q, q = q->next)
    {
        cmp = strcasecmp (type->name->entry, q->name->entry);
        if ( ! cmp )
        {
            sprintf (error, "type %s already exists", type->name->entry);
            do_error (error);
        }
        if (cmp < 0) break;
    }

    type->next = q;
    p->next = type;

} // add_type()

void add_union_member_of (type_list t, type_list T)
{
    // t est un membre de l'union T, i.e. T - (either t ...)
    // t est insr dans les sous-types de T
    
    add_subtype_of (t, T);

} // add_union_member_of()


/**
 * Cre un type 'name' qui est :
 *   + soit un sous-type de "OBJECTS"
 *   + soit un sous-type du singleton "types"
 *   + soit l'union de la liste "types"
 **/

void trt_type_def (name_list name, name_list types)
{
    //printf ("traitement du type %s\n", name->item->entry);
    type_list type = new_type (name->item, NULL);
    if (NULL == types)
    {
        add_subtype_of (type, Domain.object);
        add_type (type);
        return;
    }

    if (NULL == types->next)
    {
        type_list t = find_type (types->item->entry);
        if (NULL == t)
        {
            sprintf (error, "type %s rfrenc mais introuvable",
                     types->item->entry);
            do_error (error);
        }
        if (t->is_union)
        {
            sprintf(error, "union %s ne peut pas admettre sous type %s",
                    types->item->entry, name->item->entry);
            do_error (error);
        }
        add_subtype_of (type, t);
        add_type (type);
        return;
    }

    type->is_union = 1;
    for ( ; types; types = types->next)
    {
        type_list t = find_type (types->item->entry);
        if (NULL == t)
        {
            sprintf (error, "type %s rfrenc mais introuvable",
                     types->item->entry);
            do_error (error);
        }
        add_union_member_of (t, type);
    }
    add_type (type);

} // trt_type_def()



/**
 * tnl contient une liste de types  crer.
 * Les maillons de tnl sont librs.
 **/

void trt_types_def (tname_list tnl)
{
    tname_list t;
    name_list n;
    for (; tnl; tnl = t)
    {
        t = tnl->next;
        for (n = tnl->names; n; n = n->next)
            trt_type_def (n, tnl->type);

        free (tnl);

    }
} // trt_types_def()


/**
 * Insre obj s'il n'existe pas dj dans type (non union)
 **/

void add_object_in (name_list obj, type_list t)
{
    int cmp;
    name_list p, q;

    //printf("adding object %s (type %s)\n", obj->item->entry, t->name->entry);
    if (NULL == t->objects 
        || (cmp = strcasecmp (obj->item->entry, t->objects->item->entry)) < 0)
    {
        obj->next   = t->objects;
        t->objects = obj;
        return;
    }

    if ( ! cmp )
    {
        /*
        printf ("Warning : object %s already exists in type %s\n",
                obj->item->entry, t->name->entry);
        */
        free (obj);
        return;
    }

    for (p = t->objects, q = p->next; q; p = q, q = q->next)
    {
        cmp = strcasecmp (obj->item->entry, q->item->entry);
        if ( ! cmp )
        {
            /*
            printf ("Warning : object %s already exists in type %s\n",
                obj->item->entry, t->name->entry);
            */
            free (obj);
            return;
        }
        if (cmp < 0) break;
    }
    obj->next = q;
    p->next   = obj;

} // add_object_in()

/**
 * Cre un objet obj dans chaque type de types.
 * INTERPRETATION PERSONNELLE : si un des types est une union (either ...),
 * alors il y a erreur car il ne peut pas avoir d'objet propre.
 * On pourrait changer a en crant l'objet dans tous les types de l'union.
 * Dtruit ou rutilise le maillon obj
 **/

void trt_object_def (name_list obj, name_list types)
{
    if (NULL == types)
    {
        add_object_in (obj, Domain.object);
        return;
    }

    for ( ; types; types = types->next)
    {
        type_list t = find_type ( types->item->entry);
        if (NULL == t)
        {
            sprintf (error, "type %s introuvable", types->item->entry);
            do_error (error);
        }

        if (t->is_union)
        {
            sprintf (error, "union %s ne peut avoir %s comme objet propre",
                     types->item->entry, obj->item->entry);
            do_error (error);
        }

        add_object_in (dico2name (obj->item, NULL), t);
        /*
        printf ("objet %s insr dans type %s\n", 
                obj->item->entry, t->name->entry);
        */
    }

    
    free (obj);

} // trt_objetc_def()


/**
 * tnl est une liste d'objets ou constantes,  insrer dans les types
 * correspondant
 *
 * Exemples :
 * (:objects pack1           - PACKAGE  
 *           paris marseille - CITY
 *           marignane       - (either CITY AIRPORT) )
 **/

void trt_objects_def (tname_list tnl)
{
    tname_list t; 
    for (; tnl ; tnl = t)
    {
        t = tnl->next;

        name_list nl, n;
        for (nl = tnl->names; nl; nl = n)
        {
            n = nl->next;
            trt_object_def (nl, tnl->type);
            // free (nl);  Le maillon est rcupr pour l'objet
        }

        free_name_list (tnl->type); // juste les maillons sans les noms
        free (tnl);
    }
        
} // trt_ojects_def()


void display_predicate (pred_list p)
{
    tvar_list      tvl;
    var_list       vl;
    ref_type_list  rtl;

    printf ("+ predicat : %s\n", p->name->entry);
    printf ("     arit : %d\n", p->arity);

    // Ecriture des types des arguments
    for (tvl = p->arguments; tvl; tvl = tvl->next)
    {
        // Pour chaque argument d'une liste ayant le meme type
        for (vl = tvl->vars; vl; vl = vl->next)
        {
            // affiche l'arguement
            printf ("     - %8s : ", vl->name->entry);
            // affiche le type (ventuellement compos)
            for (rtl = tvl->type; rtl; rtl = rtl->next)
            {
                printf ("%s ",rtl->type->name->entry);
            }
            printf ("\n");
        }
    }
    printf ("\n");

} // display_predicate()

void display_predicates (pred_list predicates)
{
    printf("Liste des prdicats :\n");
    for ( ; predicates; predicates = predicates->next)
        display_predicate (predicates);

    printf ("Fin de la liste des prdicats\n"); fflush(stdout);

} // display_predicates()

void display_type (type_list type)
{
    ref_type_list rtf;
    name_list     nl;

    printf ("+ type : %s\n", type->name->entry);
    if (type->is_union)
    {
        printf ("  union de :");
        for (rtf = type->subtypes; rtf; rtf = rtf->next)
            printf (" %s", rtf->type->name->entry);
        printf ("\n");

        if (type->objects)
            printf ("  --> PROBLEME D'OBJETS\n");
    }
    else
    {
        printf ("  objects : ");
        for (nl = type->objects; nl; nl = nl->next)
            printf (" %s", nl->item->entry);
        printf ("\n");
        printf ("  sous-types : ");
        for (rtf = type->subtypes; rtf; rtf = rtf->next)
            printf (" %s", rtf->type->name->entry);
        printf ("\n");
    }
    printf ("\n");

} // display_type()


void display_domain ()
{
    #if DEBUG_PDDL_PARSER
        type_list tl;
        printf ("Liste des types :\n");
        for (tl = Domain.types; tl; tl = tl->next)
        {
            display_type (tl);
        }
        printf ("Fin des types\n\n");

        display_predicates (Domain.predicates);
    #endif

} // display_domain()

/* transforme une liste de littraux en liste de faits */
void lit_list2list_fact_t (lit_list litx, list_fact_t * lf, int but,
                           int write,     char * prefixe, char * suffixe,
                           int noadd)
{
    lit_list l;

    for (l = litx; l; l = l->next)
    {
        atomic_formula2string (l, tmp_chlit);

        if (write)
        {
            printf ("%s", prefixe);
            if (l->signe == NEGATIF)
                printf ("(not %s)", tmp_chlit);
            else
                printf ("%s", tmp_chlit);
            printf ("%s", suffixe);
        }

        add_fact_in_list (tmp_chlit, l->signe, but, lf, noadd);
    }

} // lit_list2list_fact_t()

void display_init (lit_list init)
{

    #if PDDL_PARSER_SHOW_INIT 
        printf ("  - Situation Initiale :");
    #endif

    lit_list2list_fact_t (init, &lf_InitState, 0, PDDL_PARSER_SHOW_INIT,
                          "\t\t", "\n", 0);

    #if PDDL_PARSER_SHOW_INIT
        printf ("\n");
    #endif

} // display_init()


void display_goal (lit_list goal)
{
    #if PDDL_PARSER_SHOW_GOALS
        printf ("  - Goals :");
    #endif

    lit_list2list_fact_t (goal, &lf_Goals, 1, PDDL_PARSER_SHOW_GOALS,
                          "\t\t", "\n", 0);

    #if PDDL_PARSER_SHOW_GOALS
        printf ("\n");
    #endif

} // display_goal()

void display_problem ()
{
    display_init  (Problem.init);
    display_goal  (Problem.goal);

} // display_problem()

/********************* 11/06 ***************************************/

pred_list find_predicate (const char * s)
{
    pred_list p;

    for (p = Domain.predicates; p; p = p->next)
        if (! strcasecmp (s, p->name->entry)) return p;

    sprintf (error, "predicat %s introuvable", s);
    do_error (error);

    return NULL; // inutile

} // find_predicate()


/*****************************************************************************/

func_list find_function (const char * s)
{
    func_list f;

    for (f = Domain.functions; f; f = f->next)
        if (! strcasecmp (s, f->name->entry)) return f;

    sprintf (error, "function %s not found", s);
    do_error(error);

    return NULL;

} // find_function()


void display_function_values (func_list func)
{
    numeric_term_value ntv;
    term_list          term;

    printf ("function %s :\n", func->name->entry);

    for (ntv = func->values; ntv; ntv = ntv->next)
    {
        printf ("  +");
        for (term = ntv->terms; term; term = term->next)
        {
            printf ("  %s", ((name_list)term->item)->item->entry);
        }
        printf("\n      -> %d\n", ntv->value);
    }
}

void display_functions_values ()
{
    func_list f;

    printf ("\nstart of functions values :\n");

    for (f = Domain.functions; f; f = f->next)
        display_function_values (f);

    printf ("\nend of functions values :\n");
}

/*****************************************************************************/
// vrifie que le type t1 est un sous-type de t2
int is_subtype_of (ref_type_list t1, ref_type_list t2)
{
    return 1;
/*
    ref_type_list t;

    for ( ; t1; t1 = t1->next)
    {
        for (t = t2; t; t = t->next)
        {
*/          

} // is_subtype_of()


/*****************************************************************************/
void instanciate_action (act_list action, var_list var, tvar_list tvl);


/*****************************************************************************/
// Renvoie 0.
// Ne sert plus qu' l'criture d'un littral, quel qu'il soit
int ecrire_litteral (lit_list ll, char * prefixe)
{
    term_list tl;

    sprintf (tmp_chlit, "(%s", ll->predicate->name->entry);
    for (tl = ll->terms; tl; tl = tl->next)
        sprintf (tmp_chlit, "%s %s", tmp_chlit, (tl->type == TERM_OBJ ? (term_object(tl))->item->entry : (term_variable(tl))->name->entry ));
    sprintf (tmp_chlit, "%s)", tmp_chlit);

    //#if (DEBUG_PDDL_PARSER || DEBUG_ACTION_LEVEL)
        if (ll->signe == NEGATIF)
            printf ("%s(not %s)", prefixe, tmp_chlit);
        else
            printf ("%s%s", prefixe, tmp_chlit);
    //#endif

    //#if  defined    INCORPORATE_FDP
    //    return SearchFact (tmp_chlit, 1);
    //#else
        return 0;
    //#endif      /* INCORPORATE_FDP */

} // ecrire_litteral()

/*****************************************************************************/
int is_new_litteral (lit_list l, lit_list litx)
{
    lit_list  lp;
    term_list t, tp;

    for (lp = litx; lp; lp = lp->next)
    {
        if (l->predicate != lp->predicate) continue;

        for (t = l->terms, tp = lp->terms; t; t = t->next, tp = tp->next)
            if (term_entry (t) != term_entry (tp))  break;

        if (tp == NULL) break;  // on a trouv le mme littral
    }

    return (lp == NULL);

} // is_new_litteral()


/*****************************************************************************/
void add_new_litteral (lit_list l, lit_list * lst)
{
    lit_list nl;

    if (is_new_litteral (l, *lst))
    {
        nl = dup_litteral (l);
        nl->next = *lst;
        *lst = nl;
    }

} // add_new_litteral()

int matching_terms (term_list tl1, term_list tl2)
{
    while (tl1 != NULL && tl2 != NULL)
    {
        if (term_dico(tl1) != term_dico(tl2))
            return 0;
        tl1 = tl1->next;
        tl2 = tl2->next;
    }
    return (tl1 == NULL && tl2 == NULL);
}

int get_function_value (func_list func, term_list terms)
{
    numeric_term_value values;

    for (values = func->values; values; values = values->next)
        if (matching_terms (values->terms, terms))
            return (values->value);

    snprintf (error, 200,
            "no corresponding value");
    do_error (error);
    return (0); // for the warning...
}

int get_action_cost_value (numeric_term cost)
{
    if (NULL == cost) return 0;

    if (NUMERIC_TERM_VALUE == cost->type)
        return (cost->value);

    return get_function_value (cost->functerm->function, cost->functerm->terms);

} // get_action_cost_value()


/*****************************************************************************/
/* new_litx : (modifi) liste des nouveaux littraux pour dtecter           */
/*            stabilisation.                                                 */
/* Nom de l'action : dpend de IPCOutput                                   */
/* Si IPCOutput, c'est : (nom param1 param2 ... paramk)                    */
/* sinon           c'est : nom(param1, param2, ..., paramk)                  */
/*****************************************************************************/
void ecrire_action (lit_list litx, act_list action)
{
    tvar_list tvl;
    var_list vl;
    lit_list ll;
    int num_lit;


    #if  defined    INCORPORATE_FDP
        pre = add = del = raz;
    #endif

    if (IPCOutput)
        sprintf (tmp_chact, "(%s", action->name->entry);
    else
        sprintf (tmp_chact, "%s(", action->name->entry);
    
    for (tvl = action->params; tvl ; )
    {
        for (vl = tvl->vars;  ; ) 
        {
            if (IPCOutput)
                sprintf (tmp_chact, "%s %s", tmp_chact, vl->instance->entry);
            else
                sprintf (tmp_chact, "%s%s", tmp_chact, vl->instance->entry);

            vl = vl->next;
            if (NULL == vl) break;
            if (! IPCOutput)
                sprintf (tmp_chact, "%s, ", tmp_chact);
        }
        tvl=tvl->next;
        if (NULL == tvl) break;
        if (! IPCOutput)
            sprintf (tmp_chact, "%s, ", tmp_chact);
    }

    sprintf (tmp_chact, "%s)", tmp_chact);

    //printf("action %s\n", tmp_chact);

    #if (DEBUG_PDDL_PARSER || DEBUG_ACTION_LEVEL)
        printf ("\n+++ ACTION : %s +++\n", tmp_chact);

        printf ("\tPRECONDITION :\n");
    #endif
   
    /*
    lit_list2list_fact_t (action->pre, &pre, 0, 
                          DEBUG_PDDL_PARSER || DEBUG_ACTION_LEVEL,
                          "\t\t", "\n", 1);
    */

        /* Traitement de la prcondition */
    for (ll = action->pre; ll; ll = ll->next)
    {
        atomic_formula2string (ll, tmp_chlit);

        if (NONE == (num_lit = SearchFact (tmp_chlit, ll->signe)))
        {
            /* cette prcondition n'a pas t trouve */
            /* ce doit tre parce qu'elle est ngative, */
            /* sinon, il y a un problme */
            #if (DEBUG_PDDL_PARSER || DEBUG_ACTION_LEVEL)
                printf("Warning : (");
                if (ll->signe == NEGATIF)
                    printf("not ");
                printf("%s) : precondition inexistante\n", tmp_chlit); 
            #endif
            if (ll->signe != NEGATIF)
            {
                printf("Probleme action %s : \n", tmp_chact);
                printf("  -> precondition inexistante : %s\n", tmp_chlit);
                exit(1);
            }

            num_lit = AddFact (tmp_chlit, ll->signe, 1 /* c'est un but */);

        }

        #if (DEBUG_PDDL_PARSER || DEBUG_ACTION_LEVEL)
            printf ("\t\t");
            if (ll->signe == NEGATIF)
                printf ("(not %s)", tmp_chlit);
            else
                printf ("%s", tmp_chlit);
            printf ("\n");
        #endif

        add_numfact_in_list (num_lit, POSITIF, &pre);
    }

    #if (DEBUG_PDDL_PARSER || DEBUG_ACTION_LEVEL)
        printf ("\n\tDELETE-LIST :\n");
    #endif

    for (ll = action->del; ll; ll = ll->next)
    {
        atomic_formula2string (ll, tmp_chlit);

        if (NONE == (num_lit = SearchFact (tmp_chlit, POSITIF)))
        {
            #if (DEBUG_PDDL_PARSER || DEBUG_ACTION_LEVEL)
                printf ("\t\t(inexistant)\t%s\n", tmp_chlit);
                //printf ("\n====> ACTION IGNOREE !\n\n");
            #endif
            ++nActionsUnknownDel;
            num_lit = AddFact (tmp_chlit, POSITIF, 1);
            add_numfact_in_list (num_lit, POSITIF, &del);
            /* restitution des listes */
            #if defined    INCORPORATE_FDP
                //free_list_fact (&pre);
                //free_list_fact (&del);
            #endif
            //return;
        }
        else
        {
            #if (DEBUG_PDDL_PARSER || DEBUG_ACTION_LEVEL)
                printf ("\t\t%s\n", tmp_chlit);
            #endif

            #if defined    INCORPORATE_FDP    
                add_numfact_in_list (num_lit, POSITIF, &del);
            #endif
        }
    }

    #if (DEBUG_PDDL_PARSER || DEBUG_ACTION_LEVEL)
        printf ("\n\tADD-LIST :\n");
    #endif

    for (ll = action->add; ll; ll = ll->next)
    {
        atomic_formula2string (ll, tmp_chlit);

        if (NONE == (num_lit = SearchFact (tmp_chlit, POSITIF)))
        {
            num_lit = AddFact (tmp_chlit, POSITIF, 0);
            #if (DEBUG_PDDL_PARSER || DEBUG_ACTION_LEVEL)
                printf ("\t\t(new)\t%s\n", tmp_chlit);
            #endif
            add_new_litteral (ll, & new_litx);
            ++nblitx;
        }
        else
        {
            if (LesFaits[num_lit].but == 1)
            {
                #if (DEBUG_PDDL_PARSER || DEBUG_ACTION_LEVEL)
                    printf ("\t\t(new)\t%s\n", tmp_chlit);
                #endif
                add_new_litteral (ll, & new_litx);
                ++nblitx;
            }
            else
            {
                #if (DEBUG_PDDL_PARSER || DEBUG_ACTION_LEVEL)
                    printf ("\t\t%s\n", tmp_chlit);
                #endif
            }
            LesFaits[num_lit].but = 0;
            LesFaits[OPPOSE(num_lit)].but = 0;
        }

        #if defined    INCORPORATE_FDP    
            add_numfact_in_list (num_lit, POSITIF, &add);
        #endif
    }

    #if (DEBUG_PDDL_PARSER || DEBUG_ACTION_LEVEL)
        printf ("\n");
    #endif


    #if defined    INCORPORATE_FDP
        AddNewAction(tmp_chact, pre.tete, add.tete, del.tete, 
                     get_action_cost_value(action->cost), 
                     action->contradicts);
    #endif

    ++nbact;

} // ecrire_action()

/*****************************************************************************/
/* Passe  la variable qui suit v dans la liste tvl        */
/* nv et ntvl contiendront les nouvelles variable et liste */
/* Note : reste ok si *nv = nv et *tvl = tvl               */
void skip_param (var_list   v, tvar_list    tvl,
                 var_list *nv, tvar_list * ntvl)
{
    *nv   = v->next;
    *ntvl = tvl;
    if (NULL == *nv)
    {
        /* il faut passer au maillon aprs tvl */
        *ntvl = tvl->next;
        if (NULL == *ntvl) /* il n'y a plus de variable */
            return;
        *nv = (*ntvl)->vars;
    }
} // skip_param() 


/*****************************************************************************/
/* tente d'unifier une prcondition et un littral (atomique)                */
/* Unifient si meme signe, meme predicat, chaque terme s'unifie              */
/* l est un littraux existant                                               */
/* pre est une prcondition d'une action, + ou - instancie.                 */
/* Modif du 15/03/05 pour corriger bug de variable instancie par un terme   */
/* qui ne correspond pas  son type.                                         */
/*****************************************************************************/
int are_unifying (lit_list l, lit_list pre)
{
    term_list tl, tp;
    dico_list val;      /* valeur du terme tp   */
    var_list  var;      /* variable du terme tp */

    if (! (   l->signe     == pre->signe
           && l->predicate == pre->predicate ) )
    {
        #if DEBUG_ACTION_LEVEL
            printf ("signe diff %d et %d ou predicat diff\n", 
                    l->signe, pre->signe);
            fflush (stdout);
        #endif
        return 0;
    }

    for (tl = l->terms, tp = pre->terms; tl; tl = tl->next, tp = tp->next)
    {
        val = term_dico(tp);
        if (NULL == val) /* tp dsigne une variable non instancie */
        {
            var = term_variable(tp);
            /* SAUVER LA VALEUR DE VARIABLE tp */
            push_adr_inst (stack, var);
            /* Il faudra remettre item  NULL */
            var->instance = term_dico(tl);
            if ( ! check_and_propagate_unification (var) ) return 0;
        }
        else /* tp n'est pas une variable ou est instancie */
        {
            if (val != term_dico (tl)) return 0;
        }
    }

    return 1;

} // are_unifying()


void instanciate_not_unifyied_params (lit_list  litx,
                                      act_list  action, 
                                      var_list  var, 
                                      tvar_list tvl);

/*
 * var est un parametre de action qui n'avait pas t unifi et qu'on
 * instancie  toutes les objets de type (et de ses sous-types).
 * Il faudra ensuite s'occuper des autres parametres de action
 */ 
void instanciate_not_unifyied_param (lit_list      litx,
                                     act_list      action,
                                     var_list      var,
                                     ref_type_list type,
                                     tvar_list     tvl)
{
    name_list     o;
    ref_type_list rtl;
    var_list  next_var;
    tvar_list next_tvl;
    //printf ("instanciation\n");

    for ( ; type; type = type->next)
    { 
        //printf("Pour tous les types\n");
        for (o = type->type->objects; o; o = o->next)
        {
            //printf ("Pour tous les objets du type\n");
            // ce n'est pas la peine de sauver la valeur de var
            var->instance = o->item;
            push_NULL (stack);
            if ( ! check_and_propagate_unification (var) )
            {
                rollback (stack);
                continue;
            }

            // var tant instanci, on avance de parametre
            skip_param (var, tvl, &next_var, &next_tvl);

            //printf ("repart dans instanciation action\n");
            instanciate_not_unifyied_params (litx, action, next_var, next_tvl);
            rollback (stack);
        }
        for (rtl = type->type->subtypes; rtl; rtl = rtl->next)
            instanciate_not_unifyied_param (litx, action, var, rtl, tvl);
    }
} // instanciate_not_unifyied_param()


/*****************************************************************************/
/* Instancie les paramtres  partir de var qui n'auraient pas */
/* t dj unifis */
void instanciate_not_unifyied_params (lit_list  litx,
                                      act_list  action, 
                                      var_list  var, 
                                      tvar_list tvl)
{
    while (1)   // recherche d'un parametre non unifi
    {
        if (NULL == var)
        {
            // plus de parametre  instancier
            ecrire_action(litx, action);
            return;
        }
        if (NULL == var->instance)
        {
            // parametre non unifi
            instanciate_not_unifyied_param (litx, action, var, tvl->type, tvl);
            // restaure la valeur de var
            var->instance = NULL;
            return;
        }
        skip_param (var, tvl, &var, &tvl);
    }

} // instanciate_not_unifyied_params()

/*****************************************************************************/
/* 
 * unify_and_instanciate_action() : 
 * cherche  unifier 'action' de toutes les faons
 * possibles pour qu'elle s'applique dans l'ensemble de faits 'litx'
 * pre est la premire prcondition de action non encore satisfaite
 */
void unify_and_instanciate_action (lit_list litx, 
                                   act_list action, 
                                   lit_list pre)
{
    lit_list l;

    #if DEBUG_ACTION_LEVEL
        printf ("Unify and instanciate action :\n"); fflush(stdout);
    #endif

    if (NULL == pre)
    {
        #if DEBUG_ACTION_LEVEL
            printf ("reste a instancier les parametres\n"); 
            fflush(stdout);
        #endif
        /* toutes les prconditions (positives) sont satisfaites      */
        /* il reste peut tre des paramtres  instancier             */
        instanciate_not_unifyied_params (litx,
                                         action,
                                         action->params == NULL ? 
                                            NULL : 
                                            action->params->vars, 
                                         action->params);
    }
    else if (NEGATIF == pre->signe)  /* MODIF du 30/03/2005 */
    {
        #if DEBUG_ACTION_LEVEL
            printf ("precondition negative : "); fflush(stdout);
            ecrire_litteral (pre, "");
            printf ("\n"); fflush(stdout);
        #endif
        unify_and_instanciate_action (litx, action, pre->next);
    }
    else
    {
        #if DEBUG_ACTION_LEVEL
            printf ("precondition positive "); fflush(stdout);
            ecrire_litteral (pre, "");
            printf ("\n"); fflush(stdout);
        #endif
        for (l = litx; l; l = l->next)
        {
            push_NULL (stack);
            if (are_unifying (l, pre))
            {
                #if DEBUG_ACTION_LEVEL
                    printf ("unifie avec : "); fflush(stdout);
                    ecrire_litteral (l, "");
                    printf ("\n"); fflush(stdout);
                #endif
                unify_and_instanciate_action (litx, action, pre->next);
            }
            else
            {
                #if DEBUG_ACTION_LEVEL
                    printf ("non unifie avec : "); 
                    fflush(stdout);
                    ecrire_litteral (l, "");
                    printf ("\n"); fflush(stdout);
                #endif
            }
            rollback (stack);
        }
        #if DEBUG_ACTION_LEVEL
            printf ("fin unification precondition positive\n"); fflush(stdout);
        #endif
    }
    #if DEBUG_ACTION_LEVEL
        printf ("fin fonction unify and instanciate action\n"); fflush(stdout);
    #endif

} // unify_action()

void append_lit_list (lit_list la, lit_list * L)
{
    lit_list l;

    if (NULL == la) return;

    for (l = la; l->next != NULL; l = l->next);
    l->next = *L;
    *L = la;
}

/*****************************************************************************/
/* Creation d'un niveau du graphe d'actions  partir d'un ensemble */
/* de littraux */
//void create_action_level (lit_list litx, act_list actions)
int create_action_level ()
{
    act_list act;
    nActionsIgnorees = 0;
    #if STATS_LEVELS
        //int      nb_litx = 0;
        //lit_list l;
    #endif

    #if DEBUG_ACTION_LEVEL
        printf ("\n");
        printf ("=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=\n");
        printf ("=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=\n");
        printf ("=-=-=-    Cration d'un niveau d'actions     -=-=-=\n");
        printf ("=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=\n");
        printf ("=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=\n");
        printf ("\n");
    #endif

    new_litx = NULL;
    nbact    = 0;
    ++levels;

    for (act = Domain.actions ; act; act = act->next)
    {
        if (ACT_OK == act->status)
        {
            #if DEBUG_ACTION_LEVEL
                printf ("Tentative d'instanciation de l'action : %s\n",
                        act->name->entry);
            #endif

            unify_and_instanciate_action (current_litx, act, act->pre);
        }
        else
        {
            #if DEBUG_ACTION_LEVEL
                printf ("Action saute car impossible : %s\n",
                        act->name->entry);
            #endif
        }
    }

    append_lit_list (new_litx, & current_litx);

    #if DEBUG_ACTION_LEVEL
        printf ("Fin du niveau d'actions\n");
    #endif

    #if STATS_LEVELS
        //nb_litx = 0;
        //for (l = current_litx; l; l = l->next) ++nb_litx;

        printf ("\t%d S\t%d A\t%d F\n", levels, nbact, nblitx);
    #endif

    return (NULL != new_litx);

} // create_action_level ()


/*****************************************************************************/
void stabilize_graph ()
{
    do
    {
        #if DEBUG_ACTION_LEVEL
            printf ("."); fflush(stdout);
        #endif

        create_action_level  ();
    }
    while (NULL != new_litx);
    //printf("end of stab\n");fflush(stdout);
} // stabilize_graph()

/*****************************************************************************/
void summarize_problem ()
{
    printf ("Number of actions generated without filtering : %u\n\n", nbact);

} // summarize_problem ()


/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/

/* MODIFICATIONS DU 10/02/2005 */

/*****************************************************************************/
/* remplace get_corresponding_variable()
   d est une variable aparaissant comme terme.
   on la recherche parmi 'tvl' pour retourner un pointeur (void *) var_list 
   TODO :  il faut vrifier que le type de la variable est compatible avec
   celui de l'argument correspondant : type(var) inclus dans type(arg)
*/
void * get_variable (dico_list d, tvar_list tvl, ref_type_list rtl)
{
    var_list vl;
    for ( ; tvl; tvl = tvl->next)
    {
        for (vl = tvl->vars; vl; vl = vl->next)
            if (d == vl->name)
            {
                if (is_subtype_of (tvl->type, rtl))
                    return ((void *) vl);
                else
                {
                    sprintf (error, "type of %s incorrect with argument", 
                                     d->entry);
                    do_error (error);
                }
            }
    }
    sprintf (error, "free variable %s are not allowed", d->entry);
    do_error (error);

    return NULL;                

} // get_variable ()


/*****************************************************************************/
/* recherche l'objet 'dl' dans 'type' et ses sous-types */
 
name_list find_object (dico_list dl, type_list type)
{
    name_list o;

    for (o = type->objects; o ; o = o->next)
        /* on devrait simplement comparer les dico_list plutt que les chaines
        if (! strcasecmp (dl->entry, o->item->entry)) return o;
        */
        if (dl == o->item) return o;

    ref_type_list rtl;

    for (rtl = type->subtypes; rtl; rtl = rtl->next)
    {
        o = find_object (dl, rtl->type);
        if (o) return o;
    }

    return NULL;

} // find_object ()

int is_object_of_type (dico_list d, ref_type_list rtl)
{

    for ( ; rtl; rtl = rtl->next)
        if (NULL != find_object (d, rtl->type)) return VRAI;

    return FAUX;

} // is_object_of_type()

/*****************************************************************************/
/* remplace get_corresponding_object()
 * INTERPRETATION PERSONNELLE : si le type de l'objet est une union, alors 
 * on ne renvoie que le 1 objet trouv du type.
 * NOTE : dans certains domaines/problemes, les objets ne sont pas tous
 * dclars avant de figurer dans un littral.
 * On peut utiliser find_or_create_object() plutt que find_object().
 */ 
void * get_object (dico_list d, ref_type_list rtl)
{
    name_list o;

    for ( ; rtl; rtl = rtl->next)
        if (NULL != (o = find_object (d, rtl->type))) return ((void *) o);

    sprintf (error, "Could not find correponding object %s", d->entry);
    do_error (error);

    return NULL;

} // get_object()


/*****************************************************************************/
void add_constraint (constraint cstr, var_list v)
{
    v->constraints = new_const_list (cstr, v->constraints);

} // add_constraint()


/*****************************************************************************/
/* La variable v a t unifie. Il faut vrifier que ses contraintes ne sont */
/* pas violes, et propager cette unification.                               */
/* Modif du 17/03/05 : Il faut vrifier que toute unification est compatible */
/* avec le type de la variable */
int check_and_propagate_unification (var_list v)
{
    var_list   v1, v2;
    name_list  o2;
    const_list cl;
    constraint c;

    to_empty_stack (file);
    push_var (file, v);

    while (! empty_stack (file))
    {
        v1 = pop_var (file);
        /* Modif du 17/03/05. */
        /* Vrifie que l'instance de v1 correspond  son type */
        if ( ! is_object_of_type (v1->instance, v1->type) ) return FAUX;

        for (cl = v1->constraints; cl; cl = cl->next)
        {
            c = cl->cstr;
            switch (c->type)
            {
                case EQU_VAR :
                    v2 = (c->m1 == v1 ? c->m2 : c->m1);
                    if (v2->instance != NULL)
                    {
                        if (v1->instance != v2->instance) return FAUX;
                    }
                    else
                    {
                        push_adr_inst (stack, v2);
                        v2->instance = v1->instance;
                        push_var (file, v2);
                    }
                    break;

                case NEQ_VAR :
                    v2 = (c->m1 == v1 ? c->m2 : c->m1);
                    if (v2->instance != NULL)
                    {
                        if (v1->instance == v2->instance) return FAUX;
                    }
                    break;

                case NEQ_OBJ :
                    o2 = c->m2;
                    if (v1->instance == o2->item) return FAUX;
                    break;
            } // switch
        }
    }
    return VRAI;

} // propagate_instance()

/*****************************************************************************/
int make_inequal (term_list t1, term_list t2)
{
    var_list   v1, v2;
    name_list  o1, o2;
    constraint cstr;

    if (t1->type == TERM_OBJ)
    {
        o1 = (name_list) t1->item;
        if (t2->type == TERM_OBJ)            // (!= o1 o2)
        {
            o2 = (name_list) t2->item;
            return (o1->item != o2->item);
        }
        else                                 // (!= o1 ?v2)
        {
            v2 = (var_list) t2->item;
            if (v2->instance != NULL) return (v2->instance != o1->item);
            cstr = new_constraint (NEQ_OBJ, v2, (void *) o1);
            add_constraint (cstr, v2);
            return VRAI;
        }
    }
    else
    {
        v1 = (var_list) t1->item;
        if (t2->type == TERM_OBJ)           // (!= ?v1 o2)
        {
            o2 = (name_list) t2->item;
            if (v1->instance != NULL) return (v1->instance != o2->item);
            cstr = new_constraint (NEQ_OBJ, v1, (void *) o2);
            add_constraint (cstr, v1);
            return VRAI;
        }
        else                                // (!= ?v1 ?v2)
        {
            v2 = (var_list) t2->item;
            if (v1->instance != NULL && v2->instance != NULL)
                return (v1->instance != v2->instance);
            cstr = new_constraint (NEQ_VAR, v1, (void *) v2);
            add_constraint (cstr, v1);
            add_constraint (cstr, v2);
            return VRAI;
        }
    }
} // make_inequal()


/*****************************************************************************/
int make_equal (term_list t1, term_list t2)
{
    var_list   v1, v2;
    name_list  o1, o2;
    constraint cstr;

    if (t1->type == TERM_OBJ)
    {
        o1 = (name_list) t1->item;
        if (t2->type == TERM_OBJ)            // (= o1 o2)
        {
            o2 = (name_list) t2->item;
            return (o1->item == o2->item);
        }
        else                                 // (= o1 ?v2)
        {
            v2 = (var_list) t2->item;
            if (v2->instance != NULL) return (v2->instance == o1->item);
            v2->instance = o1->item;
            return check_and_propagate_unification (v2);
        }
    }
    else
    {
        v1 = (var_list) t1->item;
        if (t2->type == TERM_OBJ)           // (= ?v1 o2)
        {
            o2 = (name_list) t2->item;
            if (v1->instance != NULL) return (v1->instance == o2->item);
            v1->instance = o2->item;
            return check_and_propagate_unification (v1);
        }
        else                                // (= ?v1 ?v2)
        {
            v2 = (var_list) t2->item;
            /* 
                la vrification est succinte. Si l'un des deux est instanci,
                l'autre prend sa valeur si possible, sinon chec.
                On ne vrifie pas les inconsistances du style :
                        (= ?v1 ?v2) (= ?v1 ?v3) (!= ?v2 ?v3)
                    ni mme : (= ?v1 ?v2) (!= ?v1 ?v2)
                elles seront dtectes plus tard.
                Il faut dire qu'en principe, ces inconsistances n'arrivent pas
                si le domaine est "srieux".
            */
            if (v1->instance != NULL)
            {
                if (v2->instance != NULL)
                    return (v1->instance == v2->instance);
                else
                {
                    v2->instance = v1->instance;
                    return (check_and_propagate_unification (v2));
                }
            }
            else
            {
                if (v2->instance != NULL)
                {
                    v1->instance = v2->instance;
                    return (check_and_propagate_unification (v1));
                }
                else
                {
                    cstr = new_constraint (EQU_VAR, v1, (void *) v2);
                    add_constraint (cstr, v1);
                    add_constraint (cstr, v2);
                    return VRAI;
                }
            }
        }
    }
} // make_equal()

/*****************************************************************************/
/* Traite la liste des prconditions en enlevant les prdicats               */
/* d'(in)galit.                                                            */
/* pre est modifi en sortie et ne contient que la liste pure des          */
/* prconditions.                                                            */
/* Renvoie ACT_IMPOSSIBLE si une contrainte est impossible.                  */
/* ATTENTION : si ACT_IMPOSSIBLE, pre n'est alors plus une liste valide.     */
/*****************************************************************************/
int trt_act_pre (tvar_list params, lit_list * pre)
{
    lit_list npre, qpre, p, q;
    term_list t1, t2;

    npre = qpre = NULL; 

    for (p = *pre; p; p = q)
    {
        q = p->next;
        if (p->predicate == Domain.equals)
        {
            t1 = p->terms;
            t2 = t1->next;
            if (p->signe == NEGATIF)
            {
                if ( ! make_inequal (t1, t2) ) return ACT_IMPOSSIBLE;
            }
            else if ( ! make_equal (t1, t2) ) return ACT_IMPOSSIBLE;
            free (t2);
            free (t1);
            free (p);
        }
        else /* p is neither 'equals' nor 'diff' */
        {
            if (npre)
                qpre = qpre->next = p;
            else
                npre = qpre = p;
            qpre->next = NULL;
        }
    }

    *pre = npre;
    return ACT_OK;

} // trt_act_pre()

int same_terms (term_list t1, term_list t2)
{
    if (t1->type == TERM_OBJ)
        if (t2->type == TERM_OBJ)
            return (term_object(t1) == term_object(t2));
        else
            return 0;
    else /* t1 is a variable */
        if (t2->type == TERM_VAR)
            return (term_variable(t1) == term_variable(t2));
        else
            return 0;

} // same_terms()

/******************************************************************************/
/* same_literals()                                                            */
/*      returns true if literlas l1 and l2 are the same                       */
/******************************************************************************/
int same_literals (lit_list l1, lit_list l2)
{
    term_list tl1, tl2;

    if (l1->predicate != l2->predicate)
        return 0;

    /* terms comparison */
    tl1 = l1->terms;
    tl2 = l2->terms;
    while (tl1 != NULL && tl2 != NULL)
    {
        if (! same_terms(tl1, tl2))
            break;

        tl1 = tl1->next;
        tl2 = tl2->next; 
    }
    return (tl1 == NULL && tl2 == NULL);

} // same_literals()

void println_literal (lit_list lit)
{
    term_list tl;

    printf ("(");
    printf ("%s", lit->predicate->name->entry);

    for (tl = lit->terms; tl; tl = tl->next)
    {
        if (tl->type == TERM_OBJ)
            printf(" %s", term_object(tl)->item->entry);
        else
            printf(" %s", term_variable(tl)->name->entry);
    }
    printf (")\n");
}

/******************************************************************************/
/* has_contradictory_effects()                                                */
/*      returns true if action has some contradictory effects                 */
/******************************************************************************/
int has_contradictory_effects (dico_list actname, lit_list adds, lit_list dels)
{
    lit_list add, del;

    for (add = adds; add; add = add->next)
    {
        for (del = dels; del; del = del->next)
        {
            if (same_literals (add, del))
            {
                /*
                printf("contradictory effects for action %s :\n", 
                            actname->entry);
                println_literal (add);
                println_literal (del);
                */
                return 1;
            }
        }
    }
    return 0;
}

/*****************************************************************************/
/*****************************************************************************/
act_list trt_act_def (dico_list name, tvar_list params, lit_list pre,
                      lit_list eff, numeric_term cost)
{
    int status = trt_act_pre (params, & pre);
    to_empty_stack (stack);  // les modifications des variables sont immuables

    // il faudra rechercher si l'action existe dj
    lit_list add, crt_add;
    lit_list del, crt_del;

    add = del = crt_add = crt_del = NULL;

    for ( ; eff; eff = eff->next)
    {
        if (eff->signe == NEGATIF)
        {
            if (NULL == del)
                crt_del = del = eff;
            else
                crt_del = crt_del->next = eff;
        }
        else
        {
            if (NULL == add)
                crt_add = add = eff;
            else
                crt_add = crt_add->next = eff;
        }
        eff->signe = POSITIF;
    }


    if (crt_add) crt_add->next = NULL;
    if (crt_del) crt_del->next = NULL;

    //printf ("traitement des effets effectue\n"); fflush (stdout);
    int nbp = 0;
    tvar_list tvl;
    var_list  vl;

    for (tvl = params; tvl; tvl = tvl->next)
        for (vl = tvl->vars; vl; vl = vl->next) ++nbp;

/*
    printf ("traitement des parametres effectue\n"); fflush (stdout);
    printf ("==> action %s cre avec %d paramtres\n", name->entry, nbp);
    for (tvl = params; tvl; tvl = tvl->next)
        for (vl = tvl->vars; vl; vl = vl->next)
        {
            printf ("......param %s (%s)\n", vl->name->entry, 
                    (tvl->type->type->is_union ? 
                                     "anonyme" : 
                    tvl->type->type->name->entry) );
        }
*/
    return new_act_list (name, status, params, pre, add, del, cost, 
                         has_contradictory_effects(name, add, del), NULL);

} // trt_act_def()



/*****************************************************************************/
/*
    Fonction utilise pour situation init.
    Si object n'existe pas dans le type correspondant, il faut le crer.
    INTERPRETATION PERSONNELLE : Si le type est compos (either ...), on cre 
    l'objet dans tous les types de l'union.
    INTERPRETATION PERSONNELLE : Il y a une erreur si un des types de l'union 
    est lui mme une union (voir aussi trt_object_defs()).
 */
name_list find_or_create_object (dico_list dl, ref_type_list rtl)
{
    name_list o;
    ref_type_list tmp_rtl;

    for (tmp_rtl = rtl; tmp_rtl; tmp_rtl = tmp_rtl->next)
        if (NULL != (o = find_object (dl, tmp_rtl->type))) return o;

    // il faut crer l'objet
    //printf ("objet %s non trouv\n", dl->entry);
    if (NULL == rtl)
    {
        o = dico2name (dl, NULL);
        add_object_in (o, Domain.object);
        return o;
    }

    for (tmp_rtl = rtl; tmp_rtl; tmp_rtl = tmp_rtl->next)
    {
        if (tmp_rtl->type->is_union)
        {
            sprintf (error, "union %s ne peut avoir %s comme objet propre",
                     tmp_rtl->type->name->entry, dl->entry);
            do_error (error);
        }
        o = dico2name (dl, NULL);
        add_object_in (o, tmp_rtl->type); 
    }
    return o;

} // find_or_create_object()


/*****************************************************************************/

ref_type_list names2reftypes (name_list names)
{
    ref_type_list head_rtf;
    ref_type_list current_rtf;

    head_rtf = current_rtf = NULL;

    if (NULL == names)
        return new_ref_type_list (Domain.object, NULL);

    for ( ; names; names = names->next)
    {
        type_list tl = find_type (names->item->entry);

        if (NULL == tl)
        {
            sprintf (error, "type %s referred to but undefined", 
                     names->item->entry);
            do_error (error);
        }

        if (NULL == current_rtf)
            head_rtf = current_rtf = new_ref_type_list (tl, NULL);
        else
        {
            current_rtf->next = new_ref_type_list (tl, NULL);
            current_rtf = current_rtf->next;
        }
    }

    return head_rtf;

} // names2reftypes()


