#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <time.h>
#include <math.h>

#include "pddl.h"

#include "cfdp.h"
#include "constantes.h"
#include "consistance.h"
#include "fdp.h"
#include "hash.h"
#include "hasht.h"
#include "main.h"
#include "search.h"
#include "sequences.h"
#include "sets.h"
#include "split.h"
#include "structures.h"
#include "stats.h"
#include "suppressions.h"
#include "threads.h"

#include "hash.h"
#include "minsteps.h"


int * minCosts = NULL;  /* couts min indices sur le nombre d'actions 
			   en partant de la fin. */

int curMinCost; /* cout minimal d'un plan valide. */
int curMinLen;  /* longueur du plan correspondant. */

int startWithFDP = 1; /* on commence avec FDP standard pour calculer 
			 partir avec une longueur min, cout min, et 
			 memoriser des etats invalides. */

int useHTinFS = 1; /* les etats qui ne permettent pas d'atteindre 
		      les buts provoquent des dead ends qui peuvent
		      perturber la recherche de terminaison occv 
		      (mais ca va plus vite que d'utiliser ces etats 
		      dans Hccv avec des couts min). */

HASHT Hccv; /* table des etats/distance/cout. */
HASHT Ho;   /* others */


int *bestCFDP, bestCFDPlen;


void recopieLaHashTable (hash_table_t H)
{
  int i;
  hash_list p;
  unsigned int *state;

  for (i = 0; i < HASH_SIZE; i ++)
    for (p = H[i].liste; p != NULL; p = p->next)
      {
	state = duplicate(p->etat, Hccv->size_states);
	addStateCostHashTable (Hccv, i, p->distance, minCosts[p->distance], 
			       NULL, NULL, state);
      }
}


void updateBestPlan (int ns, int thread)
{
  int i;
  S_GR * G = &(GRT[thread]);

  for (i = 0; i < ns; i ++)
    bestCFDP [i] = G->tAi [i][0];
  bestCFDPlen = ns;
}



void * CFDP (void * args)
{
  int thread = *((int *) args), ns;
  THREAD T = theThreads+thread;

  SizeGraph = T->nsmax;


  TOP_DEPART_THREAD(T);
  CTOP_DEPART;

  FDP_Allocations_Initializations(T->nsmax, thread);
  Hccv = newHashTable (HASH_SIZE, nFaits, RSV_INIT_SIZE);
  Ho = newHashTable (HASH_SIZE, nFaits, RSV_INIT_SIZE);
  bestCFDP = malloc(T->nsmax * sizeof(int));
  assert(bestCFDP != NULL);

  curMinCost = T->nsmax * maxActCost + 1;
  curMinLen = bestCFDPlen = NONE;

  if (startWithFDP)
    {
      FDP4CFDP (thread);
      if ( ! T->fdp_solvable)
	goto FINISH;
      curMinCost = GRT[thread].cost;
      curMinLen = T->ns;
      updateBestPlan (curMinLen, thread);
      if (1) printf("\nFDP  %d %d$\n", T->ns, GRT[thread].cost);
    }

  if (minActCost == maxActCost)
    goto FINISH;

  evaluateBackwardMinCosts (sizeMinCosts, thread);

  if (startWithFDP)
    recopieLaHashTable (HT);

  ns = costIDForwardSearch (thread);

 FINISH:

  freeHashTable (Hccv);
  freeHashTable (Ho);

/*   RestitueEnumerate(T->nsmax, thread); */
/*   RestitueFilesConsistance(thread); */
/*   RestituePilesSuppressions(thread); */

  TOP_ARRIVEE_THREAD(T);
  totalTime += T->Ctemps;
  totalChecks += T->nChecks;
  totalNE += T->NE;
  CTOP_ARRIVEE;
  StopSearch = 1;
  fdp_solvable = (curMinLen != NONE);

  return NULL;
}



/* ************************************************************** */
/*                               FDP                              */
/* ************************************************************** */

void FDP4CFDP (int thread)
{
  int ns;
  THREAD T = theThreads+thread;


  for (ns = 1; ns <= T->nsmax; ns ++)
    {
      T->ns = ns;
      T->fdp_solvable = 0;
      InitGrapheDActionsThreads(ns, InitState, Goals, thread);
      InitializeSousButs(Goals, ns, thread);

      if (MakeConsistant(ns, thread))
	Enumerate(ns, 0, 0, thread);

      if (T->fdp_solvable)
	break; 
    }

}



/* ************************************************************** */
/*                  RECHERCHE EN AVANT D'UN PLAN                  */
/* ************************************************************** */


int costIDForwardSearch (int thread)
{
  THREAD T = theThreads+thread;
  int ns;

/*   if (curMinLen != NONE) */
/*     ns = curMinLen; */
/*   else */
    ns = 1;

  while (ns <= T->nsmax)
    {
      T->ns = ns;

      if (1) printf("%d steps    cost %d     %.1fs\n", ns, curMinCost,
		    (float) (clock()-Ct_debut) / CLOCKS_PER_SEC);

      T->fdp_solvable = StopSearch = 0;
      InitGrapheDActionsThreads(ns, InitState, NULL, thread);
      TOP_DEPART;
      initializeSousButsLimit(Goals, T->nsmax, ns-1, thread);

      StopSearch = 1;

      if (MakeConsistant(ns, thread))
/* 	if (forwardSearch (ns, 0, 0, thread)) */
/* 	  break; */
	forwardSearch (ns, 0, 0, thread);
      else
	StopSearch = 0;

      if (StopSearch)
	break;

      ns ++;
    }

  return ns;
}


int EnumerateActionsCFS(int ns, int e, int c, int thread)
{
  int i, nEA, a, *oh;
#ifdef MULTI_THREADS_FILE
  S_CONS * TC = &(CST[thread]);
#endif
  S_GR * G = &(GRT[thread]);
  int *pp, *qq;
  int occv = 1;

  EXPAND_PSUPPR(thread);

  nEA = G->nAi[e];
  for (pp = G->tAi[e], qq = G->ChoixEnumActions[e]; pp < G->tAi[e]+nEA; )
    *qq ++ = *pp ++;
  *qq = FNIL;

  for (i = 0; i < nEA; i++)
    {
      a = G->ChoixEnumActions[e][i];

      if (G->Ai[e][a])
	{
	  /* the cost constraint. */
	  if (minCosts)
	    if (c+LesActions[a].cost+minCosts[ns-(e+1)] >= curMinCost)
	      continue;

	  InitFilesConsistanceThreads(thread);

	  for (oh = G->ChoixEnumActions[e]; *oh != FNIL; oh ++)
	    if ((*oh != a) && (G->Ai[e][*oh]))
	      { ADD_FILE_ACTION(*oh, e, theThreads[thread], TC); }

	  if ( ! MakeConsistant (ns, thread))
	    { occv = 0; StopSearch = 0; goto DEFAIRE; }
	  if (! forwardSearch (ns, e+1, c+LesActions[a].cost, thread))
	    occv = 0;
	      
	DEFAIRE:
	  DeFaireLastSuppressionsThreads(ns, thread);
	}
    }

  DeFaireLastSuppressionsThreads(ns, thread);
  UNEXPAND_PSUPPR(thread);

  return occv;
}


int SplitActionsSetCFS(int ns, int e, int c, int thread)
{
  int nEA, type, f;
  S_GR * G = &(GRT[thread]);
  int *S1, *S2;
  int occv = 1;

  if ( ((nEA = G->nAi[e]) < SizeSplitMin) || 
       ((f = ChooseFactForSplit(e, ACTIONS_2SPLIT_COMMON_SUPP, 
				&type, thread)) == NONE))
    return EnumerateActionsCFS(ns, e, c, thread);

  EXPAND_PSUPPR(thread);

  SplitAiFromFact(e, f, type, &S1, &S2, ns, thread);

  InitFilesConsistanceThreads(thread);
  AddActionsSetToQueue(S2, e, thread);
  if (MakeConsistant(ns, thread))
    {
      if ( ! SplitActionsSetCFS (ns, e, c, thread))
	occv = 0;
    }
  else
    { StopSearch = 0; occv = 0; }

  DeFaireLastSuppressionsThreads(ns, thread);

  InitFilesConsistanceThreads(thread);
  AddActionsSetToQueue(S1, e, thread);
  if (MakeConsistant(ns, thread))
    {
      if ( ! SplitActionsSetCFS (ns, e, c, thread))
	occv = 0;
    }
  else
    { StopSearch = 0; occv = 0; }

  DeFaireLastSuppressionsThreads(ns, thread);
  UnsplitAiFromFact(thread);

  UNEXPAND_PSUPPR(thread);

  return occv;
}


int forwardSearch (int ns, int p, int c, int thread)
/* returns only cost constraints violations */
{
  THREAD T = theThreads+thread;
  S_GR * G = &(GRT[thread]);
  hash_list helt, hoelt;
  DLIST dce, odce;
  unsigned int * state = NULL;
  int h, cc, cco, isccv;

  if (p == ns)
    {
      if ((G->nFixedGoals[p] == nGoals) && (c < curMinCost))
	{
/* 	  printf("%d steps    minCost  %d  -->  %d\n", ns, curMinCost, c); */
	  curMinCost = c;
	  curMinLen = ns;
	  updateBestPlan (curMinLen, thread);
	}
      return 1;
    }

  VERIF_TIME_LIMIT (timelimit);

  h = hashSteph (nFaits, G->Fi[p], G->CurrentState);

  if (useHTinFS && search_hash_state(HT, h, ns-p, &helt, G->CurrentState))
    {
      StopSearch = 0;
      return 0; 
    }

  cc = searchStateCostHashTable(Hccv, h, ns-p, &helt, &dce, G);
  cco = searchStateCostHashTable(Ho, h, ns-p, &hoelt, &odce, G); /*deplacer */
  if (helt != NULL)
    {
      if ((dce != NULL) && (dce->val1 == ns-p))
	if (cc + c >= curMinCost)
	  return 1;
      if ((hoelt != NULL) && (odce != NULL) && (odce->val1 == ns-p))
	if (cco + c >= curMinCost)
	  { StopSearch = 0; return 0; }
    }
  else
    {
      if ((hoelt != NULL) && (odce != NULL) && (odce->val1 == ns-p))
	if (cco + c >= curMinCost)
	  { StopSearch = 0; return 0; }
    }

/*     if (( ! MinStepsRemoveActions) || (MakeConsistant(ns, thread))) PB */
  if (MinStepMaxAdds(p, ns, thread, MS_DONT_REMOVE_ACTIONS) == MS_FAILURE)
    {
      StopSearch = 0;
      return 0;
    }

  /* on duplique systematiquement, il faudra voir les cas ou c'est inutile. */
  state = duplicate(G->CurrentState, Hccv->size_states);

  switch (T->decMode)
    {
    case ACTIONS_ENUM:
      isccv = EnumerateActionsCFS (ns, p, c, thread);
      break;
    default: /* SPLIT ACTIONS SET */
      isccv = SplitActionsSetCFS (ns, p, c, thread);
      break;
    }

  if (isccv)
    {
      addStateCostHashTable (Hccv, h, ns-p, curMinCost-c, helt, dce, state);
      if (helt != NULL)
	free (state);
    }
  else 
    {
      addStateCostHashTable (Ho, h, ns-p, curMinCost-c, helt, dce, state);
      if (hoelt != NULL)
	free (state);
    }
/*     if (helt == NULL) */
/*       free(state); */


  return isccv;
}


/* ************************************************************** */
/*        CALCUL DES COUTS MINIMAUX DES FINS DE SEQUENCES         */
/* ************************************************************** */

int minCost;
int reachTimeLimitBMC;

void evaluateBackwardMinCosts (int m, int thread)
/* Search plans of lengths at most m, ignoring the initial state, and 
   memorize their minimal costs. */
{
  THREAD T = theThreads+thread;
  int ns;

  minCosts = malloc((T->nsmax+1) * sizeof(int));
  assert(minCosts != NULL);

  reachTimeLimitBMC = 0;

  for (ns = 1; ns <= m; ns ++)
    {
      T->ns = ns;

      T->fdp_solvable = 0;
      InitGrapheDActionsThreads(ns, NULL, Goals, thread);
      InitializeSousButs(Goals, ns, thread);
      minCost = ns*(maxActCost+1);
/*       printf("init minCost = %d\n", minCost); */

      if ( ! MakeConsistant(ns, thread))
	printf("aie\n");

      enumerateBackward(ns, ns-1, 0, thread);

      if (reachTimeLimitBMC)
	break;

      minCosts[ns] = minCost;

      if (T->fdp_solvable)
	{
	  fdp_solvable = 1;
	  break; 
	}
    }

  while (ns <= T->nsmax)
    {
      minCosts[ns] = minCosts[ns-1]+minActCost;
      ns ++;
    }

  TOP_ARRIVEE_THREAD(T);
  totalTime += T->Ctemps;
  totalChecks += T->nChecks;
  totalNE += T->NE;
  CTOP_ARRIVEE;
  StopSearch = 1;

}



void enumerateBackward(int ns, int e, int c, int thread)
{
  int i, nEA, a, *oh;
#ifdef MULTI_THREADS_FILE
  S_CONS * TC = &(CST[thread]);
#endif
  S_GR * G = &(GRT[thread]);
  int *pp, *qq;

  if (c+(e+1)*minActCost >= minCost)
    return ;

  if ((float) (clock()-theThreads[thread].Ct_debut) / CLOCKS_PER_SEC > 
      (float) timeLimitMinCosts)
    {
      reachTimeLimitBMC = 1;
      return ;
    }

  if (e < 0)
    {
      if (c < minCost)
	{
	  minCost = c;
	}
      return ;
    }

  EXPAND_PSUPPR(thread);

  nEA = G->nAi[e];
  for (pp = G->tAi[e], qq = G->ChoixEnumActions[e]; pp < G->tAi[e]+nEA; )
    *qq ++ = *pp ++;
  *qq = FNIL;

  for (i = 0; i < nEA; i++)
    {
      a = G->ChoixEnumActions[e][i];
      if (G->Ai[e][a])
	{
	  InitFilesConsistanceThreads(thread);

	  for (oh = G->ChoixEnumActions[e]; *oh != FNIL; oh ++)
	    if ((*oh != a) && (G->Ai[e][*oh]))
	      { ADD_FILE_ACTION(*oh, e, theThreads[thread], TC); }

	  if ((Sequences2ActionsInEnumerate) && (e > 0))
	    FilterPreviousStepUnorderedActions(a, e, ns, thread);
	  if (MakeConsistant(ns, thread))
	    if (e >= 0) 
	      enumerateBackward (ns, e-1, c+LesActions[a].cost, thread);

	  DeFaireLastSuppressionsThreads(ns, thread);
	}

      if (reachTimeLimitBMC)
	break;
    }

  DeFaireLastSuppressionsThreads(ns, thread);
  UNEXPAND_PSUPPR(thread);
}



