#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "consistance.h"
#include "main.h"
#include "sequences.h"
#include "sets.h"
#include "stats.h"
#include "structures.h"
#include "suppressions.h"
#include "search.h"

/* Variables. */

struct maillon_file_faits *FF, *FFF ; /* File des faits a supprimer. */
struct maillon_file_actions *AA, *AAA; /* File des actions a supprimer. */
struct maillon_file_faits *pF, *dF; /* Premier et dernier de la file. */
struct maillon_file_actions *pA, *dA; /* Premier et dernier de la file. */

SET FaitsFile;
SET ActionsFile;

int MKASupp; /* Nombre de suppressions d'actions juste sur ce coup. */


/* #define DEBUG_PROPAGATION */
/* #define FULL_DEBUG_PROPAGATION */


S_CONS * CST;


void AllocFilesConsistanceThreads(int ne, int thread)
{
  S_CONS * T = &(CST[thread]);
  THREAD Th = &(theThreads[thread]);
  int sizePileFaits = 2*Th->nFaits*(ne+1);
  int sizePileActions = Th->nActions*ne;

  FF = T->FF = malloc(sizePileFaits*sizeof(struct maillon_file_faits));
  assert(T->FF != NULL);
  FFF = T->FFF = T->FF+sizePileFaits;
  AA = T->AA = malloc(sizePileActions*sizeof(struct maillon_file_actions));
  assert(T->AA != NULL);
  AAA = T->AAA = T->AA+sizePileActions;
  FaitsFile = T->FaitsFile = AllocSet(2*Th->nFaits*(ne+1));
  ActionsFile = T->ActionsFile = AllocSet(Th->nActions*(ne));
}


void InitFilesConsistanceThreads(int thread)
{
  pF = dF = FF;
  pA = dA = AA;
}

void videLaFile (S_CONS * C, THREAD T)
{
  int e;
  
  while (pF != dF)
    {
      e = pF->s*2*T->nFaits + pF->f;
      RETe(e, FaitsFile);
      pF ++;
    }
  while (pA != dA)
    {
      e = pA->s * T->nActions + pA->a;
      RETe(e, ActionsFile);
      pA ++;
    }
}

int MakeConsistant(int ns, int t)
     /* On suppose ici que les files ont ete initialisees ET les faits/actions
	a supprimer ont ete inseres dans FF et AA. 
	t : le thread. */
{

  int f, a, s;
  S_CONS * C = &(CST[t]);
  S_GR * G = &(GRT[t]);
  THREAD Th = theThreads+t;

  MKASupp = 0;
  while ((pF != dF) || (pA != dA))
    {
      Th->nChecks ++;

      if (pF != dF)
	{
	  RETRACT_FILE_FACT(f, s, t, C);
	  if (Fi[s][OPPOSE(f)] == 0)
	    { videLaFile(C, Th); return 0; }
	  RETRACT_FACT(f, s, t, G);
	  PropagerSuppressionDuFait(f, s, ns, t);
	}
      else
	{
	  RETRACT_FILE_ACTION(a, s, t, C);
	  if ((nAi[s]) == 1)
	    { videLaFile(C, Th); return 0; }
	  RETRACT_ACTION(a, s, t, G);
	  PropagerSuppressionDeLAction(a, s, ns, t);
	  MKASupp ++;
	}
    }

  return 1;
}


void PropagerSuppressionDuFait(int f, int s, int ns, int thread)
{
  int *a;
  int of = OPPOSE(f);

  if (s < ns)
    for (a = LesFaits[f].pre; *a != FNIL; a ++)
      if (Ai[s][*a])
	ADD_FILE_ACTION(*a, s, theThreads[thread], TC); /* regle-actions 3. */

  if (s > 0)
    for (a = LesFaits[f].add; *a != FNIL; a ++)
      if (Ai[s-1][*a])
	ADD_FILE_ACTION(*a, s-1, theThreads[thread], TC); /* regle-actions 1. */

  if ((s < ns) && (Fi[s+1][f]) && (nbAM[s+1][of] == 0))
    ADD_FILE_FACT(f, s+1, theThreads[thread], TC); /* regle-faits 1. */

  if ((s > 0) && (Fi[s-1][f]) && (nbAM[s][f] == 0))
    ADD_FILE_FACT(f, s-1, theThreads[thread], TC); /* regle-faits 3. */

  if ((s < ns) && (Fi[s+1][of] == 0))
    for (a = tAi[s]; a < tAi[s]+nAi[s]; a ++)
      if ( ! ((THE_ACTION(*a))->isadd[f]))
	ADD_FILE_ACTION(*a, s, theThreads[thread], TC); /* regle-actions 4 */

  if ((s > 0) && (Fi[s-1][of] == 0))
    for (a = tAi[s-1]; a < tAi[s-1]+nAi[s-1]; a ++)
      if ( ! ((THE_ACTION(*a))->issup[f]))
	ADD_FILE_ACTION(*a, s-1, theThreads[thread], TC); /* regle-actions 5 ? */
}


/* ******************************************************************** */
/*               PROPAGATION DES SUPPRESSIONS D'ACTIONS.                */
/* ******************************************************************** */

void PropagerSuppressionDeLAction(int a, int s, int ns, int t)
/* t : le thread */
{
  int *p, op;
  FLIST q;
  S_GR * G = &(GRT[t]);
  int * nbARStepS = nbAR[s];
  int ** EstSousBut =  G->EstSousBut;
  int * NbAjoutsSousButsStepSMoins1;

  if (s > 0)
    NbAjoutsSousButsStepSMoins1 = G->NbAjoutsSousButs[s-1];

  for (p = LesActions[a].add; *p != FNIL; p ++)
    {
      op = OPPOSE(*p);
      (nbAM[s+1][op]) --;

      if (nbAM[s+1][op] == 0)
	{
	  /* Plus aucune action n'ajoute *p. */
	  if (Fi[s+1][*p] == 1)
	    {
	      if ((Fi[s][*p] == 0) || (nbAM[s+1][*p] == nAi[s]))
		/* p present a l'etape s+1 et (1) absent a l'etape s
		   ou (2) toutes les actions le suppriment. */
		{ 
		  ADD_FILE_FACT(*p, s+1, theThreads[t], TC); 
		}
	      else
		{
		  /* p est present a l'etape s et a l'etape s+1.
		     aucune action ne l'ajoute,
		     mais il y a des actions qui ne le suppriment pas. */
		  ADD_FLIST_NOTHREADS(*p, s+1);
		  * tSChg++ = (void *) (*(p));
		  * tSChg++ = (void *) (2*((s)+1)+1);
		}
	    }
	  if ((Fi[s][op] == 1) && (Fi[s+1][op] == 0))
	    /* Il faudrait qu'il y ait une action qui le supprime
	       et il n'y en a plus. */
	    ADD_FILE_FACT(op, s, theThreads[t], TC); /* regle-faits 3. */
	}

      if ((IS_IN_FLIST_NOTHREADS(op, s+1)) && (Fi[s+1][op] == 1))
	{
	  RETR_FLIST_NOTHREADS(op, s+1);
	  ADD_FLIST_NOTHREADS(op, s+1);
	}
    }


  for (p = LesActions[a].pre; *p != FNIL; p ++)
    {
      (nbARStepS [*p]) --;

      if ((s > 0) && (nbARStepS [*p] == 0) && ( ! EstSousBut[s+1][*p]))
	/* Forcement EstSousBut[s][*p] valait vrai. */
	{
	  int *pa;

	  EstSousBut[s][*p] = 0;
	  * tSChg++ = EstSousBut[s] + (*p);
	  for (pa = LesFaits[*p].add; *pa != FNIL; pa++)
	    if (Ai[s-1][*pa])
	      {
		NbAjoutsSousButsStepSMoins1[*pa] --;
		* tSChg++ = NbAjoutsSousButsStepSMoins1 + (*pa);
		if (NbAjoutsSousButsStepSMoins1[*pa] == 0)
		  {
		    ADD_FILE_ACTION(*pa, s-1, theThreads[t], TC);
		  }
	      }
	}
    }

  for (q = G->P[s+1][nAi[s]].suiv; q != &(G->P[s+1][nAi[s]]); q = q->suiv)
    if (Fi[s+1][q->fait])
      ADD_FILE_FACT(q->fait, s+1, theThreads[t], TC); /* regle-faits 2 */

  if (Sequences2Actions)
    {
      if (s > 0)
	for (p = Prev2Actions[a]; *p != FNIL; p ++)
	  if (-- (NbSucc2Seq[s-1][*p]) == 0)
	    if (Ai[s-1][*p])
	      ADD_FILE_ACTION(*p, s-1, theThreads[t], TC);
      if (s < ns-1)
	for (p = Next2Actions[a]; *p != FNIL; p ++)
	  if (-- (NbPre2Seq[s+1][*p]) == 0)
	    if (Ai[s+1][*p])
	      ADD_FILE_ACTION(*p, s+1, theThreads[t], TC);
    }
}



