#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <time.h>
#include <math.h>

#include "pddl.h"


#include "constantes.h"
#include "consistance.h"
#include "fdp.h"
#include "main.h"
#include "search.h"
#include "sequences.h"
#include "sets.h"
#include "split.h"
#include "structures.h"
#include "stats.h"
#include "suppressions.h"
#include "threads.h"

#include "hash.h"
#include "minsteps.h"




int theNoop;

int fdp_solvable;
int fdp_ns;
int fdp_makespan;





void RestitueEnumerate(int ns, int thread)
{
  int i;

  free(StepLevel);
  for (i = 0; i < ns; i++)
    {
      free(ChoixEnumActions[i]);
      free(EvalEnumActions[i]);
    }
  free(EvalEnumFacts);
  free(ChoixEnumFacts);
  free(EvalEnumActions);
  free(ChoixEnumActions);
}



void InitializeSousButs(int * finalstate, int ns, int thread)
{
  int i, s, *a, *p;
  int n = 0;
#ifdef MULTI_THREADS_FILE
  S_CONS * TC = &(CST[thread]);
#endif
  S_GR * G = &(GRT[thread]);
  int na = theThreads[thread].nActions;

  for (i = 0; i < 2*nFaits; i++)
    G->Maturite[i] = -1;
  for (i = 0; i < na; i++)
    G->NbAjoutsSousButs[ns-1][i] = 0;
  for (i = 0; i < 2*nFaits; i++)
    G->EstSousBut[ns][i] = 0;

  for (i = 0; i < 2*nFaits; i++)
    if (finalstate [OPPOSE(i)] == 0)
      {
	G->EstSousBut[ns][i] = 1;
	G->Maturite[i] = ns;
	/* Marquage des actions qui ajoutent le but. */
	for (a = LesFaits[i].add; *a != FNIL; a++)
	  G->NbAjoutsSousButs[ns-1][*a] ++;
      }

  for (i = 0; i < na; i++)
    if ((G->NbAjoutsSousButs[ns-1][i] == 0) && (i != theNoop))
      {
	ADD_FILE_ACTION(i, ns-1, theThreads[thread], TC);
	n ++;
      }

  /* On va marquer les sous buts iterativement en partant de la fin. */
  for (s = ns-1; s >= 0; s--)
    {
      for (i = 0; i < 2*nFaits; i++)
	G->EstSousBut[s][i] = G->EstSousBut[s+1][i];
      /* On calcule les sous buts a l'etape precedente. */
      for (i = 0; i < na; i++)
/* 	if ((Ai[s][i]) && (NbAjoutsSousButs[s][i] > 0)) */
	/* BUG : puisque on a l'action, il faut considerer ici que ses 
	   prerequis sont des sous-buts. On traitera le fait qu'il ne sont
	   pas sous-buts quand on fera la suppression effective de l'action. */
	/* COMMENTER 
	   if (NbAjoutsSousButs[s][i] > 0)
	*/
	if (G->Ai[s][i])
	  for (p = LesActions[i].pre; *p != FNIL; p++)
	    {
	      G->EstSousBut[s][*p] = 1;
	      if (G->Maturite[*p] < s)
		G->Maturite[*p] = s;
	    }

      /* On calcule les bonnes actions a l'etape precedente. */
      if (s > 0)
	{
	  for (i = 0; i < na; i++)
	    G->NbAjoutsSousButs[s-1][i] = 0;
	  for (i = 0; i < 2*nFaits; i++)
	    if (G->EstSousBut[s][i])
	      for (a = LesFaits[i].add; *a != FNIL; a++)
		G->NbAjoutsSousButs[s-1][*a] ++;
	  for (i = 0; i < na; i++)
	    if ((G->NbAjoutsSousButs[s-1][i] == 0) && (i != theNoop))
	      {
		ADD_FILE_ACTION(i, s-1, theThreads[thread], TC);
		/* printf ("Supp %d, %s\n", s-1, (LesActions+i)->nom); */
		n++;
	      }
	}
    }
}


void initializeSousButsLimit (int * finalstate, int ns, int limit, int thread)
/* Version speciale utilisee quand on ne sait pas ou chercher les buts.
   On ne met pas dans les files les actions qui apparaissent au dela 
   de la limite donnee. */
{
  int i, s, *a, *p;
  int n = 0;
#ifdef MULTI_THREADS_FILE
  S_CONS * TC = &(CST[thread]);
#endif
  S_GR * G = &(GRT[thread]);
  int na = theThreads[thread].nActions;

  for (i = 0; i < 2*nFaits; i++)
    G->Maturite[i] = -1;
  for (i = 0; i < na; i++)
    G->NbAjoutsSousButs[ns-1][i] = 0;
  for (i = 0; i < 2*nFaits; i++)
    G->EstSousBut[ns][i] = 0;

  for (i = 0; i < 2*nFaits; i++)
    if (finalstate [OPPOSE(i)] == 0)
      {
	G->EstSousBut[ns][i] = 1;
	G->Maturite[i] = ns;
	/* Marquage des actions qui ajoutent le but. */
	for (a = LesFaits[i].add; *a != FNIL; a++)
	  G->NbAjoutsSousButs[ns-1][*a] ++;
      }

  for (i = 0; i < na; i++)
    if ((G->NbAjoutsSousButs[ns-1][i] == 0) && (i != theNoop))
      if (ns-1 <= limit)
	{
	  ADD_FILE_ACTION(i, ns-1, theThreads[thread], TC);
	  n ++;
	}

  /* On va marquer les sous buts iterativement en partant de la fin. */
  for (s = ns-1; s >= 0; s--)
    {
      for (i = 0; i < 2*nFaits; i++)
	G->EstSousBut[s][i] = G->EstSousBut[s+1][i];
      /* On calcule les sous buts a l'etape precedente. */
      for (i = 0; i < na; i++)
/* 	if ((Ai[s][i]) && (NbAjoutsSousButs[s][i] > 0)) */
	/* BUG : puisque on a l'action, il faut considerer ici que ses 
	   prerequis sont des sous-buts. On traitera le fait qu'il ne sont
	   pas sous-buts quand on fera la suppression effective de l'action. */
	/* COMMENTER 
	   if (NbAjoutsSousButs[s][i] > 0)
	*/
	if (G->Ai[s][i])
	  for (p = LesActions[i].pre; *p != FNIL; p++)
	    {
	      G->EstSousBut[s][*p] = 1;
	      if (G->Maturite[*p] < s)
		G->Maturite[*p] = s;
	    }

      /* On calcule les bonnes actions a l'etape precedente. */
      if (s > 0)
	{
	  for (i = 0; i < na; i++)
	    G->NbAjoutsSousButs[s-1][i] = 0;
	  for (i = 0; i < 2*nFaits; i++)
	    if (G->EstSousBut[s][i])
	      for (a = LesFaits[i].add; *a != FNIL; a++)
		G->NbAjoutsSousButs[s-1][*a] ++;
	  for (i = 0; i < na; i++)
	    if ((G->NbAjoutsSousButs[s-1][i] == 0) && (i != theNoop))
	      if (s-1 <= limit)
		{
		  ADD_FILE_ACTION(i, s-1, theThreads[thread], TC);
		  /* printf ("Supp %d, %s\n", s-1, (LesActions+i)->nom); */
		  n++;
		}
	}
    }
}



void FDP_Allocations_Initializations(int nsmax, int thread)
{
  AllocFilesConsistanceThreads(nsmax, thread);
}


void Solution(int ns, int mode, int thread)
{
  theThreads[thread].fdp_solvable = 1;

  if (theThreads[thread].justTest)
    {
      if ( ! noThreads)
	printf("ATTENTION : justTest peut stopper les autres threads !\n");
      StopSearch = 1;
      return ;
    }

#ifdef TRACE_THREADS
  printf("** T%d ns=%d     fdp_ns=%d  ", thread, ns, fdp_ns);
  PrintThreadsPlanLengths();
#endif
      nSol += 1;
      fdp_ns = ns;
      GRT[thread].cost = evaluatePlanCost (ns, thread);
      if (FirstSolution)
	StopSearch = 1;
}



void EnumerateAndMemorize(int ns, int p, int e, int thread)
{
  THREAD T = theThreads+thread;
  S_GR * G = &(GRT[thread]);

  if ((p < ns) && (G->nFi[p] == nFaits))
    {
      int h, dist;
      unsigned int * state = NULL;
      hash_list curHash;

      h = hashSteph (nFaits, G->Fi[p], G->CurrentState);
      if (search_hash_state(HT, h, ns-p, &curHash, G->CurrentState))
	return ;

/*       if (regressionProc) */
/* 	if (StateCannotReachGoals (G->Fi[p], ns-p)) */
/* 	  { totalRegTPrunes ++; return ; } */

      if (curHash == NULL)
	state = duplicate(G->CurrentState, sizeStates);
      else
	dist = curHash->distance;

      ajouter_etat_hash(HT, state, curHash, ns-p, h, 1, ns);
      if (MinStepMaxAdds(p, ns, thread, MS_REMOVE_ACTIONS) != MS_FAILURE)
	{
	  if (( ! MinStepsRemoveActions) || (MakeConsistant(ns, thread)))
	    Enumerate(ns, p, e, thread);
	  if (T->fdp_solvable)
	    {
	      /* on defait la memorisation de l'etat courant. */
	      if (curHash == NULL)
		{
		  hash_list tmp = HT [h].liste;
		  HT [h].liste = tmp->next;
		  free (tmp);
		}
	      else
		curHash->distance = dist;
	    }
	}
    }
  else
    Enumerate(ns, p, e, thread);
}
  


void Enumerate(int ns, int p, int e, int thread)
{
  THREAD T = theThreads+thread;
  S_GR * G = &(GRT[thread]);

  if (G->nAiSingles == ns)
      {
	Solution(ns, T->mode, thread);
	return;
      }

  if (timelimit > 0) 
    {
      if ((float) (clock()-Ct_debut) / CLOCKS_PER_SEC > (float) timelimit)
	{ 
	  printf("TIME LIMIT EXCEEDED %d seconds  (%.2f)\n", timelimit,
		 (float) (clock()-Ct_debut) / CLOCKS_PER_SEC); 
	  exit(0); 
	}
    }
  
  EXPAND_PSUPPR(thread);
  switch (T->decMode)
    {
    case ACTIONS_ENUM:
      EnumerateActions(ns, p, thread);
      break;
    default:
      SplitActionsSet(ns, e, thread);
      break;
    }
  UNEXPAND_PSUPPR(thread);
}


#ifdef MULTI_THREADS
void EnumerateActions(int ns, int p, int thread)
{
  /* Choix tres simple : on enumere sur chaque actions de l'ensemble Ai
     le plus petit et qui contient plus d'une action. */
  int i, e, nEA, a, *oh;
#ifdef MULTI_THREADS_FILE
  S_CONS * TC = &(CST[thread]);
#endif
  S_GR * G = &(GRT[thread]);
  THREAD Th = theThreads+thread;
  int *pp, *qq;

  e = p;

  nEA = G->nAi[e];
  for (pp = G->tAi[e], qq = G->ChoixEnumActions[e]; pp < G->tAi[e]+nEA; )
    *qq++ = *pp++;
  *qq = FNIL;


  for (i = 0, Th->NE += 1; i < nEA; i++, Th->NE++)
    {
      if (G->Ai[e][G->ChoixEnumActions[e][i]])
	{
	  a = G->ChoixEnumActions[e][i];

	  EXPAND_PSUPPR(thread);
	  InitFilesConsistanceThreads(thread);
	  for (oh = G->ChoixEnumActions[e]+i+1; *oh != FNIL; oh++)
	    if (G->Ai[e][*oh]) /* Une action precedente peut l'avoir supprimee */
	      { ADD_FILE_ACTION(*oh, e, theThreads[thread], TC); }
	  if (e < ns-1) {
	    if (Sequences2ActionsInEnumerate)
	      FilterNextStepUnorderedActions(a, e, ns, thread);
	  }
	  if (MakeConsistant(ns, thread))
	    EnumerateAndMemorize(ns, e+1, e+1, thread);

	  DeFaireLastSuppressionsThreads(ns, thread);
	  UNEXPAND_PSUPPR(thread);
	  if ((StopSearch) && (ns >= fdp_ns))
	    break;
	  if (i < nEA-1)
	    {
	      InitFilesConsistanceThreads(thread);
	      ADD_FILE_ACTION(a, e, theThreads[thread], TC);
	      if ( ! MakeConsistant(ns, thread))
		break;
	    }
	}
    }
  DeFaireLastSuppressionsThreads(ns, thread);
}
#else
void EnumerateActions(int ns, int p, int thread)
{
  /* Choix tres simple : on enumere sur chaque actions de l'ensemble Ai
     le plus petit et qui contient plus d'une action. */
  int i, e, nEA, a, *oh;
#ifdef MULTI_THREADS_FILE
  S_CONS * TC = &(CST[thread]);
#endif
  THREAD Th = theThreads+thread;
  int *pp, *qq;

  e = p;

  nEA = nAi[e];
  for (pp = tAi[e], qq = ChoixEnumActions[e]; pp < tAi[e]+nEA; )
    *qq++ = *pp++;
  *qq = FNIL;


  for (i = 0, Th->NE += 1; i < nEA; i++, Th->NE++)
    {
      if (Ai[e][ChoixEnumActions[e][i]])
	{
	  a = ChoixEnumActions[e][i];

	  EXPAND_PSUPPR(thread);
	  InitFilesConsistanceThreads(thread);
	  for (oh = ChoixEnumActions[e]+i+1; *oh != FNIL; oh++)
	    if (Ai[e][*oh]) /* Une action precedente peut l'avoir supprimee */
	      { ADD_FILE_ACTION(*oh, e, theThreads[thread], TC); }
	  if (e < ns-1) {
	    if (Sequences2ActionsInEnumerate)
	      FilterNextStepUnorderedActions(a, e, ns, thread);
	  }
	  if (MakeConsistant(ns, thread))
	    EnumerateAndMemorize(ns, e+1, e+1, thread);

	  DeFaireLastSuppressionsThreads(ns, thread);
	  UNEXPAND_PSUPPR(thread);
	  if ((StopSearch) && (ns >= fdp_ns))
	    break;
	  if (i < nEA-1)
	    {
	      InitFilesConsistanceThreads(thread);
	      ADD_FILE_ACTION(a, e, theThreads[thread], TC);
	      if ( ! MakeConsistant(ns, thread))
		break;
	    }
	}
    }
  DeFaireLastSuppressionsThreads(ns, thread);
}
#endif



#define SHUNTE_SPLIT_MEMORISATION /* ca a l'air un peu mieux */

void SplitActionsSet(int ns, int e, int thread)
{
  /* L'idee est de splitter l'ensemble Ai en deux parties: l'une contenant les 
     actions qui ont en commun la suppression du fait f qui apparait le plus 
     parmi les suppressions. L'autre contient les actions restantes.
     Le choix de Ai est delicat : 
     -- le plus grand (pour simplifier beaucoup le graphe au debut),
     -- le plus petit (choix habituel de la variable la plus contrainte),
     -- celui ou on a le plus grand nombre d'actions qui ont une suppression 
     commune,
     -- celui ou le split est le plus equilibre.
  */
  THREAD T = theThreads+thread;
  int f = 0, ftype;
  S_GR * G = &(GRT[thread]);
/*   int PreserveFirst; */
/*   int FirstSet = ADDOTHERS; */
  int *S1, *S2;
  
  if (G->nAi[e] < SizeSplitMin)
    {
      EnumerateActions(ns, e, thread);
      return ;
    }

  if ((f = ChooseFactForSplit(e, T->decMode, &ftype, thread)) == NONE)
    {
      EnumerateActions(ns, e, thread);
      return ;
    }

  SplitAiFromFact(e, f, ftype, &S1, &S2, ns, thread);
  T->NE ++ ;
  InitFilesConsistanceThreads(thread);
  AddActionsSetToQueue(S2, e, thread);
  if (MakeConsistant(ns, thread))
    {
      if (Sequences2ActionsInEnumerate && (G->nAi[e] <= MaxSize2SeqFiltering)
	  && (e+1 < ns))
	if (FilterNextStep2Sequences(e, ns, thread))
	  if ( ! MakeConsistant(ns, thread))
	    goto DEFAIRE_SUPPRESSIONS;
#ifdef SHUNTE_SPLIT_MEMORISATION
      Enumerate( ns, e+1, e, thread);
#else
      EnumerateAndMemorize( ns, e+1, e, thread);
#endif
      if ((StopSearch) && (ns >= fdp_ns))
	return;
    }
 DEFAIRE_SUPPRESSIONS:
  DeFaireLastSuppressionsThreads(ns, thread);
  /* On garde les actions qui ne suppriment pas f. */
  T->NE ++ ;
  InitFilesConsistanceThreads(thread);
  AddActionsSetToQueue(S1, e, thread);
  if (MakeConsistant(ns, thread))
    {
      if (Sequences2ActionsInEnumerate && (G->nAi[e] <= MaxSize2SeqFiltering)
	  && (e+1 < ns))
	if (FilterNextStep2Sequences(e, ns, thread))
	  if ( ! MakeConsistant(ns, thread))
	    goto DEFAIRE_SUPPRESSIONS2;
#ifdef SHUNTE_SPLIT_MEMORISATION
      Enumerate( ns, e+1, e, thread);
#else
      EnumerateAndMemorize( ns, e+1, e, thread);
#endif
      if ((StopSearch) && (ns >= fdp_ns))
	return;
    }
 DEFAIRE_SUPPRESSIONS2:
  DeFaireLastSuppressionsThreads(ns, thread);
  UnsplitAiFromFact(thread);
  return ;
}


