#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <pthread.h>

#include "constantes.h"
#include "hash.h"
#include "main.h"

hash_table_t HT; /* La table des echecs. */
pthread_mutex_t * HTmutexes; /* Table des mutex variables pour gerer l'acces 
				a la meme liste de la table. */


unsigned int *CurrentState; /* Etat courant sous forme de champ de bits */
hash_list CurrentHash; /* Le maillon correspondant a CurrentState 
			  dans la table. */
int sizeStates; /* Nombre de (unsigned int) pour les etats */

unsigned int sizeHT = 0;
unsigned int changesHT = 0;

int (* hash) (int, int *, int , unsigned int * );


unsigned int * duplicate(unsigned int *E, int size)
{
  unsigned int *D = malloc(size * sizeof(unsigned int));

  assert(D != NULL);
  memcpy(D, E, (size_t) size * sizeof(unsigned int));
  return D;
}


hash_table_t alloc_hash_table (size_t size, int nfaits)
{
  int          i;
  hash_table_t htable, p;

  htable = (hash_table_t) malloc (size * sizeof (struct hash_table_elt));
  assert (htable != NULL);
  p = htable;

  for (i = 0; i < size; ++i)
    {
      p->nombre = 0;
      p->liste  = NULL;
      ++ p;
    }

  sizeStates = (nfaits+SIZE_SEG_HASH-1)/SIZE_SEG_HASH;

  CurrentState = malloc(sizeStates*sizeof(unsigned int));
  assert(CurrentState != NULL);

  HTmutexes = malloc(size * sizeof(pthread_mutex_t));
  assert(HTmutexes != NULL);
  for (i = 0; i < size; ++ i)
    pthread_mutex_init( & (HTmutexes[i]), NULL);

  return htable;
}

int hashSteph (int nfaits, int *etat, unsigned int *p)
{
  int decal, i;
  unsigned int * pp = p;
/*   unsigned int *p = CurrentState; propre au thread maintenant. */

  for (i = decal = 0, *p = (unsigned int) 0; i < 2*nfaits; i += 2)
    {
      if (etat[i])
	*p += ((unsigned int) 1 << decal);
      if (decal == 31)
	{
	  * ++ p = (unsigned int) 0;
	  decal = 0;
	}
      else
	decal ++;
    }

  while (p < pp+sizeStates-1)
    * ++ p = (unsigned int) 0;

  return ((int) hash_bits (pp, sizeStates));
}

unsigned int hash_bits (unsigned int *etat, int sizeStates)
{
  int i;
  unsigned int h = 0;

  for (i = 0; i < sizeStates; i ++)
    {
      h = (h*65537 + *etat) % HASH_SIZE;
      etat ++;
    }
  return (h);
}


hash_list new_hash_list (unsigned int *etat, int distance, 
			 unsigned char * dists, hash_list next)
{
    hash_list p = (hash_list) malloc (sizeof (struct hash_list_elt));
    assert (p != NULL);
    p->etat  = etat;
    p->distance = distance;
    p->dists = dists;
    p->next  = next;

    return p;

}


void ajouter_etat_hash (hash_table_t table, unsigned int *state, hash_list ch,
			int dist, int h, int all, int ns)
/* all : indique si on doit memoriser la plage de distance ou juste dist. 
   dist : nombre d'etapes qui separe etat des buts.
   prev_dist : si vaut &i_NONE il faut creer une nouvelle entree
   sinon il suffit de mettre a jour. */
{
  if ( ! noThreads)
    pthread_mutex_lock (&(HTmutexes[h]));
  if (ch == NULL)
    {
      table [h].nombre ++;
      table[h].liste = new_hash_list(state, dist, NULL, table [h].liste);
      sizeHT ++;
    }
  else
    {
      if (dist < ch->distance) printf("HOUBA !!\n");
      ch->distance = dist;
      changesHT ++;
    }
  if ( ! noThreads)
   pthread_mutex_unlock (&(HTmutexes[h]));
}



int meme_bits_etat (unsigned int *e1, unsigned int *e2)
{
  int i;

  for (i = 0; i < sizeStates; ++ i)
    if (*(e1++) != *(e2++))
      return 0;
  return 1;
}



int search_hash_state (hash_table_t table, int h, int distance, 
				hash_list * CH, unsigned int *CS)
{
  int res;

  if ( ! noThreads)
    pthread_mutex_lock (&(HTmutexes[h]));

  for (*CH = table [h].liste; *CH != NULL; *CH = (*CH)->next)
    if (meme_bits_etat (CS, (*CH)->etat))
      break;
  if (*CH == NULL)
    res = 0;
  else
  /* C'est bon si la distance est inferieure a la distance dans la table
     ou si la distance est marquee dans la table. */
    res =  (distance <= (*CH)->distance);

  if ( ! noThreads)
    pthread_mutex_unlock (&(HTmutexes[h]));
  return res;
}



void dump_hash_table (hash_table_t table, size_t size)
{
  int i = 0, j = 0;
  printf ("dumping hash_table\n");
  for (i = 0; i < size; ++i)
    {
      if (table[i].nombre)
	{
	  j++;
	  printf ("[%d] -> %d\n", i, table[i].nombre);
	}
    }
  printf ("%.2f\n", (float) 100.0*j/size);

}

void AffStatsHash(hash_table_t table, size_t size)
{
  double s = 0, e = 0, S = 0;
  int i, n = 0;
  
  for (i = 0; i < size; ++i)
    {
      if (table[i].nombre > 0)
	{
	  S += table[i].nombre;
	  n++;
	}
      s += table[i].nombre;
    }
  s = s/size;
  for (i = 0; i < size; ++i)
    e += (s-table[i].nombre)*(s-table[i].nombre);
  if (UseHashCyril) printf ("[HC ");
  else printf ("[HS ");
  printf("%f  (%f/%d)]  ", e/size, S/n, n);
}


