#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "constantes.h"
#include "hash.h"
#include "hasht.h"
#include "main.h"
#include "structures.h"
#include "utils.h"



HASHT newHashTable (int size, int size_sets, int srsv)
{
  HASHT H;
  
  H = malloc(sizeof(struct hasht));
  assert(H != NULL);

  H->t = malloc (size * sizeof (struct hash_table_elt));
  assert(H->t != NULL);
  H->size = size;

  H->size_states = (size_sets+SIZE_SEG_HASH-1)/SIZE_SEG_HASH;

  H->size_rsv = srsv;
  H->rsv = malloc (srsv * sizeof(struct hash_list_elt));
  assert(H->rsv != NULL);

  H->top_rsv = H->rsv;

  H->rsv_list = NULL;

  return H;
}

void freeHashTable (HASHT H)
{
  LIST p, pp;

  p = H->rsv_list;
  while (p != NULL)
    {
      pp = p;
      p = p->suiv;
      free ((struct hash_list_elt *) (pp->val));
      free(pp);
    }

  free(H->t);
  free(H->rsv);

  free(H);
}


void reallocHashTableReserve (HASHT H)
{
  H->rsv_list = CreerMaillon ((int) H->rsv, NULL, H->rsv_list);
  H->rsv = malloc (H->size_rsv * sizeof(struct hash_list_elt));
  assert(H->rsv != NULL);

  H->top_rsv = H->rsv;
}

hash_list newHashElt (HASHT H, unsigned int *etat, int distance, 
			 unsigned char * dists, hash_list next)
{
    hash_list p;
    
    if (H->top_rsv == H->rsv + H->size_rsv)
      reallocHashTableReserve (H);

    
    p = H->top_rsv;
    p->etat  = etat;
    p->distance = distance;
    p->dists = dists;
    p->next  = next;

    H->top_rsv ++;

    return p;

}

void retractHashElt (HASHT H, int h, hash_list the)
{
  hash_list p, pp;

  pp = NULL;
  p = H->t[h].liste;
  while (p != the)
    {
      pp = p;
      p = p->next;
    }
  if (pp == NULL)
    H->t[h].liste = p->next;
  else
    pp->next = p->next;

  /* Un jour il faudra faire un truc pour recuperer ces maillons libres. */
}


/* ************************************************************** */
/*            Tables des etats avec distances et couts.           */
/* ************************************************************** */


int searchStateCostHashTable (HASHT H, int h, int distance, hash_list * pp,
			      DLIST *cur, S_GR * G)
{
  hash_list p;
  DLIST q;
  unsigned int *state = G->CurrentState;

  for (p = H->t[h].liste; p != NULL; p = p->next)
    if ( ! memcmp (state, p->etat, sizeof(unsigned int) * H->size_states))
      break;

  *pp = p;

  if (p == NULL)
    return 0;

  q = p->distcosts;
  while (q != NULL)
    {
      if (q->val1 == distance)
	{
	  *cur = q;
	  return (q->val2);
	}
      if (q->val1 < distance)
	break;
      q = q->suiv;
    }
  *cur = q;
  return 0;

}



void addStateCostHashTable (HASHT H, int h, int d, int c, hash_list e, 
			    DLIST cur, unsigned int *state)
{
  DLIST q, qq;

  if (e == NULL)
    {
/*       printf("add state : %d %d %d\n", h, d, c); */
      H->t [h].nombre ++;
      e = H->t [h].liste = newHashElt(H, state, 0, NULL, H->t [h].liste);
      e->distcosts = CreerDMaillon (d, c, NULL, NULL);
      return ;
    }

  if ((cur != NULL) && (cur->val1 == d))
    {
      if (c > cur->val2)
	cur->val2 = c;
      return;
    }

  qq = NULL;
  q = e->distcosts;
  while (q != NULL)
    {
      if (q->val1 < d)
	break;
      qq = q;
      q = q->suiv;
    }

  if (qq == NULL)
    e->distcosts = CreerDMaillon (d, c, NULL, q);
  else
    qq->suiv = CreerDMaillon (d, c, NULL, q);
}
