#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <string.h>
#include <sys/times.h>
#include <unistd.h>


#include "pddl.h"

#include "consistance.h"
#include "constantes.h"
#include "fdp.h"
#include "cfdp.h"
#include "hash.h"
#include "main.h"
#include "minsteps.h"
#include "search.h"
#include "sequences.h"
#include "stats.h"
#include "structures.h"
#include "suppressions.h"
#include "threads.h"


extern char * NomFicProbleme;

struct thread * theThreads;
pthread_t * theThreads_t;
int * threadsNum;

int StopSearch; /* Force tous les recherches a stopper. */

float totalTime = 0;
float totalChecks = 0;
float totalNE = 0;
float totalRegTPrunes = 0;
float totalRegTTests = 0;

struct tms    now;  


int Search(int lecturemode, int searchmode)
{
  int ns;

  ns = Lecture(lecturemode);

  StopSearch = 0;
  fdp_ns = ns+1;
  fdp_solvable = 0;

  HT = alloc_hash_table (HASH_SIZE, nFaits);


  nThreads = 1;
  AllocThreads(nThreads, nFaits);


  theThreads[0].nFaits = nFaits;
    theThreads[0].nActions = nActions;
  theThreads[0].nsmax = ns;
  threadsNum[0] = 0;
  theThreads[0].decMode = ModeDecomposition;
  theThreads[0].justTest = 0;
  AllocPilesSuppressionsThread(0);
  GRT[0].Sequences2Actions = 1;
  AllocGrapheDActionThreads(theThreads[0].nsmax, 0);
  GRT[0].OrdActionsEnum = 0;
  GRT[0].LevelOrdering = 0;
  GRT[0].ComputeParallelPlan = ComputeParallelPlan;
      
  if (searchmode & ((unsigned int) FDP_MINCOST_SEARCH))
    CFDP (&(threadsNum[0]));

  times(&now);
  assert( times(&now) != ( (clock_t) -1 ));
  Ctemps = (float) ( now.tms_utime + now.tms_stime + now.tms_cutime + now.tms_cstime ) / sysconf(_SC_CLK_TCK);
  CnChecks = totalChecks;
  CNE = totalNE;


  if (IPCOutput)
    PrintIPCOutput(searchmode);
  PrintStandardOutput(searchmode);
  return 1;
}


void AllocThreads(int n, int nf)
{
  theThreads = malloc(n*sizeof(struct thread));
  assert(theThreads != NULL);

  theThreads_t = malloc(n*sizeof(pthread_t));
  assert(theThreads_t != NULL);

  threadsNum = malloc(n*sizeof(int));
  assert(threadsNum != NULL);

  /* Table des piles des suppressions. */
  SST = malloc(n*sizeof(S_SUPP));
  assert(SST != NULL);

  /* Table des piles de consistance. */
  CST = malloc(n*sizeof(S_CONS));
  assert(CST != NULL);

  /* Graphes d'actions */
  GRT = malloc(n*sizeof(S_GR));
  assert(GRT != NULL);

  /* Tables pour l'heuristique Minsteps. */
  nGoalsAdded = calloc(n, sizeof(int *));
  assert(nGoalsAdded != NULL);
  
  /* ALLOCATIONS COMMUNES AUX DIVERS THREADS. */
  AllocTheGoals(nf);

  if ((Sequences2Actions) || (Sequences2ActionsInEnumerate)
      || (ConstructPred2SeqSets))
    ConstructionSequences2Actions();

}


void PrintIPCOutput(int searchmode)
{
  if (fdp_solvable)
    {
      FILE * f;
      char * s;

      s = malloc(strlen(NomFicProbleme)*sizeof(char)+5+1);
      assert(s != NULL);
      strcpy(s, NomFicProbleme);
/*       strcpy(s+strlen(s)-4, "soln"); */
      strcpy(s+strlen(s), ".soln");
      f = fopen(s, "w");
      assert(f != NULL);
      fPrintIPCOutput( searchmode, f );
      fclose(f);
    }
}

void fPrintIPCOutput(int searchmode, FILE * f)
/* ; Time <total-planner-time> */
/* ; ParsingTime <parsing-time> */
/* ; NrActions <nr-actions-in-plan> */
/* ; MakeSpan <makespan-of-plan> */
/* ; MetricValue <metricvalue-of-plan> */
/* ; PlanningTechnique <plan-generation-method(s)> */
/* <starttime>: (<action-name> <param_1> <param_2> .. <param_k>) [<action-duration>] */
/* . */
/* . */
/* . */
/* <starttime>: (<action-name> <param_1> <param_2> .. <param_k>) [<action-duration>] */
{
  /*   S_GR * G = &(GRT[thread]); */

  int i;

  fprintf(f, "; Time %.2f\n", tempsLecture+Ctemps);
  fprintf(f, "; ParsingTime %.2f\n", tempsLecture);
  fprintf(f, "; NrActions %d\n", bestCFDPlen);
  fprintf(f, "; MakeSpan\n");
  fprintf(f, "; MetricValue\n");
  fprintf(f, "; PlanningTechnique\n");
  for (i = 0; i < bestCFDPlen; i++)
    fprintf(f, "%d: %s [1]\n", i, LesActions[bestCFDP[i]].nom);
}

void PrintStandardOutput(int searchmode)
{
  printf ("%da %df ", theThreads->nActions, theThreads->nFaits);
  PrintOptions();

  if (searchmode & ((unsigned int) FDP_MINCOST_SEARCH))
    {
      printf(" %d %d$]  O  %d %.2fs\n", curMinLen, curMinCost, 
	     CnChecks, tempsLecture+Ctemps);
      return ;
    }


  if (fdp_solvable)
    {
      printf(" %d/%d]  O  %d %d %.2fs", fdp_ns, fdp_makespan, 
	     CNE, CnChecks, tempsLecture+Ctemps);
    }
  else
    {
      printf(" %d/%d]  X  %d %d %.2fs", fdp_ns, fdp_makespan, CNE, CnChecks, Ctemps);
    }
  printf("\n");
}

int Lecture(int mode)
{
  int nsmax;
  int StabiliseDAbord = 1;

  CTOP_DEPART;
  AllocTablesActionsFaits();
  maxActCost = minActCost = NONE;

  nsmax = analyser();
  if (nStepsImposed) 
    nSteps = nsmax = nStepsImposed;
  
  if (StabiliseDAbord)
    {
      stabilize_graph();
      ConstructionTablesNewActions(0, nActions-1, SizeAllocFaits);
    }
  
  CTOP_ARRIVEE;
  tempsLecture = Ctemps;

  if (AfficherLeProbleme)
    AfficheLeProbleme();

  return nsmax;
}
