#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "consistance.h"
#include "fdp.h"
#include "main.h"
#include "search.h"
#include "structures.h"
#include "sequences.h"

/* 2SEQUENCES */
#ifdef ORDERED2SEQ_ARRAYS
char **Seq2Actions; /* Table de validite des sequences de 2 actions. */
#endif
#ifdef ORDERED2SEQ_SETS
SET * Seq2Actions;
SET * pred2SeqSet;
#endif
int **Next2Actions; /* Les actions qui peuvent apparaitre apres. */
int *nNext2Actions;
int **Prev2Actions; /* Les actions qui peuvent apparaitre avant. */
int *nPrev2Actions;

int **NbSucc2Seq, **NbPre2Seq; /* Dans le graphe d'actions, pour chaque 
				  action on a le nombre d'actions de l'etat
				  precedent avec qui ca fait une 2sequence
				  et idem avec l'etape suivante. */




/* ************************************************************* */
/*                          2-sequences                          */
/* ************************************************************* */

#ifdef MULTI_THREADS

int FilterNextStep2Sequences(int e, int ns, int t)
/* Met dans la file de propagation les actions de l'etape e+1 qui n'ont pas 
   de support 2sequence a l'etape precedente. 
   t : le thread. */
/* Renvoie 1 si il y a eu des eliminations. */
{
  int *p, *q, iana = 0;
#ifdef MULTI_THREADS_FILE
  S_CONS * TC = &(CST[t]);
#endif
  S_GR * G = &(GRT[t]);

  if (e == ns-1)
    return 0;
  for (p = G->tAi[e+1]; p < G->tAi[e+1]+G->nAi[e+1]; p++)
    {
      for (q = G->tAi[e]; q < G->tAi[e]+G->nAi[e]; q++)
#ifdef ORDERED2SEQ_SETS
	if (IN(*p, Seq2Actions[*q]))
#endif
#ifdef ORDERED2SEQ_ARRAYS
	  if (Seq2Actions[*q][*p])
#endif
	    goto GO_AFTER;
      
    GO_AFTER:
      if (q == G->tAi[e]+G->nAi[e])
	{ 
	  ADD_FILE_ACTION(*p, e+1, theThreads[t], TC);
	  iana = 1;
	}
    }
  return iana;
}

int OOFilterNextStepUnorderedActions(int a, int e, int ns, int t)
/* Met dans la file de propagation les actions de l'etape e+1 qui n'ont pas 
   de support 2sequence avec l'action a choisie a l'etape e.
   t : le thread. */
/* Renvoie 1 si il y a eu des eliminations. */
{
  int *p, iana = 0;
#ifdef MULTI_THREADS_FILE
  S_CONS * TC = &(CST[t]);
#endif
  S_GR * G = &(GRT[t]);

  for (p = G->tAi[e+1]; p < G->tAi[e+1]+G->nAi[e+1]; p++)
    {
#ifdef ORDERED2SEQ_SETS
	    if ( ! IN(*p, Seq2Actions[a]))
#endif
#ifdef ORDERED2SEQ_ARRAYS
	    if ( ! Seq2Actions[a][*p])
#endif
	      { 
		ADD_FILE_ACTION(*p, e+1, theThreads[t], TC);
		iana = 1;
	      }
    }
  return iana;
}



int FilterPreviousStepUnorderedActions(int a, int e, int ns, int t)
/* Met dans la file de propagation les actions de l'etape e-1 qui n'ont pas 
   de support 2sequence avec l'action a choisie a l'etape e.
   t : le thread. */
/* Renvoie 1 si il y a eu des eliminations. */
{
  int *p, iana = 0;
#ifdef MULTI_THREADS_FILE
  S_CONS * TC = &(CST[t]);
#endif
  S_GR * G = &(GRT[t]);

  for (p = G->tAi[e-1]; p < G->tAi[e-1]+G->nAi[e-1]; p++)
    {
#ifdef ORDERED2SEQ_SETS
	  if ( ! IN(a, Seq2Actions[*p]))
#endif
#ifdef ORDERED2SEQ_ARRAYS
	    if ( ! Seq2Actions[*p][a])
#endif
	      { 
		ADD_FILE_ACTION(*p, e-1, theThreads[t], TC);
		iana = 1;
	      }
	  break;
    }
  return iana;
}



#else



int FilterNextStep2Sequences(int e, int ns, int t)
/* Met dans la file de propagation les actions de l'etape e+1 qui n'ont pas 
   de support 2sequence a l'etape precedente. 
   t : le thread. */
/* Renvoie 1 si il y a eu des eliminations. */
{
  int *p, *q, iana = 0;
#ifdef MULTI_THREADS_FILE
  S_CONS * TC = &(CST[t]);
#endif

  if (e == ns-1)
    return 0;
  for (p = tAi[e+1]; p < tAi[e+1]+nAi[e+1]; p++)
    {
      for (q = tAi[e]; q < tAi[e]+nAi[e]; q++)
#ifdef ORDERED2SEQ_SETS
	if (IN(*p, Seq2Actions[*q]))
#endif
#ifdef ORDERED2SEQ_ARRAYS
	  if (Seq2Actions[*q][*p])
#endif
	    goto GO_AFTER;
      
    GO_AFTER:
      if (q == tAi[e]+nAi[e])
	{ 
	  ADD_FILE_ACTION(*p, e+1, theThreads[t], TC);
	  iana = 1;
	}
    }
  return iana;
}


int FilterNextStepUnorderedActions(int a, int e, int ns, int t)
/* Met dans la file de propagation les actions de l'etape e+1 qui n'ont pas 
   de support 2sequence avec l'action a choisie a l'etape e.
   t : le thread. */
/* Renvoie 1 si il y a eu des eliminations. */
{
  int *p, iana = 0;
#ifdef MULTI_THREADS_FILE
  S_CONS * TC = &(CST[t]);
#endif

  for (p = tAi[e+1]; p < tAi[e+1]+nAi[e+1]; p++)
    {
#ifdef ORDERED2SEQ_SETS
	    if ( ! IN(*p, Seq2Actions[a]))
#endif
#ifdef ORDERED2SEQ_ARRAYS
	    if ( ! Seq2Actions[a][*p])
#endif
	      { 
		ADD_FILE_ACTION(*p, e+1, theThreads[t], TC);
		iana = 1;
	      }
    }
  return iana;
}


int FilterPreviousStepUnorderedActions(int a, int e, int ns, int t)
/* Met dans la file de propagation les actions de l'etape e-1 qui n'ont pas 
   de support 2sequence avec l'action a choisie a l'etape e.
   t : le thread. */
/* Renvoie 1 si il y a eu des eliminations. */
{
  int *p, iana = 0;
#ifdef MULTI_THREADS_FILE
  S_CONS * TC = &(CST[t]);
#endif

  for (p = tAi[e-1]; p < tAi[e-1]+nAi[e-1]; p++)
    {
#ifdef ORDERED2SEQ_SETS
	  if ( ! IN(a, Seq2Actions[*p]))
#endif
#ifdef ORDERED2SEQ_ARRAYS
	    if ( ! Seq2Actions[*p][a])
#endif
	      { 
		ADD_FILE_ACTION(*p, e-1, theThreads[t], TC);
		iana = 1;
	      }
	  break;
    }
  return iana;
}


#endif




/* ********************************************************************* */
/*                     PROPRIETES DE 2-SEQUENCES                         */
/* ********************************************************************* */



void ConstructionSequences2Actions(void)
     /* Ici on s'interesse aux sequences de 2 actions apriori
	valides, i.e. telles
	que (1) c'est possible et (2) si les actions sont independantes 
	alors la sous sequence respecte l'ordre des actions (ceci afin 
	d'eviter les enumerations redondantes de sequences equivalentes). */
{
  int i, j;

#ifdef ORDERED2SEQ_ARRAYS
  Seq2Actions = malloc(nActions*sizeof(char *));
  assert(Seq2Actions != NULL);
  for (i = 0; i < nActions; i++)
    {
      Seq2Actions[i] = calloc(nActions, sizeof(char));
      assert(Seq2Actions[i] != NULL);
    }
#endif
#ifdef ORDERED2SEQ_SETS
  Seq2Actions = malloc(nActions*sizeof(SET));
  assert(Seq2Actions != NULL);
  if (ConstructPred2SeqSets) 
    {
      pred2SeqSet = malloc(nActions*sizeof(SET));
      assert(pred2SeqSet != NULL);
    }
  for (i = 0; i < nActions; i++)
    {
      Seq2Actions[i] = AllocSet(nActions);
      assert(Seq2Actions[i] != NULL);
      if (ConstructPred2SeqSets) 
	{
	  pred2SeqSet[i] = AllocSet(nActions);
	  assert(pred2SeqSet[i] != NULL);
	}
    }
#endif
  Next2Actions = malloc(nActions*sizeof(int *));
  assert(Next2Actions != NULL);
  nNext2Actions = calloc(nActions, sizeof(int));
  assert(nNext2Actions != NULL);
  Prev2Actions = malloc(nActions*sizeof(int *));
  assert(Prev2Actions != NULL);
  nPrev2Actions = calloc(nActions, sizeof(int));
  assert(nPrev2Actions != NULL);

  for (i = 0; i < nActions-1; i++)
    for (j = i+1; j < nActions; j++)
      {
	BonneSequence2Actions(i, j);
#ifdef ORDERED2SEQ_ARRAYS
	if (Seq2Actions[i][j])
	  nNext2Actions[i]++, nPrev2Actions[j]++;
	if (Seq2Actions[j][i])
	  nNext2Actions[j]++, nPrev2Actions[i]++;
#endif
#ifdef ORDERED2SEQ_SETS
	if (IN(j,Seq2Actions[i]))
	  nNext2Actions[i]++, nPrev2Actions[j]++;
	if (IN(i,Seq2Actions[j]))
	  nNext2Actions[j]++, nPrev2Actions[i]++;
#endif
      }
  for (i = 0; i < nActions; i++)
    {
      Next2Actions[i] = malloc((nNext2Actions[i]+1)*sizeof(int));
      assert(Next2Actions[i] != NULL);
      Prev2Actions[i] = malloc((nPrev2Actions[i]+1)*sizeof(int));
      assert(Prev2Actions[i] != NULL);
      nNext2Actions[i] = nPrev2Actions[i] = 0;
    }
  for (i = 0; i < nActions; i++)
    for (j = 0; j < nActions; j++)
#ifdef ORDERED2SEQ_ARRAYS
      if ((i != j) && (Seq2Actions[i][j]))
#endif
#ifdef ORDERED2SEQ_SETS
	if ((i != j) && (IN(j, Seq2Actions[i])))
#endif
	{
	  Next2Actions[i][nNext2Actions[i]++] = j;
	  Prev2Actions[j][nPrev2Actions[j]++] = i;
	}
  for (i = 0; i < nActions; i++)
    Next2Actions[i][nNext2Actions[i]] = Prev2Actions[i][nPrev2Actions[i]] =FNIL;
}


/* La version utilisee en 2005-2006, notamment pour IPC-5, avec 
   la correction de septembre 2006, la correction de janvier 2007, et 
   l'optimisation pour le parcours des listes de preconditions, ajouts, retraits. */
#define REMOVE_INCONSISTENT_SEQUENCES 

void BonneSequence2Actions(int i, int j)
/* La sequence <p,q> est bonne dans les cas suivants
   (1) sa permutation n'est pas consistante
   (2) p < q
   (3) un ajout de p est une precondition de q
*/
{
  ACTION p = LesActions+i;
  ACTION q = LesActions+j;
  int *l1, *l2, sij, sji;
  int hoe, /* tous leurs effets sont opposes*/ 
    psuppreq, qsupprep, /* l'une supprime une precondition de l'autre */
    h1oe, /* un effet oppose */
    paddpreq, qaddprep; /* ajout d'une precondition de l'autre. */

  l1 = p->add;
  l2 = q->sup;
  hoe = 1;
  h1oe = 0;
  while ((*l1 != FNIL) && (*l2 != FNIL))
    if (*l1 == *l2)
      { h1oe = 1; if (hoe == 0) break; l1 ++, l2 ++; }
    else 
      {
	hoe = 0;
	if (h1oe == 1) break;
	if  (*l1 < *l2)  l1 ++;  else l2 ++;
      }
  if (hoe && (*l1 == FNIL) && (*l2 == FNIL))
    { sji = sij = 0; goto FIN; }
#ifndef REMOVE_INCONSISTENT_SEQUENCES 
  /* SEPT_2006 */
  if (h1oe)
    { sji = sij = 1; goto FIN; }
#endif

  /* Calcul de psuppreq et paddpreq */
  /* Si p supprime une precondition de q on arrete la. */
  l1 = q->pre;
  l2 = p->sup;
  psuppreq = paddpreq = 0;
  while ((*l1 != FNIL) && (*l2 != FNIL))
    if (*l1 == *l2)
      { psuppreq = 1; break; }
    else if (*l1 == OPPOSE(*l2))
      { paddpreq = 1; l1 ++; l2 ++; }
    else if (*l1 < *l2)
      l1++;
    else
      l2++;

  /* Calcul de qsupprep et qaddprep. */
  /* Si q supprime une precondition de p on s'arette la. */
  l1 = p->pre;
  l2 = q->sup;
  qsupprep = qaddprep = 0;
  while ((*l1 != FNIL) && (*l2 != FNIL))
    if (*l1 == *l2)
      { qsupprep = 1; break; }
    else if (*l1 == OPPOSE(*l2))
      { qaddprep = 1; l1 ++; l2 ++; }
    else if (*l1 < *l2)
      l1++;
    else
      l2++;
  
  if (psuppreq)
    if (qsupprep)
      { sij = sji = 0; goto FIN; }
    else
      { sij = 0; sji = 1; goto FIN; }
  else if (qsupprep)
      { sij = 1; sji = 0; goto FIN; }
  
  /* Note I : ici aucune des deux ne supprime une precondition de l'autre. */

#ifdef REMOVE_INCONSISTENT_SEQUENCES 
  /* SEPT_2006 */
  if (h1oe)
    /* on peut le mettre ici car de toutes facons puisqu'on doit 
       considerer les deux sequences, si une action supprime une
       precondition de l'autre action la sequence correspondante
       ne sera pas consistante. */
    { sji = sij = 1; goto FIN; }
#endif

  /* Note II : ici p et q n'ont aucun effet oppose. */

  /* JANVIER_2007 */
  if (qaddprep)
    {
      sji = 1;
      if (paddpreq)
	sij = 1;
      else
	sij = 0;
    }
  else if (paddpreq)
    {
      sij = 1;
      sji = 0;
    }
  
  /* si sji==1 alors 
       si sij==0 alors 
          interdire ij (ce qui est deja fait ! )
	  idem dans l'autre sens. */
  /* En effet si q ajoute une precondition a p il faut autoriser q,p
     Si de plus p et q n'ont pas d'effets contradictoires, et q ne 
     supprime pas de precondition de p (ce qui rendrait inconsistante
     la sequence q,p) et si p 'najoute pas de precondition a q, alors 
     on peut interdire p,q. */

  else
    if (((i < j) && (Mode2SeqIPC5Croissante)) ||
	((i > j) && ( ! Mode2SeqIPC5Croissante)))
      sij = 1, sji = 0;
    else
      sji = 1, sij = 0;
  
 FIN:
#ifdef ORDERED2SEQ_ARRAYS
  if (sij == 0)
    Seq2Actions[i][j] = 0;
  else
    Seq2Actions[i][j] = 1;
  if (sji == 0) 
    Seq2Actions[j][i] = 0;
  else
    Seq2Actions[j][i] = 1;
#endif
#ifdef ORDERED2SEQ_SETS
  if (sij != 0)
    {
      ADDe(j, Seq2Actions[i]);
      if (ConstructPred2SeqSets)
	ADDe(i, pred2SeqSet[j]);
    }
  if (sji != 0)
    {
      ADDe(i, Seq2Actions[j]);
      if (ConstructPred2SeqSets)
	ADDe(j, pred2SeqSet[i]);
    }
#endif
}



