#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>

#include "consistance.h"
#include "qsort.h"
#include "hashstrings.h"
#include "main.h"
#include "suppressions.h"
#include "structures.h"
#include "constantes.h"
#include "minsteps.h"
#include "sets.h"
#include "utils.h"
#include "search.h"
#include "sequences.h"
#include "threads.h"

/* VARIABLES. */

struct action *LesActions;
hs_table actions_htable; /* hashcodes des noms d'actions. */
int nActions;
int maxActCost; /* cout maximal pour les actions. */
int minActCost; /* cout maximal pour les actions. */
int nActionsIgnorees;
int nActionsUnknownDel; /* nombre d'actions ignors a cause */
                        /* suppression inexistante */
int SizeAllocActions;
int maxGAdds = 0;       /* nombre max de buts ajoutes par une action
			   (cf. ConstructionDesTablesDeBooleensDeLAction). */


/* Comme chaque fait a un oppose, on suppose que LesFaits[2*i] et 
   LesFaits[2*i+1] sont respectivement le fait i et son oppose. */
struct fait *LesFaits;
hs_table facts_htable; /* hash codes des noms de faits. Si elt dans 
			  facts_htable alors 2*elt et 2*elt+1. */
int nFaits; /* Nombre de faits (sans les opposes). */
int SizeAllocFaits;


int *InitState, *Goals; /* Tables de presence. */
int *TheGoals;
int nGoals;

int cadd = 0; /* Nombre d'ajouts toutes actions confondues.
		 Utile pour l'allocation de la pile des suppressions. */
int cadel = 0; /* Cumul des nombres d'actions qui suppriment chaque fait. */


/* Graphe d'actions. */

int nE; /* Nombre d'etapes (F0,...,F{nE}),(A0,...,A{nE-1}).. */
int **Fi; /* Table de presence des faits indicee sur l'etape et sur le numero 
	     du fait dans le tableau LesFaits. */
int *nFi; /* Leurs nombres a chaque etape. */
int **Ai; /* Table de presence des actions indicee sur l'tape. */
int *nAi; /* Leurs nombres a chaque etape. */
int **nbAP, **nbAM, **nbAR; /* Compteurs nbA+ et nbA- et nbAR des nombres 
			       d'actions qui ajoutent ou suppriment le fait i 
			       a l'etape s-1 ou qui le requierent a l'etape s. */
int nAiSingles; /* Le nombre de Ai de cardinal 1. */
int **tAi; /* Table des actions a chaque etape. */
int **itAi; /* Les indices des actions dans tAi. */

int *StepLevel; /* Pour les plans paralleles, numero de l'etape pour chaque 
		   niveau du graphe d'actions. */

int SizeGraph; /* Taille d'allocation du graphe. */


/* GESTION DES FAITS PERSISTANTS. */
struct fact_maillon **P; 

struct fact_maillon **TheM; 



int **EstSousBut; /* Marque des sous-buts : est sous-but un fait qui 
		     est requis par au moins une action ou qui apparait 
		     a l'etape suivante comme sous-but. */
int **NbAjoutsSousButs; /* Nombre de sous-buts ajoutes par l'action : 
			   si ca vaut 0 on supprime l'action. */
int *Maturite; /* Derniere etape ou le fait est un but. */
int *nButs; /* Nombre de buts (Goals) a chaque etape. */
int *nFixedGoals;
int AllGoals; /* Premiere etape a laquelle on a tous les buts. */


/* Piles pour ordonner les choix (actions ou faits). */
int **ChoixEnumActions;
int **EvalEnumActions;
int *ChoixEnumFacts;
float *EvalEnumFacts;



S_GR * GRT; /* Table des graphes d'actions. */

void AllocGrapheDActionThreads (int ns, int thread)
{
  int i;
  S_GR * G = &(GRT[thread]);
  int na = theThreads[thread].nActions;
  int nf = theThreads[thread].nFaits;

  Fi = G->Fi = malloc((ns+1)*sizeof(int *));
  assert(G->Fi != NULL);
  nFi = G->nFi = malloc((ns+1)*sizeof(int ));
  assert(G->nFi != NULL);
  Ai = G->Ai = malloc((ns)*sizeof(int *));
  assert(G->Ai != NULL);
  tAi = G->tAi = malloc((ns)*sizeof(int *));
  assert(G->tAi != NULL);
  itAi = G->itAi = malloc((ns)*sizeof(int *));
  assert(G->itAi != NULL);
  nAi = G->nAi = malloc((ns)*sizeof(int ));
  assert(G->nAi != NULL);
  nbAP = G->nbAP = malloc((ns+1)*sizeof(int *));
  assert(G->nbAP != NULL);
  nbAM = G->nbAM = malloc((ns+1)*sizeof(int *));
  assert(G->nbAM != NULL);
  nbAR = G->nbAR = malloc((ns+1)*sizeof(int *));
  assert(G->nbAR != NULL);
  P = G->P = malloc((ns+1)*sizeof(struct fact_maillon *));
  assert(G->P != NULL);
  TheM = G->TheM = malloc((ns+1)*sizeof(struct fact_maillon *));
  assert(G->TheM != NULL);


  EstSousBut = G->EstSousBut = malloc((ns+1)*sizeof(int *));
  assert(G->EstSousBut != NULL);
  NbAjoutsSousButs = G->NbAjoutsSousButs = malloc((ns+1)*sizeof(int *));
  assert(G->NbAjoutsSousButs != NULL);
  Maturite = G->Maturite = malloc((2*nf+1)*sizeof(int *));
  assert(G->Maturite != NULL);
  nButs = G->nButs = malloc((ns+1)*sizeof(int));
  assert(G->nButs != NULL);
  nFixedGoals = G->nFixedGoals = malloc((ns+1)*sizeof(int));
  assert(G->nFixedGoals != NULL);

  /************************************************************** */

/*   G->sS4S = (1+(int) logf((float) na)) * (ns+1) + 1; */
  G->sS4S = (1+na) * (ns+1) + 1;
  G->tS4S = G->S4S = calloc(G->sS4S, sizeof(int **));
  assert(G->S4S != NULL);

  if (G->Sequences2Actions)
    {
      NbSucc2Seq = G->NbSucc2Seq = malloc((ns+1)*sizeof(int *));
      assert(G->NbSucc2Seq != NULL);
      NbPre2Seq = G->NbPre2Seq = malloc((ns+1)*sizeof(int *));
      assert(G->NbPre2Seq != NULL);
    }

  G->CurrentState = malloc(((nf+31)/32)*sizeof(unsigned int));
  assert(G->CurrentState != NULL);
  
  G->LesEtapes = malloc(ns*sizeof(int));
  assert(G->LesEtapes != NULL);
  ChoixEnumActions = G->ChoixEnumActions = malloc(ns*sizeof(int *));
  assert(G->ChoixEnumActions != NULL);
  G->EvalEnumActions = malloc(ns*sizeof(int *));
  assert(G->EvalEnumActions != NULL);
  G->ChoixEnumFacts = malloc((2*nf+1)*sizeof(int));
  assert(G->ChoixEnumFacts != NULL);
  G->EvalEnumFacts = malloc((2*nf+1)*sizeof(float));
  assert(G->EvalEnumFacts != NULL);
  for (i = 0; i < ns; i++)
    {
      G->ChoixEnumActions[i] = malloc((na+1)*sizeof(int));
      assert(G->ChoixEnumActions[i] != NULL);
      G->EvalEnumActions[i] = malloc((na+1)*sizeof(int *));
      assert(G->EvalEnumActions[i] != NULL);
      G->ChoixEnumActions[i][0] = FNIL;
    }


  for (i = 0; i < ns+1; i++) 
    AllocGraphFactLevelThreads(na, nf, i, thread);
  for (i = 0; i < ns; i++) 
    AllocGraphActionLevelThreads(na, i, thread);

  G->SizeGraph = ns;
}

void AllocGraphActionLevelThreads(int na, int i, int thread)
/* 0<=i<ns */
{
  S_GR * G = &(GRT[thread]);

  G->Ai[i] = malloc(na*sizeof(int));
  assert(G->Ai[i] != NULL);
  G->tAi[i] = malloc(na*sizeof(int));
  assert(G->tAi[i] != NULL);
  G->itAi[i] = malloc(na*sizeof(int));
  assert(G->itAi[i] != NULL);
  G->NbAjoutsSousButs[i] = malloc(na*sizeof(int));
  assert(G->NbAjoutsSousButs[i] != NULL);
  if (G->Sequences2Actions)
    {
      /* On alloue sans reflechir, meme si i vaut 0 ou ns. */
      G->NbSucc2Seq[i] = malloc(na*sizeof(int ));
      assert(G->NbSucc2Seq[i] != NULL);
      G->NbPre2Seq[i] = malloc(na*sizeof(int ));
      assert(G->NbPre2Seq[i] != NULL);
    }
}

void AllocGraphFactLevelThreads(int na, int nf, int i, int thread)
/* 0<=i<ns+1 */
/* On ne s'interesse pas au fait que i peut valoir 0 ou le nombre total 
   d'etapes. De toutes facons dans ce dernier cas il faudra tout allouer si on 
   incremente le nombre d'etapes. */
{
  S_GR * G = &(GRT[thread]);

  G->Fi[i] = malloc(2*nf*sizeof(int));
  assert(G->Fi[i] != NULL);

  G->nbAR[i] = calloc(2*nf,sizeof(int));
  assert(G->nbAR[i] != NULL);
  G->nbAP[i] = calloc(2*nf,sizeof(int));
  assert(G->nbAP[i] != NULL);
  G->nbAM[i] = calloc(2*nf,sizeof(int));
  assert(G->nbAM[i] != NULL);


  G->EstSousBut[i] = malloc(2*nf*sizeof(int));
  assert(G->EstSousBut[i] != NULL);

  G->P[i] = malloc((na+1)*sizeof(struct fact_maillon));
  assert(G->P[i] != NULL);
  G->TheM[i] = malloc(2*nf*sizeof(struct fact_maillon));
  assert(G->TheM[i] != NULL);
}

void AllocTheGoals(int nf)
{
  int i;

  TheGoals = malloc((nf+1) * sizeof(int));
  assert(TheGoals != NULL);
  for (i = nGoals = 0; i < nf; i++)
    if ( ! Goals[2*i])
      TheGoals[ nGoals ++ ] = 2*i+1;
    else if ( ! Goals[2*i+1])
      TheGoals[ nGoals ++ ] = 2*i;
  TheGoals[nGoals] = FNIL;
}


int removeFactList(int s, int *I, int t)
/* t : le thread */
{
  int i;
  THREAD T = theThreads+t;
#ifdef MULTI_THREADS_FILE
  S_CONS * TC = &(CST[t]);
#endif

  for (i = 0; i < 2*T->nFaits; i++)
    if (I[i] == 0)
      {
	if (GRT[t].Fi[s][i])
	  {
	    ADD_FILE_FACT(i, s, (*T), TC);
	  }
      }

  return 1;
}


int countFixedGoals (int s, int * goal, int t)
{
  int i, n;
  THREAD T = theThreads+t;
  S_GR * G = &(GRT[t]);

  for (i = n = 0; i < 2*T->nFaits; i++)
    if ( ! goal[OPPOSE(i)] )
      if ( ! G->Fi[s][OPPOSE(i)] )
	n ++;
  return n;
}

void fixGoals (int ns, int *G, int t)
{
  int i;
  THREAD T = theThreads+t;
#ifdef MULTI_THREADS_FILE
  S_CONS * TC = &(CST[t]);
#endif

  for (i = 0; i < 2*T->nFaits; i++)
    if (G[i] == 0)
      ADD_FILE_FACT(i, ns, (*T), TC);
}

void InitGrapheDActionsThreads(int ns, int *I, int *G, int t)
/* t : le thread */
{
  int i;
  THREAD T = theThreads+t;
#ifdef MULTI_THREADS_FILE
  S_CONS * TC = &(CST[t]);
#endif
  S_GR * GA = &(GRT[t]);

  GA->nAiSingles = nAiSingles = 0;
  for (i = 0; i <= ns; i++)
    InitGenericFactStateThreads(i, T->nActions, T->nFaits, ns, t);
  for (i = 0; i < ns; i++)
    InitGenericActionStateThreads(i, ns, t);

  /* Nombres de buts presents a chaque etape. */
  for (i = 0; i < ns+1; i++)
    {
      GA->nButs[i] = nGoals;
      GA->nFixedGoals[i] = 0;
    }
  GA->AllGoals = AllGoals = 1;

  InitFilesConsistanceThreads(t);

  if (I != NULL)
    for (i = 0; i < 2*T->nFaits; i++)
      if (I[i] == 0)
	ADD_FILE_FACT(i, 0, (*T), TC);

  if (G != NULL)
    for (i = 0; i < 2*T->nFaits; i++)
      if (G[i] == 0)
	ADD_FILE_FACT(i, ns, (*T), TC);

  InitPilesSuppressionsThread(t);
}

void InitGenericFactStateThreads(int s, int na, int nf, int ns, int thread)
{
  int i;
  S_GR * G = &(GRT[thread]);
  
  /* On met tous les faits dans Fs. */
  for (i = 0; i < nf; i++)
    G->Fi[s][2*i] = G->Fi[s][2*i+1] = 1;
  G->nFi[s] = 2*nf;
  /* Initialisation des listes de P(s) a vide. */
  if (s > 0)
    for (i = 0; i <= na; i++)
      (G->P[s][i]).suiv = (G->P[s][i]).prec = &(G->P[s][i]);
  /* Initialisation (pas tres utile car dans TheM l'indice donne le rang du 
     fait correspondant) des maillons de TheM. */
  if (s > 0)
    for (i = 0; i < nf; i++)
      {
	(G->TheM[s][2*i]).fait = 2*i;
	(G->TheM[s][2*i]).suiv = NULL;
	(G->TheM[s][2*i+1]).fait = 2*i+1;
	(G->TheM[s][2*i+1]).suiv = NULL;
      }
  /* Initialisation des compteurs nbA+ et nbA-, et eventuellement 
     si nbA+(f)=0 on met le fait dans la pile P(s). */
  if (s > 0)
      for (i = 0; i < nf; i++)
	{
	  /* NE PAS CHANGER L'ORDRE DES INSTRUCTIONS SUIVANTES (ADD_FLIST utilise nbAM[s][f]). */
	  G->nbAM[s][2*i] = LesFaits[2*i].nsup;
	  if ((G->nbAP[s][2*i] = LesFaits[2*i].nadd) == 0)
	    ADD_FLIST(2*i, s, G);
	  G->nbAP[s][2*i] = LesFaits[2*i].nadd;
	  G->nbAM[s][2*i+1] = LesFaits[2*i+1].nsup;
	  if ((G->nbAP[s][2*i+1] = LesFaits[2*i+1].nadd) == 0)
	    ADD_FLIST(2*i+1, s, G);
	  G->nbAP[s][2*i+1] = LesFaits[2*i+1].nadd;
	}
  /* Initialisation des compteurs nbAR. */
  if (s < ns+1)
    for (i = 0; i < nf; i++)
      {
	G->nbAR[s][2*i] = LesFaits[2*i].npre;
	G->nbAR[s][2*i+1] = LesFaits[2*i+1].npre;
      }

  /* Initialisation des sous-buts : tous les faits sont sous-buts 
     initialement.*/
  if (s < ns+1)
    for (i = 0; i < nf; i++)
      {
	G->EstSousBut[s][2*i] = 1;
	G->EstSousBut[s][2*i+1] = 1;
      }
}

void InitGenericActionStateThreads(int s, int ns, int thread)
     /* Initialise As, en y mettant toutes les actions. */
     /* O(|A|) */
{
  int i, na;
  S_GR * G = &(GRT[thread]);

  na = theThreads[thread].nActions;

  for (i = 0; i < na; i++)
    {
      G->Ai[s][i] = 1;
      G->tAi[s][i] = i;
      G->itAi[s][i] = i;
    }
  G->nAi[s] = na;
  if (G->nAi[s] == 1)
    {
      G->nAiSingles ++;
      nAiSingles ++;
    }
  if (G->Sequences2Actions)
    {
      if (s > 0)
	for (i = 0; i < na; i++)
	  G->NbPre2Seq[s][i] = nPrev2Actions[i];
      if (s < ns)
	for (i = 0; i < na; i++)
	  G->NbSucc2Seq[s][i] = nNext2Actions[i];
    }
}


/* calcul du cout du plan courant (suppose qu'on a un plan). */
int evaluatePlanCost (int ns, int thread)
{
  S_GR * G = &(GRT[thread]);
  int i, c;

  for (i = c = 0; i < ns; i ++)
    c += LesActions[ G->tAi[i][0] ].cost;
  return (c);
}




/* Fonctions de creations des actions et des faits a la lecture. */
void AllocTablesActionsFaits(void)
{
  nFaits = nActions = nActionsUnknownDel = nActionsIgnorees = 0;
  SizeAllocFaits = 2000;
  SizeAllocActions = 1000;
  SizeGraph = 10;
  LesFaits = malloc(SizeAllocFaits*sizeof(struct fait));
  assert(LesFaits != NULL);

  Maturite = malloc(SizeAllocFaits*sizeof(int));
  assert(Maturite != NULL);

  LesActions = malloc(SizeAllocActions*sizeof(struct action));
  assert(LesActions != NULL);
  actions_htable = alloc_hs_table();
  facts_htable = alloc_hs_table();
}





/* ************************************************************ */
/*                                                              */
/*        REPRESENTATION DU PROBLEME : ACTIONS, FAITS,...       */
/*                                                              */
/* ************************************************************ */

void MaJFactsActionsLists(int a)
{
  /* Mise a jour des tables d'actions des faits apres l'ajout d'une 
     nouvelle action a. */
  /* IMPORTANT : en principe, vu que les actions sont ajoutees dans l'ordre 
     de leurs numeros, l'ordre est respecte dans les listes. */
  FAIT p;
  int *q;

  for (q = (LesActions+a)->pre; *q != FNIL; q++)
    {
      p = LesFaits+*q;
      p->npre ++;
      p->pre = realloc(p->pre, (p->npre+1)*sizeof(int));
      p->pre[p->npre-1] = a;
      p->pre[p->npre] = FNIL;
    }
  for (q = (LesActions+a)->add; *q != FNIL; q++)
    {
      p = LesFaits+*q;
      p->nadd ++;
      cadd ++;
      p->add = realloc(p->add, (p->nadd+1)*sizeof(int));
      p->add[p->nadd-1] = a;
      p->add[p->nadd] = FNIL;
    }
  for (q = (LesActions+a)->sup; *q != FNIL; q++)
    {
      p = LesFaits+*q;
      p->nsup ++;
      cadel ++;
      p->sup = realloc(p->sup, (p->nsup+1)*sizeof(int));
      p->sup[p->nsup-1] = a;
      p->sup[p->nsup] = FNIL;
    }
}

void ConstructionDesTablesDeBooleensDeLAction(ACTION a, int nf)
/* Allocation et construction des tables de booleens des actions, indiquant 
   les faits qui sont prerequis, ajouts ou suppressions de l'action a. */
{
  int *l;
  int nga;
  
  a->isadd = calloc(2*nf, sizeof(char));
  assert(a->isadd != NULL);
  a->issup = calloc(2*nf, sizeof(char));
  assert(a->issup != NULL);
  a->ispre = calloc(2*nf, sizeof(char));
  assert(a->ispre != NULL);
  for (l = a->pre; *l != FNIL; l++)
    a->ispre[*l] = 1;
  for (l = a->add, nga = 0; *l != FNIL; l++)
    {
      if ( ! Goals [OPPOSE(*l)])
	nga ++;
      a->isadd[*l] = 1;
    }
  for (l = a->sup; *l != FNIL; l++)
    a->issup[*l] = 1;

  if (nga > maxGAdds)
    maxGAdds = nga;
}

void ConstructionTablesNewActions(int first, int last, int nf)
/* Construit les tables de booleens des actions pour les actions 
   LesActions[first],...,LesActions[last] */
{
  ACTION a;

  for (a = LesActions+first; a <= LesActions+last; a++)
    ConstructionDesTablesDeBooleensDeLAction(a, nf);
}


#define KEEP_ADD_EFFECT

#ifndef KEEP_ADD_EFFECT
int AddNewAction(char *s, FLIST P, FLIST A, FLIST S, int cost, int keepActionsWithOppositeEffects)
/* Ici on suppose qu'il n'y a que des faits positifs dans A et S. */
{
  ACTION a;
  int *t_add, *t_sup, *t_pre, nadd, nsup, npre;

  if (search_action_hs_table(s, actions_htable, LesActions) != NOT_IN_HS)
    return ACTION_PRESENTE;
  
  t_add = AddFactsFromLists(A, S, &nadd);
  t_sup = AddFactsFromLists(S, A, &nsup);
  t_pre = AddFactsFromList(P, &npre);

  if (npre-1 >= 0) QSort(t_pre, 0, npre-1);
  npre = ListesFiltreRedundancies(t_pre);

  if (nadd-1 >= 0) QSort(t_add, 0, nadd-1);
  nadd = ListesFiltreRedundancies(t_add);
  if ( keepActionsWithOppositeEffects)
    nadd = removeSimultaneousOppositeEffects(t_add);

  if (nsup-1 >= 0) QSort(t_sup, 0, nsup-1);
  nsup = ListesFiltreRedundancies(t_sup);
  if ( keepActionsWithOppositeEffects)
    {
    nsup = removeSimultaneousOppositeEffects(t_sup);
    }

  if (( ! keepActionsWithOppositeEffects) && 
      ( ! ListesDisjointes(t_add, t_sup)))
    {
      free(t_add), free(t_sup), free(t_pre);
      RestitueFList(P);
      RestitueFList(A);
      RestitueFList(S);
      nActionsIgnorees++;
      return ACTION_INCONSISTANTE;
    }

  if (nActions == SizeAllocActions)
    {
      SizeAllocActions += 100;
      LesActions = realloc(LesActions, SizeAllocActions*sizeof(struct action));
      assert(LesActions != NULL);
    }

  a = LesActions+nActions;  
  a->add = t_add;
  a->nadd = nadd;
  a->sup = t_sup;
  a->nsup = nsup;
  a->pre = t_pre;
  a->npre = npre;

  if ( ! randomCosts)
    {
      a->cost = cost;
      if (0 && cost <= 0)
	printf("%s  cost = %d\n", s, cost);
    }
  else
    a->cost = rand()%7+1;

  if (strcmp (s, "noop") != 0)
    {
      if ((a->cost > maxActCost) || (maxActCost == NONE))
	maxActCost = a->cost;
      if ((a->cost < minActCost) || (minActCost == NONE))
	minActCost = a->cost;
    }

  a->nom = malloc((strlen(s)+1)*sizeof(char));
  assert(a->nom != NULL);
  strcpy(a->nom, s);
  a->hcode = hs_hash(s);
  add_hs_table(s, actions_htable, nActions);
  MaJFactsActionsLists(nActions);

  RestitueFList(P);
  RestitueFList(A);
  RestitueFList(S);
  nActions ++;
  return (nActions-1);
}
#endif


#ifdef KEEP_ADD_EFFECT

int * AddEffectsFromLists(FLIST lp, FLIST lm, FLIST lr, int *n);
int * makeDeletionsList (int *A, int n, int * size);

int AddNewAction(char *s, FLIST P, FLIST A, FLIST S, int cost, 
		 int keepActionsWithOppositeEffects)
/* Ici on suppose qu'il n'y a que des faits positifs dans A et S. */
{
  ACTION a;
  int *t_add, *t_sup, *t_pre, nadd, nsup, npre;

  if (search_action_hs_table(s, actions_htable, LesActions) != NOT_IN_HS)
    return ACTION_PRESENTE;
  
  t_add = AddEffectsFromLists(A, S, P, &nadd);
  t_sup = makeDeletionsList (t_add, nadd, &nsup);

  t_pre = AddFactsFromList(P, &npre);

  if (npre > 1) 
    {
      QSort(t_pre, 0, npre-1);
      npre = ListesFiltreRedundancies(t_pre);
    }

  if (nadd > 1) 
    {
      QSort(t_add, 0, nadd-1);
      nadd = ListesFiltreRedundancies(t_add);
      QSort(t_sup, 0, nsup-1);
      nsup = ListesFiltreRedundancies(t_sup);
    }

  if (nActions == SizeAllocActions)
    {
      SizeAllocActions += 100;
      LesActions = realloc(LesActions, SizeAllocActions*sizeof(struct action));
      assert(LesActions != NULL);
    }

  a = LesActions+nActions;  
  a->add = t_add;
  a->nadd = nadd;
  a->sup = t_sup;
  a->nsup = nsup;
  a->pre = t_pre;
  a->npre = npre;

  if ( ! randomCosts)
    a->cost = cost;
  else
    a->cost = rand()%7+1;

  if (strcmp (s, "noop") != 0)
    {
      if ((a->cost > maxActCost) || (maxActCost == NONE))
	maxActCost = a->cost;
      if ((a->cost < minActCost) || (minActCost == NONE))
	minActCost = a->cost;
    }

  a->nom = malloc((strlen(s)+1)*sizeof(char));
  assert(a->nom != NULL);
  strcpy(a->nom, s);
  a->hcode = hs_hash(s);
  add_hs_table(s, actions_htable, nActions);
  MaJFactsActionsLists(nActions);

  RestitueFList(P);
  RestitueFList(A);
  RestitueFList(S);
  nActions ++;
  return (nActions-1);
}


int * AddEffectsFromLists(FLIST lp, FLIST lm, FLIST lr, int *nb)
{
  /* COnstruit la table tt avec les faits de la liste lp et les opposes des 
     faits de la liste lm. Laisse comme effet les faits de lp meme s'ils apparaissent 
     dans lm. */
  int *t, *tt, *q, s;
  FLIST p, r;

  s = 0;
  for (p = lp; p != NULL; p = p->suiv)
    s ++;

  for (p = lm; p != NULL; p = p->suiv)
    s ++;

  tt = malloc((s+1) * sizeof(int));
  assert(tt != NULL);

  for (t = tt, p = lp; p != NULL; p = p->suiv)
    *t ++ = p->fait;

  for (p = lm; p != NULL; p = p->suiv)
    {
      /* on ne conserve la suppression que si son oppose n'est pas ajoute */
      for (q = tt; q != t; q ++)
	if (p->fait == *q)
	  break;
      if (q == t)
	*t ++ = OPPOSE(p->fait);
    }

  *t = FNIL;
  *nb = t-tt;
  return tt;
}


int * makeDeletionsList (int * A, int n, int * size)
{
  int * D, i;

  D = malloc((n+1) * sizeof(int));
  assert(D != NULL);

  for (i = 0; i < n; i ++)
    D[i] = OPPOSE(A[i]);

  D[n] = FNIL;

  *size = n;
  return D;
}

#endif

int nFaitsCommuns(int *s, int *p)
/* Calcule le nombre de faits communs aux deux listes (triees). */
{
  int n = 0;
  while ((*s != FNIL) && (*p != FNIL))
    if (*s == *p)
      {
	n++;
	s++, p++;
      }
    else if (*s < *p)
      s++;
    else
      p++;
  return n;
}

int ListesDisjointes(int *a, int *s)
/* Verifie que a et s sont disjointes (elles sont supposees triees). */
{
  while ((*a != FNIL) && (*s != FNIL))
    if (*a == *s)
      return 0;
    else if (*a < *s)
      a++;
    else
      s++;
  return 1;
}

int ListeIncluse(int *a, int *s)
/* Verifie que a est inclue dans s (elles sont supposees triees). */
{
  while (*a != FNIL)
    {
      if ((*s > *a) || (*s == FNIL))
	return 0;
      if (*a == *s)
	a++;
      s++;
    }
  return 1;
}

int ListesFiltreRedundancies(int *a)
/* Verifie que le meme element n'apparait pas plusieurs fois. */
{
  int *p = a, *tmp = a;

  if (*a == FNIL)
    return 0;

  while (*a != FNIL)
    {
      if (*p != *a)
	* ++p = *a;
      a ++;
    }
  * ++p = FNIL;

  return (p-tmp);
}

int removeSimultaneousOppositeEffects(int *a)
{
  int *p, *first = a;

  if (*a == FNIL)
    return 0;

  p = a+1;
  while (*(p-1) != FNIL)
    if ((*p == FNIL) || (*(p-1) != OPPOSE(*p)))
      {
	*a ++ = *(p-1);
	p ++;
      }
    else
      {
	if (0) printf("opposite effects  %d %d\n", *(p-1), *p);
/* 	*a ++ = *p; */
	p += 2;
      }

  *a = FNIL;
  return (a-first);
}

int * AddFactsFromBothLists(FLIST lp, FLIST lm, int *n)
{
  /* COnstruit la table tt avec les faits des listes lp et lm. */
  int *t, *tt;
  FLIST p;

  if (lp != NULL)
    for (*n = 1, p = lp->suiv; ((p != NULL) && (p != lp)); p = p->suiv)
      (*n)++;
  else
    *n = 0;
  if (lm != NULL)
    for (*n = (*n)+1, p = lm->suiv; ((p != NULL) && (p != lm)); p = p->suiv)
      (*n)++;
  tt = t = malloc((*n+1)*sizeof(int));
  assert(t != NULL);
  if (lp != NULL)
    for (p = lp->suiv, *t++ = lp->fait; ((p != NULL) && (p != lp)); p = p->suiv)
      *t++ = p->fait;
  if (lm != NULL)
    for (p = lm->suiv, *t++ = lm->fait; ((p != NULL) && (p != lm)); p = p->suiv)
      *t++ = p->fait;
  *t = FNIL;
  return tt;
}

int * AddFactsFromLists(FLIST lp, FLIST lm, int *n)
{
  /* COnstruit la table tt avec les faits de la liste lp et les opposes des 
     faits de la liste lm. Chacune des listes peut etre circulaire. */
  int *t, *tt;
  FLIST p;

  *n = 0;
  if (lp != NULL)
    for (*n = (*n)+1, p = lp->suiv; ((p != NULL) && (p != lp)); p = p->suiv)
      (*n) ++;

  if (lm != NULL)
    for (*n = (*n)+1, p = lm->suiv; ((p != NULL) && (p != lm)); p = p->suiv)
      (*n) ++;

  tt = t = malloc((*n+1) * sizeof(int));
  assert(t != NULL);
  if (lp != NULL)
    for (p = lp->suiv, *t++ = lp->fait; ((p != NULL) && (p != lp)); p = p->suiv)
      *t ++ = p->fait;
  if (lm != NULL)
    for (p = lm->suiv, *t++ = OPPOSE(lm->fait); ((p != NULL) && (p != lm)); p = p->suiv)
      *t ++ = OPPOSE(p->fait);
  *t = FNIL;
  return tt;
}

int * AddFactsFromList(FLIST l, int *n)
{
  /* Stocke les faits de la liste l dans la table t apres l'avoir allouee.
     Ajoute FNIL a la fin de t. Attention : la liste l peut etre circulaire. */
  int *t, *tt;
  FLIST p;

  *n = 0;
  if (l != NULL)
    for (*n = 1, p = l->suiv; ((p != NULL) && (p != l)); p = p->suiv)
      (*n)++;

  tt = t = malloc((*n+1)*sizeof(int));
  assert(t != NULL);

  if (l != NULL)
    for (p = l->suiv, *t++ = l->fait; ((p != NULL) && (p != l)); p = p->suiv)
      *t++ = p->fait;
  *t = FNIL;
  return tt;
}

int SearchFact(char *s, int signe)
/* Ici on renvoie juste le numero du fait.
   'signe' vaut 1 pour le fait, et 0 pour son oppose. */
{
  int i;

  if ((i = search_fact_hs_table(s, facts_htable, LesFaits)) != NOT_IN_HS)
    return (2*i+1-signe);
  return NONE;
  return (AddFact(s, signe, 0));
}

int AddFact(char *s, int signe, int but)
/* Ajout d'un nouveau fait. */
{
  FAIT p;
  FAIT q;

  if (2*nFaits == SizeAllocFaits)
    {
      SizeAllocFaits += 200;
      LesFaits = realloc(LesFaits, SizeAllocFaits*sizeof(struct fait));
      assert(LesFaits != NULL);

      Maturite = realloc(Maturite, SizeAllocFaits*sizeof(int));
      assert(Maturite != NULL);
    }
  p = LesFaits+2*nFaits;
  q = LesFaits+2*nFaits+1;
  q->nom = p->nom = malloc((strlen(s)+1)*sizeof(char));
  assert(p->nom != NULL);
  strcpy(p->nom, s);
  p->npre = p->nadd = p->nsup = 0;
  q->npre = q->nadd = q->nsup = 0;
  {
    p->pre = malloc(sizeof(int));assert(p->pre != NULL);p->pre[0] = FNIL;
    p->add = malloc(sizeof(int));assert(p->add != NULL);p->add[0] = FNIL;
    p->sup = malloc(sizeof(int));assert(p->sup != NULL);p->sup[0] = FNIL;
    q->pre = malloc(sizeof(int));assert(q->pre != NULL);q->pre[0] = FNIL;
    q->add = malloc(sizeof(int));assert(q->add != NULL);q->add[0] = FNIL;
    q->sup = malloc(sizeof(int));assert(q->sup != NULL);q->sup[0] = FNIL;
  }
  p->but = q->but = but;
  p->hcode = q->hcode = hs_hash(s);
  add_hs_table(s, facts_htable, nFaits);
  nFaits++;
  return (2*(nFaits-1)+1-signe);
}


/* ******************************************************* */
/*                  Fonctions d'affichage.                 */
/* ******************************************************* */

void AfficheLeFait(FAIT f)
{
  int *p;
  if ((f-LesFaits)%2 == 0)
    printf("+%s :\n   ", f->nom);
  else
    printf("-%s :\n   ", f->nom);
  for (p = f->pre; *p != FNIL; p++)
    printf("%s ", (LesActions+*p)->nom);
  printf("\n   ");
  for (p = f->add; *p != FNIL; p++)
    printf("%s ", (LesActions+*p)->nom);
  printf("\n   ");
  for (p = f->sup; *p != FNIL; p++)
    printf("%s ", (LesActions+*p)->nom);
  printf("\n");
}

void AfficheLAction(ACTION a)
{
  int *p;
  printf("%s :\n   P ", a->nom);
  for (p = a->pre; *p != FNIL; p++)
    printf("%c%s ", ((*p)%2==0)?'+':'-', (LesFaits+*p)->nom);
  printf("\n   A ");
  for (p = a->add; *p != FNIL; p++)
    printf("%c%s ", ((*p)%2==0)?'+':'-', (LesFaits+*p)->nom);
  printf("\n   S ");
  for (p = a->sup; *p != FNIL; p++)
    printf("%c%s ", ((*p)%2==0)?'+':'-', (LesFaits+*p)->nom);
  printf("\n");
}

void printAction(int a)
{
  int *p;
  printf("%s :\n   P ", (LesActions+a)->nom);
  for (p = (LesActions+a)->pre; *p != FNIL; p++)
    printf("%d ", *p);
  printf("\n   A ");
  for (p = (LesActions+a)->add; *p != FNIL; p++)
    printf("%d ", *p);
  printf("\n   S ");
  for (p = (LesActions+a)->sup; *p != FNIL; p++)
    printf("%d ", *p);
  printf("\n");
}

void AfficheLeProbleme(void)
{
  int i;

  printf("*** %d faits ***\n", nFaits);
  for (i = 0; i < nFaits; i++)
    {
      AfficheLeFait(LesFaits+(2*i));
      AfficheLeFait(LesFaits+(2*i+1));
    }
  printf("*** %d actions ***\n", nActions);
  for (i = 0; i < nActions; i++)
    AfficheLAction(LesActions+i);
  printf("*** Etat initial / Buts ***\n");
  printf("  I ");
  for (i = 0; i < nFaits; i++)
    {
      if (InitState[2*i])
	printf("+%s ", (LesFaits+(2*i))->nom);
      if (InitState[2*i+1])
	printf("-%s ", (LesFaits+(2*i+1))->nom);
    }
  printf("\n  G ");
  for (i = 0; i < nFaits; i++)
    {
      if (Goals[2*i])
	printf("+%s ", (LesFaits+(2*i))->nom);
      if (Goals[2*i+1])
	printf("-%s ", (LesFaits+(2*i+1))->nom);
    }
  printf("\n*** *** *** *** *** \n");
}

void AfficheLEtat(int e)
{
  int i;

  printf("[%d] ", e);
  for (i = 0; i < nFaits; i++)
    {
      if (Fi[e][2*i])
	printf("10");
      else
	printf("01");
      if (i%4 == 3)
	printf("-");
    }
  printf("\n");
}

void printFactsAtLevel (int e, int thread)
{
  S_GR * G = &(GRT[thread]);
  int i;

  printf("F%d =  ", e);
  for (i = 0; i < nFaits; i++)
    {
      if ( ! G->Fi[e][2*i])
	printf("%d ", 2*i+1);
      else if ( ! G->Fi[e][2*i+1])
	printf("%d ", 2*i);
    }
  printf("\n");
}




void AfficheLeGrapheDActions(int ne, int mode, int thread)
{
  int i, j;
  S_GR * G = &(GRT[thread]);
  THREAD T = theThreads+thread;

  for (i = 0; i <= ne; i ++)
    {
      printf("F%d (%d)  ", i, G->nFi[i]);
      for (j = 0; j < T->nFaits; j ++)
	if ((mode != 2) || (( ! G->Fi[i][2*j]) || ( ! G->Fi[i][2*j+1])))
	{
	  if (G->Fi[i][2*j])
	    printf("+%s ", LesFaits[2*j].nom);
	  if (mode)
	    if (G->Fi[i][2*j+1])
	      printf("-%s ", LesFaits[2*j].nom);
	}
      if (i < ne)
	{
	  printf("\nA%d (%d)  ", i, G->nAi[i]);
	  for (j = 0; j < T->nActions; j ++)
	    {
	      if (G->Ai[i][j])
		printf("%s ", LesActions[j].nom);
	    }
	}
      printf("\n");
    }
}


void AfficheSizesGraphe(int ne)
/* Affiche juste les degres de libertes de chaque Fi et Ai. */
{
  int i;
  char s[100];
  
  for (i = 0; i < ne; i++)
    {
      sprintf(s, "%d [%d] ", nFi[i]-nFaits, nAi[i]);
      printf("%8s", s);
    }
  printf("%3d\n", nFi[ne]-nFaits);
}

/* Utilitaires listes. */

FLIST CreerFLMaillon(int f, FLIST p, FLIST s)
{
  FLIST l;

  l = malloc(sizeof(struct fact_maillon));
  assert(l != NULL);
  l->fait = f;
  l->prec = p;
  l->suiv = s;
  return l;
}

void RestitueFList(FLIST l)
{
  FLIST p, q;
  
  if (l != NULL)
    {
      p = l->suiv;
      while ((p != l) && (p!=NULL))
	{
	  q = p->suiv;
	  free(p);
	  p = q;
	}
      free(l);
    }
}

int * FromListToBooleans(FLIST l, int n)
/* Alloue un tableau de presence pour les n elements. */
/* Marque presents les elements qui sont dans l. */
/* Exemple : InitState = FromListToBooleans(lInit, nFaits*2). */
{
  int *T;
  int i; 

  T = calloc(n, sizeof(int));
  assert(T != NULL);
  for(i = 1; i < n; i += 2)
    T[i] = 1;

  while (l != NULL)
  {
      if (l->fait < n) 
      {
            T[l->fait] = 1;
            T[OPPOSE(l->fait)] = 0;
      }
      l=l->suiv;
  }
  return T;
}


int * FromGoalsToBooleans(FLIST l, int n)
/* Alloue un tableau de presence pour les n elements. */
/* Marque presents les elements qui sont dans l. */
/* Exemple : InitState = FromListToBooleans(lInit, nFaits*2). */
{
  int *T;
  int i;
  T = calloc(n, sizeof(int));
  assert(T != NULL);
  for (i = 0; i < n; ++i) 
    T[i] = 1;

  while (l != NULL)
  {
    if (l->fait < n) 
      T[OPPOSE(l->fait)] = 0;
    l=l->suiv;
  }
  return T;
}

