
#ifndef _H_structures
#define _H_structures

#include "hash.h"
#include "sets.h"
#include "utils.h"

/* OPtions de compilation. */



/* Utilisation d'une pile pour les modifications des sous-buts lors du 
   traitement des preconditions des actions supprimees. */
#define STACK_CHANGES_PRECONDITIONS_PROCESSING
#define STACK_CHANGES_ADDFLIST


/* MACROS. */

#define ISSUPP 0
#define ISADD 1
#define ISPRE 2

/* Drapeaux d'ajouts d'actions et de faits. */
#define ACTION_INCONSISTANTE (-1)
#define ACTION_PRESENTE (-2)

/* STRUCTURES. */

typedef struct fait *FAIT;
typedef struct action *ACTION;

#include "hashstrings.h"

struct fait {
  char *nom;
  int hcode;
  int but; /* Marque signalant si le fait est un but mais aucune action ne 
	      l'ajoute pour l'instant, et il n'est pas dans l'etat initial. */
  int *pre, *add, *sup; /* Les action dont f est prerequis, ajout, suppr. */
  int npre, nadd, nsup; /* Les nombres d'actions qui requierent f,.. */
};

struct action {
  char *nom;
  int hcode;
  int *pre, *add, *sup; /* Les prerequis, ajouts et suppressions. */
  int npre, nadd, nsup;  /* Leurs nombres respectifs. */
  char *isadd, *issup, *ispre; /* Tables booleennes pour determiner si un  */
  /* fait est ajoute ou supprime par l'action. */
  int cost;
};

/* Listes des faits qui doivent persister. */
typedef struct fact_maillon *FLIST;

struct fact_maillon {
  int fait;
  FLIST prec, suiv;
};

/* VARIABLES. */

extern struct action *LesActions;
extern hs_table actions_htable;
extern int nActions;
extern int maxActCost;
extern int minActCost;
extern int SizeAllocActions;
extern int maxGAdds;
extern int nActionsIgnorees;   /* Modif Cyril */
extern int nActionsUnknownDel;   /* Modif Cyril */
extern struct fait *LesFaits;
extern hs_table facts_htable;
extern int nFaits;
extern int SizeAllocFaits;
extern int *InitState, *Goals;
extern int *TheGoals;
extern int nGoals;

extern int cadd;
extern int cadel;

extern int sizeStruct; 
extern int nE;
extern int **Fi; 
extern int *nFi; 
extern int **Ai;
extern int *nAi;
extern int **nbAP, **nbAM, **nbAR;
extern int nAiSingles;
extern int **tAi;
extern int **itAi;
extern int *StepLevel;
extern int SizeGraph;

extern struct fact_maillon **P; 
extern struct fact_maillon **TheM; 

extern struct fact_maillon **FNR;
extern struct fact_maillon **TheFNR;



extern int **EstSousBut;
extern int **NbAjoutsSousButs;
extern int *Maturite;
extern int *nButs;
extern int *nFixedGoals;

extern int AllGoals;


/* Piles pour ordonner les choix (actions ou faits). */
extern int **ChoixEnumActions;
extern int **EvalEnumActions;
extern int *ChoixEnumFacts;
extern float *EvalEnumFacts;



/* ********************************************************* */

typedef struct structure_graphe_dactions {
  int sizeStruct; 
  int nE;
  int **Fi; 
  int *nFi; 
  int **Ai;
  int *nAi;
  int **nbAP, **nbAM, **nbAR;
  int nAiSingles;
  int **tAi;
  int **itAi;
  int SizeGraph;
  int cost;

  struct fact_maillon **P;
  struct fact_maillon **TheM;

  int **EstSousBut;
  int **NbAjoutsSousButs;
  int *Maturite;
  int *nButs;
  int *nFixedGoals;
  int AllGoals;

  /* Ensembles d'actions lors du split. */
  int *** S4S; /* Sets for split. */
  int *** tS4S; /* top. */
  int sS4S;    /* La taille de la table allouee. */

  /* 2-sequences dans la propagation. */
  int Sequences2Actions;
  int **NbSucc2Seq, **NbPre2Seq; /* Pour Sequences2Actions (pas utilise). */

  /* Tables de hashage. */
  unsigned int *CurrentState;
  hash_list CurrentHash;

  /* Piles pour ordonner les choix (actions ou faits). */
  int OrdActionsEnum;
  int *LesEtapes;
  int nLesEtapes;
  int **ChoixEnumActions;
  int **EvalEnumActions;
  int *ChoixEnumFacts;
  float *EvalEnumFacts;
  int *iFaitsInvalides, *sFaitsInvalides;

  /* Level ordering */
  int LevelOrdering;
  int * DGLevel; 
  int ** DGL;
  int DGState;

  /* Plan paralleles... */
  int ComputeParallelPlan;
  int * StepLevel;
}
  S_GR;

extern S_GR * GRT;

void AllocGrapheDActionThreads(int ne, int thread);
int removeFactList(int s, int *I, int t);
void InitGrapheDActionsThreads(int ns, int *I, int *G, int t);
void AllocTheGoals(int nf);
void AllocGraphActionLevelThreads(int na, int i, int thread);
void AllocGraphFactLevelThreads(int na, int nf, int i, int thread);
void InitGenericFactStateThreads(int s, int na, int nf, int ns, int thread);
void InitGenericActionStateThreads(int s, int ns, int thread);
int evaluatePlanCost (int ns, int thread);


/* Prototypes. */
void fixGoals (int ns, int *G, int t);
int countFixedGoals (int s, int * goals, int t);


int AddNewAction(char *s, FLIST P, FLIST A, FLIST S, int cost, int keepActionsWithOppositeEffects);
int *AddFactsFromList(FLIST l, int *n);
int *AddFactsFromLists(FLIST lp, FLIST lm, int *n);
int *AddFactsFromBothLists(FLIST lp, FLIST lm, int *n);
int ListesDisjointes(int *a, int *s);
int ListeIncluse(int *a, int *s);
int ListesFiltreRedundancies(int *a);
int removeSimultaneousOppositeEffects(int *a);
int nFaitsCommuns(int *s, int *p);
int SearchFact(char *s, int signe);
int AddFact(char *s, int signe, int but);
void AllocTablesActionsFaits(void);
void ConstructionDesTablesDeBooleensDeLAction(ACTION a, int nf);
void ConstructionTablesNewActions(int first, int last, int nf);
void MaJFactsActionsLists(int a);

void AfficheLeFait(FAIT f);
void AfficheLAction(ACTION a);
void printAction(int a);
void AfficheLeProbleme(void);
void AfficheLeGrapheDActions(int ne, int mode, int thread);
void AfficheSizesGraphe(int ne);
void AfficheLEtat(int e);
void printFactsAtLevel (int e, int thread);

/* FLIST CreerMaillon(int f, FLIST p, FLIST s); */
FLIST CreerFLMaillon(int f, FLIST p, FLIST s);
void RestitueFList(FLIST l);
int * FromListToBooleans(FLIST l, int n);
int * FromGoalsToBooleans(FLIST l, int n);


/* Macros. */

#define FNIL (-1)

#define NUM_FACT(f) ((f)-LesFaits)
#define NUM_ACTION(a) ((a)-LesActions)
#define THE_FACT(i) (LesFaits+(i))
#define THE_ACTION(i) (LesActions+(i))
#define OPPOSE(f) ((((f)%2)==1)?((f)-1):((f)+1))

#define INSERER_FLIST(p,l) { \
  p->suiv = l->suiv;\
  l->suiv->prec = p;\
  l->suiv = p;\
  p->prec = l;\
}

/* Macro d'ajout de f a l'etape s dans les listes P des faits qui doivent persister
   (pas d'action qui les ajoute). */
#define ADD_FLIST_NOTHREADS(f, s) {			\
  FLIST lADD_FLIST = &(P[(s)][nbAM[(s)][(f)]]);\
  FLIST pADD_FLIST = &(TheM[(s)][(f)]);\
  INSERER_FLIST(pADD_FLIST,lADD_FLIST);\
}
#define RETR_FLIST_NOTHREADS(f, s) {			\
  FLIST pADD_FLIST = &(TheM[(s)][(f)]);\
  pADD_FLIST->prec->suiv = pADD_FLIST->suiv;\
  pADD_FLIST->suiv->prec = pADD_FLIST->prec;\
  pADD_FLIST->suiv = NULL;\
}
#define IS_IN_FLIST_NOTHREADS(f, s) (TheM[(s)][(f)].suiv != NULL)
#define ADD_FLIST(f, s, G) {			\
  FLIST lADD_FLIST = &(G->P[(s)][G->nbAM[(s)][(f)]]);\
  FLIST pADD_FLIST = &(G->TheM[(s)][(f)]);\
  INSERER_FLIST(pADD_FLIST,lADD_FLIST);\
}
#define RETR_FLIST(f, s, G) {			\
  FLIST pADD_FLIST = &(G->TheM[(s)][(f)]);\
  pADD_FLIST->prec->suiv = pADD_FLIST->suiv;\
  pADD_FLIST->suiv->prec = pADD_FLIST->prec;\
  pADD_FLIST->suiv = NULL;\
}
#define IS_IN_FLIST(f, s, G) (G->TheM[(s)][(f)].suiv != NULL)


/* Macros d'affichage. */
#define PRINT_FACT(s,f,i) printf("%s%c%s @ F%d\n", (s), ((f)%2==0)?'+':'-', (LesFaits+(f))->nom, (i));
#define PRINT_ACTION(s,a,i) printf("%s%s @ A%d\n", (s), (LesActions+(a))->nom, (i));


#endif
